{  ppmdb -- quick access proper motion database
   Version 20210727
   Copyright (C) 2020-2021 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{$mode objfpc}
program mkppmdb;

uses sysutils,zlib,getparam,ppmdb,dateutils;

{ ************************************************************************* }
{ ***** main ************************************************************** }
{ ************************************************************************* }
const bufsize = 8*1024;
      mag_n   = 20;

{ returns false on error, result will be unchanged }
function val2(const s:shortstring; var v:longint):boolean;
var w  : word;
    v2 : longint;
begin
    val(s,v2,w);
    result:=w=0;
    if result then v:=v2;
end;

function val2(const s:shortstring; var v:double):boolean;
var w  : word;
    v2 : double;
begin
    val(s,v2,w);
    result:=w=0;
    if result then v:=v2;
end;


var { parameters }
    par                            : tparams;
    dir,ppmxl                      : ansistring;
    dec_n,dec_i1,dec_i2,epoch,pmma : longint;
    mm_i                           : longint;
    pm_f                           : double; // 3600*1000/pmma
    mm                             : double;
    { i/o, classes }
    indexer                        : CPPMDB_Indexer;
    gzf                            : gzfile;
    { parsing variables }
    buf                            : array[0..bufsize-1] of char;
    i,j,k,len,res,line,col,flags   : longint;
    scol                           : shortstring;
    err,miv,mb0v                   : boolean;
    ra,dec,pmra,pmdec,mi,mb,mb0,mr : double;
    r                              : double;
    mbv,mrv                        : longint;
    i_pm,i_m0,i_mm,ni              : longint; // ignored / not ignored sources
    { DB creator }
    tiles                          : ^CPPMDB_Creator;
    te                             : TPPMDB_TEntry;
    dec_j,ra_j,ra_n                : longint;
    dec_i,ra_i                     : word;
    ofn                            : ansistring;

begin
    with par do
    begin
        init;
        dir:='.';     defparam('-d','Database director. Default: current directory',dir);
        dec_n:=36;    defparam('-n','Number of tiles in DEC direction. Default: 36 (5° tile height).',dec_n);
        dec_i1:=0;    defparam('-i1','Start DEC index for database generation. Default: 0. Used to to save memory',dec_i1);
        dec_i2:=-1;   defparam('-i2','End DEC index for database generation. Default: n-1. Used to to save memory',dec_i2);
        mm:=22;       defparam('-m','Maximim magnitude. Default: 22',mm);
        epoch:=-1;    defparam('-e','Reference year. Default: current year + 3',epoch);
        pmma:=20;     defparam('-p','PM resolution in mas/year. Default: 20',pmma);
        ppmxl:='';    defparam('-ppmxl','Convert from PPMXL format',ppmxl);
                      defparam('-h','Help',helpf);
        run;
        done;
    end;

    if epoch<0 then
    begin
        epoch:=yearof(now)+3;
        writeln('Reference year: ',epoch);
    end;

    pm_f:=3600*1000/pmma;
    mm_i:=round((mm+2)*mag_n+0.49);
    if dec_i1<0 then dec_i1:=0;
    if (dec_i2<0) or (dec_i2>=dec_n) then dec_i2:=dec_n-1;
    if dec_i2<dec_i1 then
    begin
        writeln(stderr,'Warning: i2<i1. Nothing to do.');
        halt(0);
    end;

    if ppmxl='' then
    begin
        writeln(stderr,'Warning: No input defined. Nothing to do');
        halt(0);
    end;

    // create the classes
    indexer:=CPPMDB_Indexer.create(dec_n);
    getmem(tiles,sizeof(pointer)*(dec_i2-dec_i1+1)*dec_n*2);
    for dec_i:=dec_i1 to dec_i2 do
    begin
        ra_n:=indexer.ra_n(dec_i);
        for ra_i:=0 to ra_n-1 do
            tiles[(dec_i-dec_i1)*dec_n*2 + ra_i]:=CPPMDB_Creator.create(epoch,dec_n,ra_i,dec_i,mag_n,pmma,mm_i, 'r','b','i');
    end;

    // read PPMXL format
    if ppmxl<>'' then
    begin
        gzf:=gzopen(@ppmxl[1],'rb');
        if gzf=nil then
        begin
            writeln(stderr, 'Error opening file ''',ppmxl,'''');
            halt(1);
        end;

        indexer:=CPPMDB_Indexer.create(dec_n);

        line:=1; col:=1; scol:=''; err:=false;
        i_pm:=0; i_m0:=0; i_mm:=0; ni:=0;
        repeat
            len:=0;
            repeat
                res:=gzread(gzf,@buf[len],bufsize-len);
                if res>0 then len+=res;
            until (res<=0) or (len>=bufsize);
            if len>bufsize then len:=bufsize;

            // parse the data
            for i:=0 to len-1 do
            begin
                if (buf[i]=#10) or (buf[i]='|') then
                begin
                    if not err then
                    begin
                        case col of
                            2 : err:=not val2(scol,ra);
                            3 : err:=not val2(scol,dec);
                            6 : val2(scol,pmra);
                            7 : val2(scol,pmdec);
                            19: mb0v:=val2(scol,mb0);
                            20: if val2(scol,mb) then mbv:=1;
                            21: if val2(scol,mr) then mrv:=1;
                            22: if val2(scol,r) then begin mr+=r; mrv+=1 end;
                            23: miv:=val2(scol,mi);
                            25: err:=not val2(scol,flags);
                        end;
                        if err then writeln(stderr,'Error parsing line ',line,', column ',col,': ',scol)
                        else if col=25 then
                        begin
                            { adjust epoch }
                            r:=cos(dec/180*pi);
                            if r<1e-10 then r:=1e-10;
                            dec+=(epoch-2000)*pmdec;
                            ra+=(epoch-2000)*pmra/r;
                            ppmdb_normalize(ra,dec);
                            { calculate dec index and position }
                            dec_j:=round((90-dec)/180*dec_n*65536);
                            if dec_j<0 then
                            begin
                                writeln(stderr,'Internal error at line ', line,': dec_j=',dec_j);
                                dec_j:=0;
                            end;
                            j:=dec_n*65536;
                            if dec_j>=j then
                            begin
                                if dec_j>j then writeln(stderr,'Internal error at line ', line,': dec_j=',dec_j);
                                dec_j:=j;  // dec_j=dec_n*65536 may happen
                            end;
                            dec_i:=dec_j shr 16;
                            te.dec:=dec_j and 65535;
                            if (dec_i>=dec_i1) and (dec_i<=dec_i2) then
                            begin
                                { calculate dec index and position }
                                ra_n:=indexer.ra_n(dec_i);
                                j:=ra_n*65536;
                                ra_j:=round(ra/360*j) mod j;
                                if ra_j<0 then ra_j+=j;  // -(a mod b) = (-a) mod b
                                ra_i:=ra_j shr 16;
                                te.ra:=ra_j and 65535;
                                { convert proper motions }
                                j:=round(pmra*pm_f);
                                k:=round(pmdec*pm_f);
                                if (j<-128) or (j>128) or (k<-128) or (k>127) then
                                begin
                                    i_pm+=1;
//                                  writeln('Warning: Line ',line,' ignored because source is moving to fast: pm_ra=',round(pmra*3600*1000),' mas/y  pm_dec=',round(pmdec*3600*1000),' mas/y');
                                end
                                else begin
                                    te.pmra:=j;
                                    te.pmdec:=k;
                                    { convert magnitudes }
                                    if flags and 2=2 then miv:=false // from PPMX, mi is Johnson V mag. (green)
                                    else if mb0v then   // catalog mag. if from PPMX
                                    begin
                                        mb+=mb0;
                                        mbv+=1;
                                    end;
                                    if mbv>0 then te.m[1]:=round((2+mb/mbv)*mag_n)
                                    else te.m[1]:=-1;
                                    if mrv>0 then te.m[0]:=round((2+mr/mrv)*mag_n)
                                    else te.m[0]:=-1;
                                    if miv then te.m[2]:=round(mi*mag_n)
                                    else te.m[2]:=-1;
//                                  if (line=1) {or ((flags and 2)=2)} then writeln(line,'  ',te.m[0],'  ',te.m[1],'   ',te.m[0]);
                                    { sort out sources and save the rest }
                                    if mrv<1 then
                                    begin
                                        i_m0+=1;
                                        if (flags and 2>0) and mb0v and (mb0<8) then writeln('Warning: No red magnitude defined for source at line ',line,'  cmag=', mb0:0:1)
{                                       write(stderr,'Warning: No red magnitude defined for source at line ',line);
                                        if mbv>0 then write(stderr, ', bmag=', te.m[1]/mag_n-2:0:1);
                                        if miv then write(stderr, ', imag=', te.m[2]/mag_n-2:0:1);
                                        if (flags and 2>0) and mb0v then write(stderr, ', cmag=', mb0:0:1);
                                        writeln(stderr); }
                                    end
                                    else if te.m[0]>mm_i then
                                    begin
                                        i_mm+=1;
//                                      writeln(stderr,'Warning: Source at line ',line,' ignored because to faint: mag=', te.m[0]/mag_n-2:0:2);
                                    end
                                    else begin
                                        tiles[(dec_i-dec_i1)*dec_n*2 + ra_i].append(@te);
                                        ni+=1;
                                    end;

                                end;
                            end;
                        end
//                      else if col=25 then writeln(stderr,'Warning: No red magnitude. Ignoring line ',line,'   flags=',flags,'   ',mbv,'   ',miv);
                    end;

                    if buf[i]='|' then col+=1
                    else begin
                        col:=1;
                        line+=1;
                    end;
                    if col=2 then  // also resets data for first line
                    begin
                        err:=false;
                        pmra:=0;
                        pmdec:=0;
                        mbv:=0;
                        mb0v:=false;
                        mrv:=0;
                        miv:=false;
                        mb:=0;
                        mr:=0;
                    end;
                    scol:='';
                end
                else if buf[i]>#32 then scol+=buf[i];
            end;
            write(stderr,#13,'Read ',(line-1)*1e-6:0:3,' M lines (dropped: ',100*(1-ni/(1e-6+ni+i_pm+i_m0+i_mm)):0:2,'%)                        ',#13);
        until (len<bufsize) or (res<=0) {or (line>1e6)};
        gzclose(gzf);
    end;
    writeln(stderr,'                                                                                   ');
    i:=ni+i_pm+i_m0+i_mm;
    writeln('Processed ',i,' of ', line-1, ' sources (',i/(line-1)*100:0:1,'%)');
    if i>0 then
    begin
        writeln('Ignored because no reg magnitude : ',i_m0:7,'  (',i_m0/i*100:5:2,'%)');
        writeln('Ignored because to faint         : ',i_mm:7,'  (',i_mm/i*100:5:2,'%)');
        writeln('Ignored because moving to fast   : ',i_pm:7,'  (',i_pm/i*100:5:2,'%)');
        writeln('Used                             : ',  ni:7,'  (',  ni/i*100:5:2,'%)');
    end;

    // write data and dispose anything
    for dec_i:=dec_i1 to dec_i2 do
    begin
        ra_n:=indexer.ra_n(dec_i);
        for ra_i:=0 to ra_n-1 do
        begin
            i:=(dec_i-dec_i1)*dec_n*2 + ra_i;
            if tiles[i].size>0 then
            begin
                ofn:=dir+pathDelim+ppmdb_fileBase(ra_i,dec_i)+'.pdb';
                writeln('Writing ',ofn);
                tiles[i].write(ofn);
            end;
            tiles[i].destroy;
        end;
    end;
    indexer.destroy;
    freemem(tiles,sizeof(pointer)*(dec_i2-dec_i1+1)*dec_n*2);
end.
