{  simg -- command line applications for deterministic image processing
   Version 20210727
   Copyright (C) 1999-2021 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{@name convert}
{@description Conversion utility}

{$I directives.pas}
uses simage,getparam,sysutils;
var par                            : TParams;
    cc                             : CSImg_Config;
    ri                             : TSImg_RDImage;
    mi                             : CSImg_MetaInfo;
    b8,b16,b32,b64,bobe,bole       : boolean;
    pnm,sif,fits,pnmle,q,fh        : boolean;
    i,j,x1,y1,x2,y2                : longint;
    sc,bias,r                      : double;
    fi,fo,mf,sa,key,comment        : ansistring;
    fmt                            : TSImg_RawFormat;
    ma,mr,md                       : TAAStr;

begin
    cc:=CSImg_Config.create(SImg_Config_Default);
    with par do
    begin
        init;
        paramusage:='[<infile> [<outfile>]]';
        fi:='';      defparam('<infile>','Input image (default: stdin)',fi);
        fo:='';      defparam('<outfile>','Output image (default: stdout)',fo);
                     defparam('-h','Help',helpf);
        pnmle:=false;defparam('-pnmle','Byte order of PNM input files is little endian',pnmle);
                     defparam('-8','Force 8 bit output',b8);
                     defparam('-16','Force 16 bit output',b16);
                     defparam('-32','Force single precision output',b32);
                     defparam('-64','Force double precision output',b64);
        bobe:=false; defparam('-be','Force big endian output (default: FITS/PNM: big endian)',bobe);
        bole:=false; defparam('-le','Force little endian output (default: FITS/PNM: big endian)',bole);
        bias:=0;     defparam('-bias','Subtract this bias',bias);
        sc:=1;       defparam('-sc','Multiply image with scalar',sc);
        ma.size:=0;  defparam('-m','Add meta info (can be defined multiple times)',ma);  deffmt('-m','<key>=<value>[/<comment>]');
        md.size:=0;  defparam('-md','Delete meta info (can be defined multiple times)',md); deffmt('-md','<key>');
        mr.size:=0;  defparam('-mr','Rename meta info (can be defined multiple times)',mr); deffmt('-mr','<oldkey>:<newkey>');
        mf:='';      defparam('-mf','Read meta info from FITS file',mf); deffmt('-mf','<FITS file>');
        fits:=false; defparam('-fits','Force fits output',fits);
        fh:=false;   defparam('-fh','Force fits header output only',fh);
        pnm:=false;  defparam('-pnm','Force pnm output',pnm);
        sif:=false;  defparam('-sif','Force sif output',sif);
        x1:=0;       defparam('-x1','Cut image, first X coordinate',x1);
        y1:=0;       defparam('-y1','Cut image, first Y coordinate',y1);
        x2:=-1;      defparam('-x2','Cut image, last X coordinate',x2);
        y2:=-1;      defparam('-y2','Cut image, last Y coordinate',y2);
                     defparam('-o','Destination image',fo); deffmt('-o','<outfile>');
        run;

        i:=0;
        if fits then inc(i);
        if fh then inc(i);
        if pnm then inc(i);
        if sif then inc(i);
        if (i>1) then
        begin
            cc.warning('convert: More than one file format defined, ignoring all');
            fits:=false;
            fh:=false;
            pnm:=false;
            sif:=false;
        end;

        fmt:=SImg_RawByte; { makes the compiler happy }
        i:=0;
        if paramdef('-8') then
        begin
            fmt:=SImg_RawByte;
            i+=1;
        end;
        if paramdef('-16') then
        begin
            fmt:=SImg_RawWord;
            i+=1;
        end;
        if paramdef('-32') then
        begin
            fmt:=SImg_RawSingle;
            i+=1;
        end;
        if paramdef('-64') then
        begin
            fmt:=SImg_RawDouble;
            i+=1;
        end;
        if (i>1) then
        begin
            cc.warning('convert: More than one data format defined, ignoring all');
            i:=0;
        end;

        if bobe and bole then
        begin
            cc.warning('convert: More then one output byte order defined, ignoring all');
            bobe:=false;
            bole:=false;
        end;
        cc.forceBE:=bobe;
        cc.forceLE:=bole;
        if pnmLE then cc.pnmLE:=true;

        cc.debug('convert: values: input='+fi+' output='+fo+' DataFormat='+condstring(i>0,fmt2str(fmt),'Auto')+
            condstring(bobe,'-BE',condstring(bole,'-LE','')) + ' bias='+real2str(bias) + ' sc='+real2str(sc) +  ' mf=' + mf +
            ' FileFormat='+condstring(fh,'FITS header',condstring(fits,'FITS', condstring(pnm,'PNM', condstring(sif,'SIF', 'Auto'))))+
            ' x1='+int2str(x1)+' y1='+int2str(y1)+' x2='+int2str(x2)+' y2='+int2str(y2));

        cc.info('convert: Reading input image '+fi);
        mi:=CSImg_MetaInfo.create;
        SImg_ReadImageFile(cc,fi,ri,mi);

        { red meta infos from file }
        try
            if mf<>'' then SImg_RHeader(cc, mf, mi);
        except
            SImg_OutputCatchedException(cc);
        end;

        { convert format }
        if i>0 then
        begin
            cc.info('convert: Converting ' + fmt2str(ri.fmt) + ' --> ' + fmt2str(fmt) );
            SImg_Convert(cc,ri,ri,fmt,mi,true,false);
        end;

        { cut image }
        if x2<0 then x2:=ri.nx-1;
        if y2<0 then y2:=ri.ny-1;
        if (x1<>0) or (y1<>0) or (x2<>ri.nx-1) or (y2<>ri.ny-1) then
        begin
            cc.info('convert: Cutting image to ('+int2str(x1)+','+int2str(y1)+') .. ('+int2str(x2)+','+int2str(y2)+')');
            SImg_RDCut(cc, ri,ri, x1,y1,x2,y2);
        end;

        { delete meta info keys }
        for i:=0 to md.size-1 do
            mi.delete(trim(md.aa[i]));

        { rename meta info keys }
        for i:=0 to mr.size-1 do
        begin
            sa:=trim(mr.aa[i]);
            j:=pos(':',sa);
            if (j<=1) or (j>=length(sa))
            then cc.warning('Invalid format of -mr '+sa+', expected -mr <oldkey>:<newkey>')
            else mi.rename(trim(copy(sa,1,j-1)), trim(copy(sa,j+1,length(sa)-j)));
        end;

        if abs(bias)>1e-9 then
        begin
            if mi.get('DATAMIN',r,@comment) then mi.add('DATAMIN', real2str(r-bias,15), comment);
            if mi.get('DATAMAX',r,@comment) then mi.add('DATAMAX', real2str(r-bias,15) ,comment);
        end;

        if abs(sc-1)>1e-9 then
        begin
            if mi.get('DATAMIN',r,@comment) then mi.add('DATAMIN', real2str(r*sc,15), comment);
            if mi.get('DATAMAX',r,@comment) then mi.add('DATAMAX', real2str(r*sc,15) ,comment);
            if (abs(sc)>1e-20) and mi.get('EGAIN',r,@comment) then mi.add('EGAIN', real2str(r/sc,15) ,comment);
            if mi.get('BUNIT',sa,@comment) then mi.add('BUNIT', '''' + real2str(sc,8) + '*' + trimquotes(sa) + '''' ,comment);
            if mi.get('RDNOISE',r,@comment) then mi.add('RDNOISE', real2str(r*sc,15) ,comment);
        end;

        { add meta infos }
        for i:=0 to ma.size-1 do
        begin
            sa:=trim(ma.aa[i]);
            j:=pos('=',sa);
            if (j<=1) or (j>=length(sa))
            then cc.warning('Invalid format of -m '+sa+', expected -mr <key>=<value>[/<comment>]')
            else begin
                key:=trim(copy(sa,1,j-1));
                sa:=trim(copy(sa,j+1,length(sa)-j));
                j:=1;
                q:=false;
                while (j<=length(sa)) and (q or (sa[j]<>'/')) do
                begin
                    if sa[j]='''' then q:=not q;
                    inc(j);
                end;
                if j=1
                then cc.warning('Invalid format of -m '+sa+', <value> must not be empty, syntax: -m <key>=<value>[/<comment>] ')
                else begin
                    comment:='';
                    if j<length(sa) then comment:=trim(copy(sa,j+1,length(sa)-j));
                    if j=0 then j:=length(sa)+1;
                    mi.add(key, trim(copy(sa,1,j-1)), comment);
                end;
            end;
        end;

        { factor, bias }
        ri.cf*=sc;
        ri.co:=(ri.co-bias)*sc;

        cc.info('convert: Writing output image '+fo);
        if fits then SImg_WFITS(cc,CSImg_FileOutput.init(fo),ri,mi).close(cc)
        else if fh then SImg_WFITSHeader(cc,CSImg_FileOutput.init(fo),mi).close(cc)
        else if pnm then SImg_WPNM(cc,CSImg_FileOutput.init(fo),ri,mi)
        else if sif then SImg_WSIF(cc,CSImg_FileOutput.init(fo),ri,mi)
        else SImg_WriteImageFile(cc,fo,ri,mi);
        SImg_Free(cc,ri);

        done;
    end;
    cc.destroy;
end.
