{  simg -- command line applications for deterministic image processing
   Version 20210727
   Copyright (C) 1999-2021 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{@ame info}
{@description Print image information

{$I directives.pas}
uses simage,getparam;
var par                 : tparams;
    cc                  : CSImg_Config;
    img                 : TSImg_FPImage;
    si                  : TSImg_StatInfo;
    mi                  : CSImg_MetaInfo;
    dim,sv,mm           : boolean;
    sm,pmi,shfd         : boolean;
    sh1d,sh2d,sh3d,scd  : boolean;
    onlymeta            : boolean;
    fi,val              : ansistring;
    bps                 : shortstring;
    sh1,sh2,sh3,shf     : double;
    r,ca,cad,sc,nf      : double;
    i,j,x,y,xo,yo       : longint;
    s                   : shortstring;
    sharpness           : array of double;
    bi                  : TSImg_BayerInfo;
    vn,qa,pmn           : TAAStr;
    q,qv                : array of double;
    ri                  : TSImg_RDImage;

function nsm(s:simgstring):simgstring;
begin
    if sm then result:=''
    else result:=s;
end;

function nsm2(s:simgstring):simgstring;
begin
    if sm then result:=' '
    else result:=s;
end;

function val2(const s:shortstring; var i:longint):boolean;      // false on error
var w  : word;
begin
    system.val(s,i,w);
    result:=w=0;
end;

function val2(const s:shortstring; var d:double):boolean;       // false on error
var w  : word;
begin
    system.val(s,d,w);
    result:=w=0;
end;


begin
    with par do
    begin
        init;
        paramusage:='[<infile>]';
        fi:='';         defparam('<infile>','Source image (default: stdin)',fi);
                        defparam('-h','Help',helpf);
        sm:=false;      defparam('-sm','Script mode',sm);
        pmi:=false;     defparam('-mi','Print all meta info',pmi);
        pmn.size:=0;    defparam('-mn','Print meta info specified by name. Can be specified multiple times. Returns "(undefined)" if undefined',pmn);
        dim:=false;     defparam('-dim','Dimension of image and date format info',dim);
        mm:=false;      defparam('-mm','Print minimal value and maximal value',mm);
        qa.size:=0;     defparam('-q','Print quantile. Can be specified multiple times.',qa);
        vn.size:=0;     defparam('-v','Return brightness at that pixel coordinate in comma separated X,Y format. Top left pixel is 0,0. ',vn);
        sv:=false;      defparam('-sv','Statistic values (average value, variance, ...)',sv);
        ca:=-1;         defparam('-ca','Get color adjustment factor (normal mode) using this smoothing radius',ca);
        cad:=-1;        defparam('-cad','Get color adjustment factor (differential mode) using this smoothing radius',cad);
        nf:=-1;         defparam('-nf','Calculate noise floor. Parameter is the super-resolution factor. 2.5 is safe in almost any case',nf);
        shf:=2;         defparam('-sf','Star FWHM in pixel. Parameter is the Minimum FWHM. (As larger as better the noise immunity).\\'+
                                       'Values less then 1.5 should be avoided. Recommended: at least 2',shf);
        sh1:=0.4;       defparam('-s1','Sharpness indicators 1 (threshold as quantile; valid: 0..1; 0.999..1 for stars)\\'+
                                       'Edge detect algorithm, best choice if sensor resolution is higher than optical resolution',sh1);
        sh2:=0.4;       defparam('-s2','Sharpness indicators 2 (threshold as quantile; valid: 0..1, 0.999..1 for stars)\\'+
                                       'Edge detect algorithm, optical resolution must be close or higher than sensor resolution', sh2);
        sh3:=0.95;      defparam('-s3','Sharpness indicators 3 (threshold as quantile; valid: 0..1, feasible: 0.999..1)\\'+
                                       'Peak detect algorithm optimized for stars, optical resolution must be close to or higher than sensor resolution',sh3);
        sc:=1.0;        defparam('-sc','Calculates differential color sharpness. The parameter is the quotient between multipliers of first and last color\\'+
                                       'This just calculates <sharpness first color>/<sharpness last color>*<-sc parameter> - 1\\'+
                                       'Has no effect for single color images or if neither -s1, s2, -s3 or -sf are given\\'+
                                       'In combination with -sf the parameter is ignored and output is <FWHM last color>/<FWHM first color>-1',sc);
        bps:='';        defparam('-bp','Bayer pattern for sharpness calculation',bps);
        run;
        shfd:=vardef(@shf);
        sh1d:=vardef(@sh1);
        sh2d:=vardef(@sh2);
        sh3d:=vardef(@sh3);
        scd:=vardef(@sc);
        done;
    end;
    onlymeta:=not (sv or mm or shfd or sh1d or sh2d or sh3d or (ca>-0.5) or (cad>-0.5) or (qa.size>0) or (vn.size>0) or (nf>simg_mina));
    if onlymeta and not (dim or pmi or (pmn.size>0)) then
    begin
        dim:=true;
        sv:=true;
        mm:=true;
        onlymeta:=false;
    end;

    cc:=CSImg_Config.create(nil);
    cc.debug(cc.name+': started');

    mi:=CSImg_MetaInfo.create();
    ri.pt:=nil;
    if onlymeta then
    begin
        cc.Info(cc.name+': Reading meta info of image '+fi);
        SImg_RHeader(cc, fi, ri, mi)
    end
    else begin
        cc.Info(cc.name+': Reading input image '+fi);
        SImg_ReadImageFile(cc, fi, ri, mi);
    end;

    if dim then writeln(nsm('Dimensions: ')+int2str(ri.nx)+nsm2('x')+int2str(ri.ny)+nsm2('x')+int2str(ri.nc)+nsm2('  format: ')+fmt2str(ri.fmt)+'-'+condstring(ri.bole,'LE','BE'));

    if pmi then
        for i:=0 to mi.size-1 do
            if mi.key[i]<>'' then
            begin
                fillchar(s,sizeof(s),#32);
                s[0]:=#12;
                s[11]:='=';
                move(mi.key[i][1],s[1],length(mi.key[i]));
                s+=mi.val[i];
                if (mi.comment[i]<>'') and (length(s)<252) then
                begin
                    if length(s)<35 then s[0]:=#35;
                    s+=' / '+mi.comment[i];
                end;
                writeln(s);
            end;

    for i:=0 to pmn.size-1 do
    begin
        dim:=mi.get(pmn.aa[i],val);
        writeln(nsm(pmn.aa[i]+' = ')+select(dim, trimquotes(val), '(undefined)'));
    end;

    if onlymeta then
    begin
        mi.destroy;
        cc.destroy;
        SImg_Halt;
    end;

    img.pt:=nil;
    SImg_Convert(cc, ri,img, mi, true,false);

    if vn.size>0 then
        for i:=0 to vn.size-1 do
        begin
            s:=trim(vn.aa[i]);
            j:=pos(',',s);
            if (j<=1) or (j>=length(s)) or (not val2(copy(s,1,j-1),x)) or (not val2(copy(s,j+1,length(s)-j),y)) then cc.error('Invalid format of -v '+s+', expected -v <X>,<Y>');
            xo:=x; yo:=y;
            if x<0 then x:=0;
            if x>=img.nx then x:=img.nx-1;
            if y<0 then x:=0;
            if y>=img.ny then y:=img.ny-1;
            if (x<>xo) or (y<>yo) then cc.warning('Pixel coordinates out of range. Cropped ('+int2str(xo)+','+int2str(yo)+') to ('+int2str(x)+','+int2str(y)+')' );
            if not sm then write('Value(s) at (' , x , ',' , y , '):');
            for j:=0 to img.nc-1 do
                write(' ',real2str(img.pt[y*img.sl + x*img.nc +j],6));
            writeln;
        end;

    if mm or sv then
    begin
        SImg_GetStatInfo(cc,img,si);
        with si do
        begin
            if mm then
            begin
                writeln(nsm('Minimal value: '),real2str(minv,5));
                writeln(nsm('Maximal value: '),real2str(maxv,5));
            end;
            if sv then
            begin
                writeln(nsm('Average value: '),real2str(av,5));
                writeln(nsm('Mean deviation: '),real2str(md,5));
                writeln(nsm('RMS: '),real2str(rms,5));
            end;
        end;
    end;

    if qa.size>0 then
    begin
        setlength(q, qa.size);
        setlength(qv, qa.size);
        for i:=0 to qa.size-1 do
            if not val2(qa.aa[i],q[i]) then cc.error(cc.name+': Number expected after -q, got "'+qa.aa[i]+'"');
        SImg_GetQuantile(cc, img, cc.threads, -1, q,qv);
        for i:=0 to qa.size-1 do
            writeln(nsm(real2str(q[i]*100,3)+'% Quantile: '),real2str(qv[i],5));
        setlength(q, 0);
        setlength(qv, 0);
    end;

    if ca>-0.5 then writeln(nsm('Color factors (normal mode): ')+SImg_ColorAdjustInfo(cc,img,false,ca));

    if cad>-0.5 then writeln(nsm('Color factors (differential mode): ')+SImg_ColorAdjustInfo(cc,img,true,cad));

    if nf>simg_mina then writeln(nsm('Noise floor: ')+real2str(SImg_NoiseFloor(cc,img,nf)));

    if shfd or sh1d or sh2d or sh3d then
    begin
        if (bps='') and mi.get('BAYERPAT',val) then bps:=trimquotes(val);
        if bps<>'' then
        begin
            bi.cf[0]:=1;
            bi.cf[1]:=1;
            bi.cf[2]:=1;
            bi.cf[3]:=1;
            bi.pattern:=SImg_BayerParsePattern(cc,bps);
            if not SImg_BayerCheckInfo(bi)
            then cc.warning(cc.name+': Invalid Bayer pattern')
            else begin
                cc.Info(cc.name+': Converting Bayer pattern image');
                SImg_BayerDownscale(cc,bi,img,img);
            end;
        end;
    end;

    setLength(sharpness,img.nc);
    if shfd then
    begin
        r:=SImg_GetSharpnessFWHM(cc,img, shf,round(shf*20+0.49), sharpness);
        write(nsm('FWHM:  '));
        if (img.nc>1) and (scd) and (sharpness[img.nc-1]>1e-10) then write(nsm('   color diff.:  '),real2str(sharpness[img.nc-1]/sharpness[0]-1,-3):8,nsm(',  RHS: '));
        write(real2str(r,-2):8);
        if img.nc>1 then
        begin
            if not sm then write(',  per color: ');
            for i:=0 to img.nc-1 do
                write(' ',real2str(sharpness[i],-2):8);
        end;
    writeln;
    end;

    if sh1d then
    begin
        r:=SImg_GetSharpness1(cc,img, sh1, sharpness);
        write(nsm('Sharpness1:  '));
        if (img.nc>1) and (scd) and (sharpness[img.nc-1]>1e-10) then write(nsm('   color diff.:  '),real2str(sharpness[0]/sharpness[img.nc-1]*sc-1,-3):8,nsm(',  geo.mean: '));
        write(real2str(r,5):8);
        if img.nc>1 then
        begin
            if not sm then write(',  per color: ');
            for i:=0 to img.nc-1 do
                write(' ',real2str(sharpness[i],5):8);
        end;
    writeln;
    end;

    if sh2d then
    begin
        r:=SImg_GetSharpness2(cc,img, sh2, sharpness);
        write(nsm('Sharpness2:  '));
        if (img.nc>1) and (scd) and (sharpness[img.nc-1]>1e-10) then write(nsm('   color diff.:  '),real2str(sharpness[0]/sharpness[img.nc-1]*sc-1,-3):8,nsm(',  geo.mean: '));
        write(real2str(r,5):8);
        if img.nc>1 then
        begin
            if not sm then write(',  per color: ');
            for i:=0 to img.nc-1 do
                write(' ',real2str(sharpness[i],5):8);
        end;
    writeln;
    end;

    if sh3d then
    begin
        r:=SImg_GetSharpness3(cc,img, sh3, sharpness);
        write(nsm('Sharpness3:  '));
        if (img.nc>1) and (scd) and (sharpness[img.nc-1]>1e-10) then write(nsm('   color diff.:  '),real2str(sharpness[0]/sharpness[img.nc-1]*sc-1,-3):8,nsm(',  geo.mean: '));
        write(real2str(r,5):8);
        if img.nc>1 then
        begin
            if not sm then write(',  per color: ');
            for i:=0 to img.nc-1 do
                write(' ',real2str(sharpness[i],5):8);
        end;
    writeln;
    end;

    SImg_Free(cc,img);
    cc.destroy;
end.
