{  simg -- command line applications for deterministic image processing
   Version 20210727
   Copyright (C) 1999-2021 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{@name maskop}
{@description Mask operations.}
{@longdescription Reads a 8 bit image and applies mask operations}

{$I directives.pas}
uses simage,getparam,sysutils;
var par         : TParams;
    cc          : CSImg_Config;
    ri          : TSImg_RDImage;
    fi          : TSImg_FPImage;
    mi          : CSImg_MetaInfo;
    ifn,ofn     : ansistring;
    f           : array[0..7] of simgfp;
    r,s,z,l,u,g : simgfp;
    i,m,b       : longint;

begin
    cc:=CSImg_Config.create(SImg_Config_Default);
    with par do
    begin
        init;
        paramusage:='[<infile> [<outfile>]]';
        ifn:='';     defparam('<infile>','( bit input image (default: stdin)',ifn);
        ofn:='';     defparam('<outfile>','Output image (default: stdout)',ofn);
        f[0]:=0;     defparam('-f0','Factor for bit 0. Default: 0',f[0]);
        f[1]:=0;     defparam('-f1','Factor for bit 1. Default: 0',f[1]);
        f[2]:=0;     defparam('-f2','Factor for bit 2. Default: 0',f[2]);
        f[3]:=0;     defparam('-f3','Factor for bit 3. Default: 0',f[3]);
        f[4]:=0;     defparam('-f4','Factor for bit 4. Default: 0',f[4]);
        f[5]:=0;     defparam('-f5','Factor for bit 5. Default: 0',f[5]);
        f[6]:=0;     defparam('-f6','Factor for bit 6. Default: 0',f[6]);
        f[7]:=0;     defparam('-f7','Factor for bit 7. Default: 0',f[7]);
        z:=0;        defparam('-z','Offset, default 0.',z);
        g:=-1;       defparam('-g','Filter radius fr Gaussian low-pass-filtering',g);
        l:=-simg_large; defparam('-l','Lower clipping limit. Applied after low-pass filtering.',l);
        u:=simg_large;  defparam('-u','Upper clipping limit. Applied after low-pass filtering.',u);
                     defparam('-o','Destination image',ofn); deffmt('-o','<outfile>');
                     defparam('-h','Help',helpf);
        run;
        done;
    end;
    cc.debug(cc.name+': input=' + ifn + ' output=' + ofn+ ' z=' + real2str(z) + ' l=' + real2str(l) + ' u=' + real2str(u)
        + ' f0=' + real2str(f[0])  + ' f1=' + real2str(f[1]) + ' f2=' + real2str(f[2]) + ' f3=' + real2str(f[3]) + ' f4=' + real2str(f[4]) + ' f5=' + real2str(f[5]) + ' f6=' + real2str(f[6]) + ' f7=' + real2str(f[7]) );

    { m=0: nothing to do. m>0 fast mode. m<0 slow mode }
    m:=0;
    r:=0;
    for i:=0 to 7 do
    begin
        f[i]/=1 shl i;
        if abs(f[i])>simg_mina then
        begin
            if m=0 then r:=f[i]
            else if abs(r-f[i])>simg_mina then
            begin
                m:=-1;
                break;
            end;
            m:=m or (1 shl i);
        end;
    end;
    if m=0 then
    begin
        cc.warning(cc.name+': No factors are defined. Nothing to do.');
        SImg_Halt;
    end;
//    writeln(stderr,m,'   ',r:0:3);

    { input }
    cc.info(cc.name+': Reading input image '+ifn);
    mi:=CSImg_MetaInfo.create;
    ri.pt:=nil;
    SImg_ReadImageFile(cc,ifn,ri,mi);
    if ri.fmt<>SImg_RawByte then cc.error(cc.name+': 8 bit image expected');

    { factors, offset }
    fi.pt:=nil;
    SImg_New(cc, fi, ri.nx,ri.ny,ri.nc, false);
    if m>0 then
    begin
        { fast mode }
        cc.info(cc.name+': Performing mask operations (fast mode)');
        for i:=0 to ri.size-1 do
        begin
            s:=z + (ri.ptb[i] and m)*r;
            fi.pt[i]:=s;
        end;
    end
    else begin
        cc.info(cc.name+': Performing mask operations (slow mode)');
        for i:=0 to ri.size-1 do
        begin
            b:=ri.ptb[i];
            s:=z;
            s+=(b and   1)*f[0];
            s+=(b and   2)*f[1];
            s+=(b and   4)*f[2];
            s+=(b and   8)*f[3];
            s+=(b and  16)*f[4];
            s+=(b and  32)*f[5];
            s+=(b and  64)*f[6];
            s+=(b and 128)*f[7];
            fi.pt[i]:=s;
        end;
    end;
    SImg_Free(cc,ri);

    { gaussian filer }
    if g>1e-3 then
    begin
        cc.info(cc.name+': Gaussian low-pass filtering using ' + int2str(cc.threads) + ' threads');
        SImg_Filter(cc, fi,fi,g);
    end;

    { clipping }
    if (l>-0.9*simg_large) or (u<0.9*simg_large) then
    begin
        cc.info(cc.name+': Clipping values');
        for i:=0 to fi.size-1 do
        begin
            s:=fi.pt[i];
            if s<l then fi.pt[i]:=l;
            if s>u then fi.pt[i]:=u;
        end;
    end;

    { output }
    cc.info(cc.name+': Writing output image '+ofn);
    SImg_WriteImageFile(cc,ofn,fi,mi);
    SImg_Free(cc,fi);
    mi.destroy;
    cc.destroy;
end.
