{  simg -- image processing library
   Version 20210727
   Copyright (C) 1999-2021 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

// define[approx_debug]

{*************************************************************************}
{******* SImg_FBGApprox **************************************************}
{*************************************************************************}
{ background approximation
    rad    : approximation radii (geometric average of x and y values)
    xyr    : x/y ratio of approximation size.
    cs     : relative approximation size in corner (feasible: 0..1)
    sf     : subtraction factors. If unset and length of rad is 1, the approximation is returned. Otherwise length must be equal to length of rad.
    frad   : radius for quantile pre-filter (<=0 to disable this feature)
    fquant : quantile for pre-filter (0.5 for median)
    fg     : approximate foreground instead of background
}
procedure SImg_FBGApprox(cc:CSImg_Config; fi:TSImg_FPImage; var fj:TSImg_FPImage; mask:TSImg_RDImage; rad:array of double; xyr,cs:double; sf:array of double; frad,fquant:double; fg:boolean);
var fk,fl,fm    : TSImg_FPImage;
    i,j,irad,ss : longint;
    r           : simgfp;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_BgApprox: fi='+SImg_InfoStr2(fi)+' fj='+SImg_InfoStr2(fj)+' mask='+SImg_InfoStr2(mask)+' rad.size='+int2str(high(rad)+1)+' xyr='+real2str(xyr,-1)+' cs='+real2str(cs,-1)
        +' sf.size='+int2str(high(sf)+1)+' frad='+real2str(frad,-1)+' fquant='+real2str(fquant,-1)+' fg='+boolean2str(fg));
    if not SImg_Recover(cc,fi) then cc.error('SImg_BgApprox: Invalid image: fi');
    if mask.pt<>nil then
    begin
        if not SImg_Recover(cc,mask) then cc.error('SImg_FBGFilter: Invalid mask');
        if mask.fmt<>SImg_RawByte then cc.error('SImg_FBGFilter: Invalid data format of mask');
        if (mask.nx<>fi.nx) and (mask.ny<>fi.ny) and (mask.nc<>1) then cc.error('SImg_FBGFilter: Invalid dimensions of mask');
    end;
    if not SImg_OverwriteAlert(fi,fj) then SImg_Free(cc,fj);
    if (high(sf)<>high(rad)) and ((high(sf)>=0) or (high(rad)<>0)) then cc.error('SImg_FBGFilter: Number of radii must be equal to number of factors');
    xyr:=sqrt(abs(xyr));
    if xyr<1e-3 then xyr:=1e-3;

    { pre filter }
    fk.pt:=nil;
    irad:=round(frad-0.45);
    if irad<0 then irad:=0;
//    ss:=round(sqrt(irad*2+0.3));
    ss:=(irad*2+1) div 3;
    for j:=0 to high(rad) do
    begin
        if xyr>1 then i:=round(rad[j]/xyr*0.25)
        else i:=round(rad[j]*xyr*0.25);
        if i<ss then ss:=i;
    end;
    if ss<1 then ss:=1;
    if irad>0 then SImg_Filter(cc,fi,fk,irad+0.5,-2,ss, fquant)
    else SImg_Duplicate(cc,fi,fk,true);

    if high(rad)=0 then
    begin
        { approximations }
        SImg_FBGFilter(cc, fk,fk, mask, rad[0]*xyr/ss,rad[0]/(xyr*ss),cs, not fg,false);

        { re-scale }
        if ss>1 then SImg_Resize(cc, fk,fk, fi.nx,fi.ny);

        { subtraction }
        if high(sf)>=0 then
        begin
            r:=sf[0];
            for i:=0 to fi.size-1 do
                fk.pt[i]:=fi.pt[i]-fk.pt[i]*r;
        end;
    end
    else begin
        fl.pt:=nil; fm.pt:=nil;
        if irad>0 then SImg_Duplicate(cc,fk,fm,true)
        else SImg_Free(cc,fj);
        for j:=0 to high(rad) do
        begin
            { approximation }
            SImg_FBGFilter(cc, fk,fl, mask, rad[j]*xyr/ss,rad[j]/(xyr*ss),cs, not fg,false);

            { subtraction }
            r:=sf[j];
            for i:=0 to fk.size-1 do
                fk.pt[i]-=fl.pt[i]*r;
        end;
        SImg_Free(cc, fl);
        if fm.pt<>nil then
        begin
            for i:=0 to fk.size-1 do
                fk.pt[i]:=fm.pt[i]-fk.pt[i];
            SImg_Free(cc, fm);
            if ss>1 then SImg_Resize(cc, fk,fk, fi.nx,fi.ny);
            for i:=0 to fi.size-1 do
                fk.pt[i]:=fi.pt[i]-fk.pt[i];
        end;
    end;
    SImg_Move(cc,fk,fj);

    SImg_DecDebugLevel;
end;


{*************************************************************************}
{******* SImg_BgApprox2 **************************************************}
{*************************************************************************}
{ constant background approximation
    res    : result buffer for at least fi.nc values
}
procedure SImg_BgApprox2(cc:CSImg_Config; fi:TSImg_FPImage; res:psimgfp);
var  fj           : TSImg_FPImage;
     d1,d2,d3,d4  : ^simgfp;
     x,y,c        : longint;
     r            : simgfp;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_BgApprox2 fi='+SImg_InfoStr2(fi));
    if not SImg_Recover(cc,fi) then cc.error('SImg_BgApprox2: Invalid image: fi');

    for c:=0 to fi.nc-1 do
        res[c]:=simg_large;

    SImg_Filter(cc, fi,fj, 4.5, -2, 9, 0.5);
    for y:=0 to fj.ny-2 do
    begin
        d1:=fj.pt + y*fj.sl;
        d2:=d1 + fj.nc;
        d3:=d1 + fj.sl;
        d4:=d2 + fj.sl;
        for x:=0 to fj.nx-2 do
            for c:=0 to fj.nc-1 do
            begin
                r:=d1^+d2^+d3^+d4^;
                d1+=1; d2+=1; d3+=1; d4+=1;
                if r<res[c] then res[c]:=r;
            end;
    end;
    SImg_Free(cc,fj);

    for c:=0 to fi.nc-1 do
        res[c]*=0.25;

    SImg_DecDebugLevel;
end;


{ constant background approximation and subtraction }
procedure SImg_BgApprox2(cc:CSImg_Config; fi:TSImg_FPImage);
var res,d : ^simgfp;
    i,c   : longint;
begin
    getmem(res,fi.nc*sizeof(simgfp));
    SImg_BgApprox2(cc,fi,res);
    d:=fi.pt;
    for i:=0 to fi.nx*fi.ny-1 do
    begin
        for c:=0 to fi.nc-1 do
            d[c]-=res[c];
        d+=fi.nc;
    end;
    freemem(res,fi.nc*sizeof(simgfp));
end;

{*************************************************************************}
{******* SImg_Approx *****************************************************}
{*************************************************************************}
{ calculate scalar products
    ya,yz     : region for this thread
    n         : highest order of radial polynomials
    sp        : scalar products (results)
    a         : factors of previous solutions
    dev       : deviation
   polynomials:
    p_0(x,y)=x
    p_1(x,y)=y
    p_2(x,y)=x*y
    p_(3+i)=(x^2+y^2)^i
    p_0 to p_2 are orthogonal to all other polynomials
}
threadprocedure SImg_approx_thread1(fi:TSImg_FPImage; ya:longint; yz:longint; n:longint; sp:pdouble; a:pdouble; dev:pdouble );
var i,j,k,c     : longint;
    x,y,r,s,t   : double;
    p3,pr       : double;
    d           : ^simgfp;
    e,f         : ^double;
begin
    for i:=0 to fi.nc*(n+4)-1 do
        sp[i]:=0;

    d:=fi.pt + ya*fi.sl;
    for i:=ya to yz do
    begin
        y:=i-(fi.ny-1)*0.5;
        for j:=0 to fi.nx-1 do
        begin
            x:=j-(fi.nx-1)*0.5;
            p3:=x*y;
            pr:=sqr(x)+sqr(y);
            e:=sp;
            f:=a;
            for c:=0 to fi.nc-1 do
            begin
                { cap limit }
                t:=dev[c];
                s:=f[0]*x + f[1]*y +  f[2]*p3 + f[3];   f+=4;
                r:=1;
                for k:=0 to n-1 do
                begin
                    r*=pr;
                    s+=f[k]*r;
                end;
                f+=n;
                r:=d^;  d+=1;
                if r>s+0.25*t then r:=s+0.25*t
                else if r<s-t then r:=s-t;
                { scalar product }
                e[0] += r*x;
                e[1] += r*y;
                e[2] += r*p3;
                e[3] += r;
                e+=4;
                for k:=0 to n-1 do
                begin
                    r*=pr;
                    e[k]+=r;
                end;
                e+=n;
            end;
        end;
    end;
end;

{ calculate scalar products
    ya,yz     : region for this thread
    n         : highest order of radial polynomials
    a         : factors of solutions
    dm        : difference mode
    ss        : step size
   polynomials:
    p_0(x,y)=x
    p_1(x,y)=y
    p_2(x,y)=x*y
    p_(3+i)=(x^2+y^2)^i
}
threadprocedure SImg_approx_thread2(fi:TSImg_FPImage; ya:longint; yz:longint; n:longint; a:pdouble; dm:boolean; ss:longint );
var i,j,k,c   : longint;
    x,y,xc,yc : double;
    r,s,rs    : double;
    p3,pr     : double;
    f1,f2     : double;
    d         : ^simgfp;
    f         : ^double;
begin
    d:=fi.pt + ya*fi.sl;
    if dm then
    begin
        f1:=1;
        f2:=-1;
    end
    else begin
        f1:=0;
        f2:=1;
    end;
    if ss<1 then ss:=1;
    rs:=1/ss;
    xc:=((fi.nx-1) div ss)*0.5 + (((fi.nx-1) mod ss) div 2)*rs;
    yc:=((fi.ny-1) div ss)*0.5 + (((fi.ny-1) mod ss) div 2)*rs;
    for i:=ya to yz do
    begin
        y:=i*rs-yc;
        for j:=0 to fi.nx-1 do
        begin
            x:=j*rs-xc;
            p3:=x*y;
            pr:=sqr(x)+sqr(y);
            f:=a;
            for c:=0 to fi.nc-1 do
            begin
                r:=f[0]*x + f[1]*y +  f[2]*p3 + f[3];   f+=4;
                s:=1;
                for k:=0 to n-1 do
                begin
                    s*=pr;
                    r+=f[k]*s;
                end;
                f+=n;
                d^:=d^*f1 + f2*r;  d+=1;
            end;
        end;
    end;
end;

{ calculate standard deviation
    ya,yz     : region for this thread
    n         : highest order of radial polynomials
    buf       : output buffers
    a         : factors of previous solutions
   polynomials:
    p_0(x,y)=x
    p_1(x,y)=y
    p_2(x,y)=x*y
    p_(3+i)=(x^2+y^2)^i
    p_0 to p_2 are orthogonal to all other polynomials
}
threadprocedure SImg_approx_thread3(fi:TSImg_FPImage; ya:longint; yz:longint; n:longint; buf:pdouble; a:pdouble );
var i,j,k,c  : longint;
    x,y,r,s  : double;
    p3,pr    : double;
    d        : ^simgfp;
    f        : ^double;
begin
    for i:=0 to fi.nc-1 do
        buf[i]:=0;

    d:=fi.pt + ya*fi.sl;
    for i:=ya to yz do
    begin
        y:=i-(fi.ny-1)*0.5;
        for j:=0 to fi.nx-1 do
        begin
            x:=j-(fi.nx-1)*0.5;
            p3:=x*y;
            pr:=sqr(x)+sqr(y);
            f:=a;
            for c:=0 to fi.nc-1 do
            begin
                { cap limit }
                s:=f[0]*x + f[1]*y +  f[2]*p3 + f[3];   f+=4;
                r:=1;
                for k:=0 to n-1 do
                begin
                    r*=pr;
                    s+=f[k]*r;
                end;
                f+=n;
                buf[c]+=abs(d^-s);
                d+=1;
            end;
        end;
    end;
end;

{ approximate an image by polynomials
    rr   : Factor for radial polynomial
    r1   : Factor for 1st order polynomial (x and y)
    r1b  : Factor for 1st order bi-polynomial (x*y)
    n    : highest order of radial polynomials ((x-xc)^2+(y-yc)^2)^i
    it   : background approximation iterations (cap values to previous approximation)
    dm   : false: approximation replaces input
           true: difference replaces input
    frad : radius median pre-filter, <=0 to disable it
   polynomials:
    p_0(x,y)=x
    p_1(x,y)=y
    p_2(x,y)=x*y
    p_(3+i)=(x^2+y^2)^i
    p_0 to p_2 are orthogonal to all other polynomials
}
procedure SImg_Approx(cc:CSImg_Config; fj:TSImg_FPImage; rr,r1,r1b:double; n,it:longint; dm:boolean; frad:double);
var d          : array[0..2] of double;         { diagonal elements d_i = <p_i,p_i>, factor }
    ma,vx,vy   : matrix;
    fi         : TSImg_FPImage;
    xo,yo,x,y  : double;
    pc,s,r     : double;
    sp,a,dev   : ^double;                       { buffer for <p_i,f> and approximation coefficients }
    i,j,k,c,tn : longint;
    ya,yz      : longint;
    irad,ss    : longint;
threadvars;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_Approx: fj='+SImg_InfoStr2(fj)+' rr='+real2str(rr)+' r1='+real2str(r1)+' r1b='+real2str(r1b)+' n='+int2str(n)+' it='+int2str(it)+' dm='+boolean2str(dm)+' rfad='+real2str(frad,-1));
    if not SImg_Recover(cc,fj) then cc.error('SImg_Approx: Invalid image: fj');

    ma.pt:=nil;
    vx.pt:=nil;
    vy.pt:=nil;

    { pre-filter }
    fi.pt:=nil;
    irad:=round(frad-0.45);
    if irad<0 then irad:=0;
    ss:=round(sqrt(irad*2+0.3));
    if irad>0 then SImg_Filter(cc, fj,fi, irad+0.5, -2, ss, 0.5)
    else fi:=fj;

    { calculate d[0] }
    xo:=(1+(fi.nx and 1))*0.5;
    d[0]:=0;
    for i:=0 to (fi.nx-2) shr 1 do
        d[0]+=sqr(xo+i);
    d[0]*=2*fi.ny;

    { calculate d[1] }
    yo:=(1+(fi.ny and 1))*0.5;
    d[1]:=0;
    for i:=0 to (fi.ny-2) shr 1 do
        d[1]+=sqr(yo+i);
    d[1]*=2*fi.nx;

    { calculate d[2] and matrix for radial polynomials }
    //rpc:=(sqr(fi.nx)+sqr(fi.ny))*0.125;
    pc:=8/(sqr(fi.nx)+sqr(fi.ny));
    d[2]:=0;
    am(vy,zmatrix(1,2*n+1));
    for i:=0 to (fi.ny-2) shr 1 do
    begin
        y:=yo+i;
        for j:=0 to (fi.nx-2) shr 1 do
        begin
            x:=xo+j;
            d[2]+=sqr(y*x);
            r:=pc*(sqr(x)+sqr(y));
            s:=1;
            for k:=0 to 2*n do
            begin
                vy.pt[k]+=s;
                s*=r;
            end
        end;
    end;
    d[2]*=4;
    am(ma,zmatrix(n+1,n+1));
    for i:=0 to n do
        for j:=0 to n do
            ma.pt[i*(n+1)+j]:=4*vy.pt[i+j];
    am(vy,zmatrix(1,n+1));

    { preparations }
    tn:=cc.threads;
    k:=n+4;
    getmem2(sp, tn*fi.nc*k*sizeof(double));
    getmem2(a, fi.nc*k*sizeof(double));
    getmem2(dev, fi.nc*sizeof(double));
    for j:=0 to fi.nc-1 do
    begin
        for i:=0 to k-1 do
            a[j*k+i]:=0;
//      a[j*k+3]:=1e300;
        dev[j]:=1e300;
    end;

    { approximation iterations }
    for i:=0 to it do
    begin
        { calculate deviation }
        if i>0 then
        begin
            ya:=0;
            for j:=0 to tn-1 do
            begin
                yz:=fi.ny*(j+1) div tn;
                dothread SImg_Approx_thread3(fi, ya,yz-1, n, sp+j*fi.nc, a);
                ya:=yz;
            end;
            threadwait;
            s:=1/(fi.nx*fi.ny);
            for k:=0 to fi.nc-1 do
            begin
                r:=0;
                for j:=0 to tn-1 do
                    r+=sp[j*fi.nc+k];
                r*=s;
                dev[k]:=r;
//              write(stderr,'  ',real2str(r,8));
            end;
//          writeln(stderr);
        end;

        { calculate <p_i,f> }
        ya:=0;
        for j:=0 to tn-1 do
        begin
            yz:=fi.ny*(j+1) div tn;
            dothread SImg_Approx_thread1(fi, ya,yz-1, n, sp+j*fi.nc*(n+4), a, dev);
            ya:=yz;
        end;
        threadwait;

        { solve for factors }
        for c:=0 to fi.nc-1 do
        begin
            x:=1;
            for k:=0 to n+3 do
            begin
                y:=0;
                for j:=0 to tn-1 do
                    y+=sp[ (j*fi.nc+c)*(n+4) + k];
                case k of
                    0 : a[c*(n+4)+0]:=y*r1/d[0];
                    1 : a[c*(n+4)+1]:=y*r1/d[1];
                    2 : a[c*(n+4)+2]:=y*r1b/d[2];
                    else begin
                        vy.pt[k-3]:=y*x;
                        x*=pc;
                    end;
                end;
            end;
            am(vx,solve(ma,vy));
            if i=it then x:=rr
            else x:=1;
            for k:=0 to n do
            begin
                a[c*(n+4)+3+k]:=vx.pt[k]*x;
                x*=pc;
            end;
        end;
    end;

    if irad>0 then SImg_free(cc,fi);
    free(ma);
    free(vx);
    free(vy);

    ya:=0;
    for j:=0 to tn-1 do
    begin
        yz:=fj.ny*(j+1) div tn;
        dothread SImg_Approx_thread2(fj, ya,yz-1, n, a, dm, ss);
        ya:=yz;
    end;
    threadwait;

    freemem2(sp, tn*fi.nc*(n+4)*sizeof(double));
    freemem2(a, fi.nc*(n+4)*sizeof(double));
    freemem2(dev, fi.nc*sizeof(double));
    SImg_DecDebugLevel;
end;


{*************************************************************************}
{******* SImg_Approx2 ****************************************************}
{*************************************************************************}
const SImg_approx2_pnmax = 16;
{ calculate scalar products
    ya,yz     : region for this thread
    pn        : number of polynomials
    sp        : scalar products (results) for vector and matrix: nc*(pn + pn*(pn+1)/2)
}
threadprocedure SImg_approx2_thread1(fi:TSImg_FPImage; fw:TSImg_FPImage; ya:longint; yz:longint; pn:longint; sp:pdouble );
var i,j,k,l,c      : longint;
    x,y,r,s,rx,ry  : double;
    p              : array[0..SImg_approx2_pnmax-1] of double;
    d,e            : ^simgfp;
    f              : ^double;
begin
    rx:=1/(fi.nx-1);
    ry:=1/(fi.ny-1);
    d:=fi.pt + ya*fi.sl;
    e:=fw.pt + ya*fi.sl;
    for i:=ya to yz do
    begin
        y:=i*ry-0.5;
        for j:=0 to fi.nx-1 do
        begin
            x:=j*rx-0.5;
            p[0]:=1;
            p[1]:=x;
            p[2]:=y;
            p[3]:=x*y;
            r:=sqr(x)+sqr(y);
            s:=1;
            for k:=4 to pn-1 do
            begin
                s*=r;
                p[k]:=s;
            end;
            f:=sp;
            for c:=0 to fi.nc-1 do
            begin
                r:=d^; d+=1;
                s:=e^; e+=1;
                for k:=0 to pn-1 do
                begin
                    x:=p[k]*s;
                    f[0]+=x*r;
                    for l:=0 to k do
                        f[l+1]+=x*p[l];
                    f+=k+2;
                end;
            end;
        end;
    end;
end;

{ calculate scalar products
    ya,yz     : region for this thread
    pn        : number of polynomials
    a         : coefficients nc*pn
    dm        : difference mode
}
threadprocedure SImg_approx2_thread2(fi:TSImg_FPImage; ya:longint; yz:longint; pn:longint; a:pdouble; dm:boolean);
var i,j,k,c        : longint;
    x,y,r,s,rx,ry  : double;
    p              : array[0..SImg_approx2_pnmax-1] of double;
    d              : ^simgfp;
    f              : ^double;
begin
    rx:=1/(fi.nx-1);
    ry:=1/(fi.ny-1);
    d:=fi.pt + ya*fi.sl;
    for i:=ya to yz do
    begin
        y:=i*ry-0.5;
        for j:=0 to fi.nx-1 do
        begin
            x:=j*rx-0.5;
            p[0]:=1;
            p[1]:=x;
            p[2]:=y;
            p[3]:=x*y;
            r:=sqr(x)+sqr(y);
            s:=1;
            for k:=4 to pn-1 do
            begin
                s*=r;
                p[k]:=s;
            end;
            f:=a;
            for c:=0 to fi.nc-1 do
            begin
                r:=0;
                for k:=0 to pn-1 do
                    r+=f[k]*p[k];
                f+=pn;
                if dm then d^-=r
                else d^:=r;
                d+=1;
            end;
        end;
    end;
end;

{ approximate an image with weight by polynomials
    fi   : input image
    fw   : weight
    pn   : number of polynomials (recommended: 1,3,4,5,..)
    dm   : false: approximation replaces input
           true: difference replaces input
   polynomials:
    p_0(x,y)=1
    p_1(x,y)=x
    p_2(x,y)=y
    p_3(x,y)=x*y
    p_3+i=(x^2+y^2)^i
}
procedure SImg_Approx2(cc:CSImg_Config; fi,fw:TSImg_FPImage; pn: longint; dm:boolean);
var ma,vx,vy   : matrix;
    sp,f       : ^double;                       { buffer for <p_i,f> and approximation coefficients }
    i,j,k,tn   : longint;
    ya,yz      : longint;
threadvars;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_Approx2: fi='+SImg_InfoStr2(fi)+' fw='+SImg_InfoStr2(fi)+' pn='+int2str(pn)+' dm='+boolean2str(dm));

    if pn<1 then
    begin
        SImg_IncDebugLevel;
        cc.warning('SImg_Approx2: Number of polynomials less than 1: Nothing to do.');
        exit;
    end;
    if pn>SImg_approx2_pnmax then pn:=SImg_approx2_pnmax;

    if not SImg_Recover(cc,fi) then cc.error('SImg_Approx2: Invalid image: fi');
    if not SImg_Recover(cc,fw) then cc.error('SImg_Approx2: Invalid image: fw');
    if fw.size<>fi.size then cc.error('SImg_Approx2: Incompatible pixel weight');

    tn:=cc.threads;
    k:=pn + pn*(pn+1) div 2;
    getmem2(sp, tn*fi.nc*k*sizeof(double));
    for i:=0 to tn*fi.nc*k-1 do
        sp[i]:=0;

    { calculate scalar products }
    ya:=0;
    for j:=0 to tn-1 do
    begin
        yz:=fi.ny*(j+1) div tn;
        dothread SImg_Approx2_thread1(fi,fw, ya,yz-1, pn, sp+j*fi.nc*k);
        ya:=yz;
    end;
    threadwait;
    for i:=0 to fi.nc*k-1 do
        for j:=1 to tn-1 do
            sp[i]+=sp[i+j*fi.nc*k];

    { solve for coefficients }
    am2(ma, zmatrix(pn,pn));
    am2(vy, zmatrix(1,pn));
    vx.pt:=nil;
    f:=sp;
    for i:=0 to fi.nc-1 do
    begin
        for j:=0 to pn-1 do
        begin
            vy.pt[j]:=f[0];
            for k:=0 to j do
            begin
                ma.pt[j*pn+k]:=f[k+1];
                ma.pt[k*pn+j]:=f[k+1];
            end;
            f+=j+2;
        end;
        am(vx,solve(ma,vy));
        for j:=0 to pn-1 do
            sp[i*pn+j]:=vx.pt[j];
    end;
    free(ma);
    free(vx);
    free(vy);

    ya:=0;
    for j:=0 to tn-1 do
    begin
        yz:=fi.ny*(j+1) div tn;
        dothread SImg_Approx2_thread2(fi, ya,yz-1, pn, sp, dm);
        ya:=yz;
    end;
    threadwait;

    freemem2(sp);
    SImg_DecDebugLevel;
end;


{*******************************************************************************}
{********* SImg_BGApprox3 ******************************************************}
{*******************************************************************************}
//ifdef[approx_debug]
const SImg_BGApprox3_first : boolean = true;
//endif
{ Background subtraction using polynomial approximation
    fi: input and output
    fw: pixel weight
    bg: Approximation method:
        1: constant
        2: gradient
        3: curved
        4+: 1+ radial polynomials
}
procedure SImg_BGApprox3(cc:CSImg_Config; fi,fw:TSImg_FPImage; bg:longint);
const weight_fact = 0.8;        { weight at noise floor after down-scaling and filtering ( \approx 1/(4*filter_size) the noise floor before down-scaling) }
      filt_rad    = 2.5;        { median filter radius }
      iterations  = 10;
      ureg        = 1e-5;       { regularization for undefined pixels relative to weighted weight }
var fi2,fw2,fw3,fj              : TSImg_FPImage;
    x,y,c,xo,yo,x2,y2,i,j       : longint;
    d,e                         : ^simgfp;
    a0,a1,a2,a3                 : simgfp;
    r,s,t,r2,s2                 : double;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_BGApprox3: Called with: fi='+SImg_InfoStr2(fi)+' fw='+SImg_InfoStr2(fw)+' bg='+int2str(bg));
    if bg>1 then bg+=1; // degrees of freedom

    if not SImg_Recover(cc,fi) then cc.error('SImg_BGApprox3: Invalid image: fi');
    if not SImg_Recover(cc,fw) then cc.error('SImg_BGApprox3: Invalid image: fw');
    if (fw.nx<>fi.nx) or (fw.ny<>fi.ny) or (fw.nc<>fi.nc) then cc.error('SImg_BGApprox3: Incompatible dimensions of fw and fi');

    { weighted average used for undefined pixels, see next comment }
    s:=0;  r2:=0;  s2:=0;
    for i:=0 to fi.size-1 do
    begin
        r:=fw.pt[i];
        r2+=fi.pt[i]*r;
        s2+=sqr(r);
        s+=r;
    end;
    if s<simg_mina then cc.error('SImg_BGApprox3: Sum of fw is negative');
    s:=1/s;
    t:=s2*s;    { weight limit in order to reduce effects of incorrectly estimated noise }
    s2:=t*ureg;
    r2*=s*s2;

    { Down-scaling by factor 4 + median filter with diameter (2*filt_rad).
      The regularization is used in order to avoid artifacts at boundaries of undefined regions.
      For that reason the 30% quantile is used as weight }
    fi2.pt:=nil; fw2.pt:=nil;
    SImg_New(cc, fi2, fi.nx shr 2, fi.ny shr 2, fi.nc, false);
    SImg_New(cc, fw2, fi2.nx, fi2.ny, fi2.nc, false);
    xo:=(fi.nx - fi2.nx*4) div 2;
    yo:=(fi.ny - fi2.ny*4) div 2;
    j:=0;
    for y:=0 to fi2.ny-1 do
    begin
        y2:=(y*4+yo)*fi.sl;
        for x:=0 to fi2.nx-1 do
        begin
            x2:=y2 + (x*4+xo)*fi.nc;
            for c:=0 to fi.nc-1 do
            begin
                d:=fi.pt + x2 + c;
                e:=fw.pt + x2 + c;
                r:=0;  s:=0;
                for i:=0 to 3 do
                begin
                    a0:=e[0];           if a0>t then a0:=t;
                    a1:=e[fi.nc];       if a1>t then a1:=t;
                    a2:=e[2*fi.nc];     if a2>t then a2:=t;
                    a3:=e[3*fi.nc];     if a3>t then a3:=t;
                    e+=fw.sl;
                    r+=d[0]*a0 + d[fi.nc]*a1 + d[2*fi.nc]*a2 + d[3*fi.nc]*a3;
                    d+=fi.sl;
                    s+=a0+a1+a2+a3;
                end;
                if s<0 then
                begin
                    s:=0;
                    r:=0;
                end;
                s+=s2;
                fw2.pt[j]:=s;
                fi2.pt[j]:=(r+r2)/s;
                j+=1;
            end;
        end;
    end;
    fw3.pt:=nil;
    SImg_Duplicate(cc, fi2,fw3, true);
    SImg_Filter(cc, fi2,fi2, filt_rad, -2, 1, 0.5);
    SImg_Filter(cc, fw2,fw2, filt_rad, -2, 1, 0.3);
//ifdef[approx_debug]
    if SImg_BGApprox3_first then SImg_WriteImageFile(cc,'bgApprox3-fi2.fits',fi2);
//endif

    { estimate noise from difference of input and output of median filter }
    t:=simg_mina; s:=0;
    for i:=0 to fw3.size-1 do
    begin
        r:=fw2.pt[i];
        t+=sqr(fi2.pt[i]-fw3.pt[i])*r;
        s+=r;
    end;
//    cc.info('sig='+real2str(sqrt(t/s)));
    t:=sqr(2*filt_rad) * s/t * ln(weight_fact);

    { 1st approximation }
    fj.pt:=nil;
    SImg_Duplicate(cc, fi2,fj, true);
    SImg_Approx2(cc, fj,fw2, bg, true);

    { approximation }
    for i:=1 to iterations do
    begin
        for x:=0 to fw2.size-1 do
        begin
            r:=fj.pt[x];
            if r<=0 then fw3.pt[x]:=fw2.pt[x]*(1-sqr(r)*t)      { u<0 }
            else fw3.pt[x]:=fw2.pt[x]*exp(sqr(r)*t);
        end;
//      SImg_WriteImageFile(cc,'bgApprox3-fw3-'+int2str(i)+'.fits',fw3);
        SImg_Approx2(cc, fj,fw3, bg, true);
    end;
    SImg_Free(cc,fw2);

    { calculate bg approximation }
    for x:=0 to fi2.size-1 do
        fi2.pt[x]-=fj.pt[x];

//ifdef[approx_debug]
    if SImg_BGApprox3_first then SImg_WriteImageFile(cc,'bgApprox3-fj.fits',fj);
    if SImg_BGApprox3_first then SImg_WriteImageFile(cc,'bgApprox3-fw3.fits',fw3);
//endif
    SImg_Free(cc,fw3);
    SImg_Free(cc,fj);

    { bg subtraction }
    d:=fi.pt;
    for y:=0 to fi.ny-1 do
    begin
        if y>yo then
        begin
            y2:=(y-yo) shr 2;
            if y2>=fi2.ny then y2:=fi2.ny-1;
        end
        else y2:=0;
        for x:=0 to fi.nx-1 do
        begin
            if x>xo then
            begin
                x2:=(x-xo) shr 2;
                if x2>=fi2.nx then x2:=fi2.nx-1;
            end
            else x2:=0;
            e:=fi2.pt + y2*fi2.sl + x2*fi2.nc;
            for c:=0 to fi.nc-1 do
                d[c]-=e[c];
            d+=fi.nc;
        end;
    end;
//ifdef[approx_debug]
    if SImg_BGApprox3_first then SImg_WriteImageFile(cc,'bgApprox3-fi.fits',fi);
    SImg_BGApprox3_first:=false;
//endif
    SImg_Free(cc,fi2);

    SImg_DecDebugLevel;
end;


{*******************************************************************************}
{********* SImg_BGApprox4 ******************************************************}
{*******************************************************************************}
{ Background approximation by image and constant
    fi: input and output
    fa: approximation image
    fw: pixel weight
    result: image factor
}
function SImg_BGApprox4(cc:CSImg_Config; fi,fa,fw:TSImg_FPImage):double;
const weight_fact = 0.8;        { weight at noise floor after down-scaling and filtering ( \approx 1/(4*filter_size) the noise floor before down-scaling) }
      filt_rad    = 2.5;        { median filter radius }
      iterations = 10;
      ureg       = 1e-5;        { regularization for undefined pixels relative to weighted weight }
var fi2,fa2,fw2,fv              : TSImg_FPImage;
    x,y,c,xo,yo,x2,y2,i,j       : longint;
    d,e,f                       : ^simgfp;
    a0,a1,a2,a3,a4              : double;
    r,s,t,u,r2,s2,t2            : double;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_BGApprox4: Called with: fi='+SImg_InfoStr2(fi)+' fw='+SImg_InfoStr2(fw)+' fa='+SImg_InfoStr2(fa));

    if not SImg_Recover(cc,fi) then cc.error('SImg_BGApprox4: Invalid image: fi');
    if not SImg_Recover(cc,fa) then cc.error('SImg_BGApprox4: Invalid image: fa');
    if (fa.nx<>fi.nx) or (fa.ny<>fi.ny) or (fa.nc<>fi.nc) then cc.error('SImg_BGApprox4: Incompatible dimensions of fa and fi');
    if not SImg_Recover(cc,fw) then cc.error('SImg_BGApprox4: Invalid image: fw');
    if (fw.nx<>fi.nx) or (fw.ny<>fi.ny) or (fw.nc<>fi.nc) then cc.error('SImg_BGApprox4: Incompatible dimensions of fa and fi');

    { weighted average used for undefined pixels, see next comment }
    s:=0;  r2:=0;  s2:=0;  t2:=0;
    for i:=0 to fi.size-1 do
    begin
        r:=fw.pt[i];
        r2+=fi.pt[i]*r;
        t2+=fa.pt[i]*r;
        s2+=sqr(r);
        s+=r;
    end;
    if s<simg_mina then cc.error('SImg_BGApprox4: Sum of fw is negative');
    s:=1/s;
    u:=s2*s;    { weight limt in order to reduce effects of incorrectly estimated noise }
    s2:=u*ureg;
    r2*=s*s2;
    t2*=s*s2;
    { Down-scaling by factor 4 + median filter with diameter 5.
      The regularization is used in order to avoid artifacts at boundaries of undefined regions.
      For that reason the 30% quantile is used as weight }
    fi2.pt:=nil; fw2.pt:=nil; fa2.pt:=nil;
    SImg_New(cc, fi2, fi.nx shr 2, fi.ny shr 2, fi.nc, false);
    SImg_New(cc, fa2, fi2.nx, fi2.ny, fi2.nc, false);
    SImg_New(cc, fw2, fi2.nx, fi2.ny, fi2.nc, false);
    xo:=(fi.nx - fi2.nx*4) div 2;
    yo:=(fi.ny - fi2.ny*4) div 2;
    j:=0;
    for y:=0 to fi2.ny-1 do
    begin
        y2:=(y*4+yo)*fi.sl;
        for x:=0 to fi2.nx-1 do
        begin
            x2:=y2 + (x*4+xo)*fi.nc;
            for c:=0 to fi.nc-1 do
            begin
                d:=fi.pt + x2 + c;
                e:=fw.pt + x2 + c;
                f:=fa.pt + x2 + c;
                r:=0; s:=0; t:=0;
                for i:=0 to 3 do
                begin
                    a0:=e[0];           if a0>u then a0:=u;
                    a1:=e[fi.nc];       if a1>u then a1:=u;
                    a2:=e[2*fi.nc];     if a2>u then a2:=u;
                    a3:=e[3*fi.nc];     if a3>u then a3:=u;
                    e+=fw.sl;
                    r+=d[0]*a0 + d[fi.nc]*a1 + d[2*fi.nc]*a2 + d[3*fi.nc]*a3;
                    d+=fi.sl;
                    t+=f[0]*a0 + f[fi.nc]*a1 + f[2*fi.nc]*a2 + f[3*fi.nc]*a3;
                    f+=fa.sl;
                    s+=a0+a1+a2+a3;
                end;
                if s<0 then
                begin
                    s:=0;
                    r:=0;
                    t:=0;
                end;
                s+=s2;
                fw2.pt[j]:=s;
                s:=1/s;
                fi2.pt[j]:=(r+r2)*s;
                fa2.pt[j]:=(t+t2)*s;
                j+=1;
            end;
        end;
    end;
    fv.pt:=nil;
    SImg_Duplicate(cc, fi2,fv, true);
    SImg_Filter(cc, fi2,fi2, filt_rad, -2, 1, 0.5);
    SImg_Filter(cc, fa2,fa2, filt_rad, -2, 1, 0.5);
    SImg_Filter(cc, fw2,fw2, filt_rad, -2, 1, 0.3);
//ifdef[approx_debug]
    SImg_WriteImageFile(cc,'bgApprox4-fi2.fits',fi2);
    SImg_WriteImageFile(cc,'bgApprox4-fa2.fits',fa2);
//endif

    { estimate noise from difference of input and output of median filter, initial weight }
    u:=simg_mina; s:=0;
    for i:=0 to fv.size-1 do
    begin
        r:=fw2.pt[i];
        u+=sqr(fi2.pt[i]-fv.pt[i])*r;
        s+=r;
        fv.pt[i]:=r;
    end;
//    cc.info('sig='+real2str(sqrt(u/s)));
    u:=sqr(2*filt_rad) * s/u * ln(weight_fact);

    { approximation iterations }
    for i:=0 to iterations do
    begin
        { approximation }
        a0:=0; a1:=0; a2:=0; a3:=0; a4:=0;
        for x:=0 to fi2.size-1 do
        begin
            r:=fi2.pt[x];
            s:=fa2.pt[x];
            t:=fv.pt[x];
            a0+=sqr(s)*t;
            a1+=s*t;
            a2+=t;
            a3+=r*s*t;
            a4+=r*t;
        end;
        r:=a0*a2-sqr(a1);
        if abs(r)<1e-100 then
        begin
            s:=0;
            t:=0;
            break;
        end;
        r:=1/r;
        s:=(a2*a3-a1*a4)*r;     { image factor }
        t:=(a0*a4-a1*a3)*r;     { constant }

        { new weight }
        if i<iterations then
        begin
            for x:=0 to fw2.size-1 do
            begin
                r:=fi2.pt[x]-s*fa2.pt[x]-t;
                if r<=0 then fv.pt[x]:=fw2.pt[x]*(1-sqr(r)*u)   { u<0 }
                else fv.pt[x]:=fw2.pt[x]*exp(sqr(r)*u);
            end;
        end;
    end;

//ifdef[approx_debug]
    SImg_WriteImageFile(cc,'bgApprox4-fv.fits',fv);
//endif
    SImg_Free(cc,fi2);
    SImg_Free(cc,fa2);
    SImg_Free(cc,fw2);
    SImg_Free(cc,fv);
    result:=s;

    SImg_DecDebugLevel;
end;


{*******************************************************************************}
{********* SImg_Imgapprox ******************************************************}
{*******************************************************************************}
//ifdef[approx_debug]
const SImg_Imgapprox_first : boolean = true;
//endif
{ Approximate fi by fa and polynomials and return the difference.
    fi:  Input image and background corrected output (1 color)
    fa:  Approximation image (1 colors)
    fw:  Weight for approximation image
    fo1: Residual output
    fo2: Background corrected output
    bg:  polynomial approximation method:
         0: none
         1: constant
         2: gradient
         3: curved
         4+: 1+ radial polynomials
    mosaic: boundary overweight factor, >=0
    bc:  Brightness calibration, calibration factor is returned. Not recommended because sensitive to noise and errors.
    result: calibration factor (1 if bc=false)
}
function SImg_ImgApprox(cc:CSImg_Config; fi,fa,fw:TSImg_FPImage; var fo1,fo2:TSImg_FPImage; bg:longint; mosaic:double; bc:boolean):double;
const f_ss      = 4;            // down-sampling. Filter diameter is 2*f_ss+1
      q_um_low  = 0.5e-3;       // lower unmask quantile for gradient calculation
      q_um_high = 1-2e-3;       // higher unmask quantile for gradient calculation

var fi2,fa2,fw2,fi1,fa1         : TSImg_FPImage;
    x,y,i,xo,yo,x2,y2           : longint;
    a,b,r,s,t                   : double;
    d,e                         : ^simgfp;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_Imgapprox: Called with: fi='+SImg_InfoStr2(fi)+' fa='+SImg_InfoStr2(fa)+' fw='+SImg_InfoStr2(fw)+' fo1='+SImg_InfoStr2(fo1)+' fo2='+SImg_InfoStr2(fo2)+' bg='+int2str(bg)+' mosaic='+real2str(mosaic)+' bc='+boolean2str(bc));
    if bg>1 then bg+=1; // degrees of freedom
    if mosaic<0 then mosaic:=0;

    if not SImg_Recover(cc,fi) then cc.error('SImg_Imgapprox: Invalid image: fi');
    if not SImg_Recover(cc,fa) then cc.error('SImg_Imgapprox: Invalid image: fa');
    if not SImg_Recover(cc,fw) then cc.error('SImg_Imgapprox: Invalid image: fw');
    if (fi.nc<>1) then cc.error('SImg_Imgapprox: Invalid dimensions fi');
    if (fa.nx<>fi.nx) or (fa.ny<>fi.ny) or (fa.nc<>1) then cc.error('SImg_Imgapprox: Invalid dimensions of fa');
    if (fw.nx<>fi.nx) or (fw.ny<>fi.ny) or (fw.nc<>1) then cc.error('SImg_Imgapprox: Invalid dimensions of fw');
    if not SImg_OverWriteAlert(fo1,fi) and not SImg_OverWriteAlert(fo1,fa) and not SImg_OverWriteAlert(fo1,fw) then
    begin
        SImg_Free(cc,fo1);
        fo1.pt:=nil;
    end;
    if not SImg_OverWriteAlert(fo2,fi) and not SImg_OverWriteAlert(fo2,fa) and not SImg_OverWriteAlert(fo2,fw) then
    begin
        SImg_Free(cc,fo2);
        fo2.pt:=nil;
    end;

    { linear down-scaling + filtering
      These operations must be linear and un-weighted in order to avoid effects caused by different PSF width's }
    fa2.pt:=nil; SImg_Filter(cc, fa ,fa2,  f_ss+1, -3, f_ss);
    fi2.pt:=nil; SImg_Filter(cc, fi ,fi2,  f_ss+1, -3, f_ss);
    fw2.pt:=nil; SImg_Filter(cc, fw ,fw2,  f_ss+1, -5, f_ss);

//ifdef[approx_debug]
    if SImg_Imgapprox_first then
    begin
        SImg_WriteImageFile(cc,'Imgapprox-fi2.fits',fi2);
        SImg_WriteImageFile(cc,'Imgapprox-fw2.fits',fw2);
        SImg_WriteImageFile(cc,'Imgapprox-fa2.fits',fa2);
    end;
//endif

    { polynomial approximation }
    fi1.pt:=nil; fa1.pt:=nil;
    SImg_Duplicate(cc, fi2,fi1,  true);
    SImg_Duplicate(cc, fa2,fa1, true);
    if bg>0 then
    begin
        SImg_Approx2(cc, fi2,fw2, bg, true);
        SImg_Approx2(cc, fa2,fw2, bg, true);
    end;

    { image approximation }
    a:=simg_mina; b:=simg_mina;
    for x:=0 to fi2.size-1 do
    begin
        t:=fw2.pt[x];
        if t<0 then t:=0;
        r:=fa2.pt[x];
        s:=fi2.pt[x];
        a+=sqr(r)*t;
        b+=r*s*t;
    end;
    b/=a;
    cc.info(cc.name+': Gradient and background subtraction. Image approximation factor: '+real2str(b,-3));
    result:=1;
    if bc and (b>simg_mina) then result:=1/b;

    { down-scaled residual }
    a:=1/result;
    for x:=0 to fi2.size-1 do
    begin
        r:=fi1.pt[x];
        s:=fa1.pt[x];
        fi2.pt[x]-=b*fa2.pt[x];         { residual }
        fa1.pt[x]:=r - b*s;             { down-scaled fo1 plus gradient }
        fa2.pt[x]:=r - a*s;             { down-scaled fo2 plus gradient minus fa }
    end;
//ifdef[approx_debug]
    if SImg_Imgapprox_first then SImg_WriteImageFile(cc,'Imgapprox-fr2.fits',fi2);
//endif

    { gradient for fo1 }
    s:=SImg_GetQuantile(cc, fi2, q_um_low);
    t:=SImg_GetQuantile(cc, fi2, q_um_high);
    for x:=0 to fw2.size-1 do
    begin
        r:=fi2.pt[x];
        if (r<s) or (r>t) then fw2.pt[x]:=0;
    end;
//ifdef[approx_debug]
    if SImg_Imgapprox_first then SImg_WriteImageFile(cc,'Imgapprox-fw3.fits',fw2);
//endif
    SImg_Approx2(cc, fa1,fw2, bg, false);

    { gradient for fo2 }
    if mosaic>1e-3 then { mosaic gradient }
    begin
        mosaic/=1+mosaic;
        s:=2/(fw2.nx-1);
        t:=2/(fw2.ny-1);
        i:=0;
        for y:=0 to fw2.ny-1 do
        begin
            r:=y*t;
            r*=(2-r)*mosaic;
            for x:=0 to fw2.nx-1 do
            begin
                a:=x*s;
                fw2.pt[i]*=1-a*(2-a)*r;
                i+=1;
            end;
        end;
//ifdef[approx_debug]
        if SImg_Imgapprox_first then SImg_WriteImageFile(cc,'Imgapprox-fw4.fits',fw2);
//endif
    end;
    SImg_Approx2(cc, fa2,fw2, bg, false);

    { output }
    SImg_Free(cc, fi1);
    SImg_Free(cc, fi2);
    SImg_Free(cc, fw2);

    if fo1.pt=nil then SImg_New(cc, fo1, fi.nx,fi.ny,1, false);
    if fo2.pt=nil then SImg_New(cc, fo2, fi.nx,fi.ny,1, false);
    d:=fo1.pt;
    e:=fo2.pt;
    i:=0;
    xo:=(fi.nx-fi2.nx*f_ss) div 2;
    yo:=(fi.ny-fi2.ny*f_ss) div 2;
    for y:=0 to fi.ny-1 do
    begin
        y2:=(y+yo) div f_ss;
        for x:=0 to fi.nx-1 do
        begin
            x2:=(x+xo) div f_ss + fi2.sl*y2;
            r:=fi.pt[i];
            s:=fa.pt[i];
            d[i]:=r - b*s - fa1.pt[x2];
            e[i]:=r       - fa2.pt[x2];
            i+=1;
        end;
    end;

//ifdef[approx_debug]
    if SImg_Imgapprox_first then SImg_WriteImageFile(cc,'Imgapprox-fg1.fits',fa1);
    if SImg_Imgapprox_first then SImg_WriteImageFile(cc,'Imgapprox-fg2.fits',fa2);
    if SImg_Imgapprox_first then SImg_WriteImageFile(cc,'Imgapprox-fo.fits',fo2);
    if SImg_Imgapprox_first then SImg_WriteImageFile(cc,'Imgapprox-fres.fits',fo1);
    SImg_Imgapprox_first:=false;
//endif
    SImg_Free(cc, fa1);
    SImg_Free(cc, fa2);

    SImg_DecDebugLevel;
end;


{*******************************************************************************}
{********* SImg_ApproxMosaic ***************************************************}
{*******************************************************************************}
{ Approximate fi by polynomials. Border will be overweighted (mosaic mode)
    fi: Input and output
    fw: Weight
    bg: polynomial approximation method:
         0: none
         1: constant
         2: gradient
         3: curved
         4+: 1+ radial polynomials
    dm: false: output approximation
        true: output difference
    pc: true if input is already pre-corrected
    mosaic: boundary overweight factor, >0
}
procedure SImg_ApproxMosaic(cc:CSImg_Config; fi,fw:TSImg_FPImage; bg:longint; mosaic:double; dm,pc:boolean);
const q_low     = 0.5e-3;               // lower unmask quantile for gradient calculation (normal mode)
      q_high    = 1-3e-3;       // higher unmask quantile for gradient calculation (normal mode)
      q_low_pc  = 1e-3;         // lower unmask quantile for gradient calculation (pre-corrected mode)
      q_high_pc = 1-5e-3;       // higher unmask quantile for gradient calculation (pre-corrected mode)
      w_fact    = 1.0;          // weight is limited to weighted weight multiplies by this parameter
var x,y,i       : longint;
    r,s,t,u     : double;
    fw2         : TSImg_FPImage;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_ApproxMosaic: Called with: fi='+SImg_InfoStr2(fi)+' fw='+SImg_InfoStr2(fw)+' bg='+int2str(bg)+' mosaic='+real2str(mosaic)+' dm='+boolean2str(dm)+' pc='+boolean2str(dm));
    if not SImg_Recover(cc,fi) then cc.error('SImg_ApproxMosaic: Invalid image: fi');
    if not SImg_Recover(cc,fw) then cc.error('SImg_ApproxMosaic: Invalid image: fw');
    if (fi.nc<>1) then cc.error('SImg_ApproxMosaic: Invalid dimensions fi');
    if (fw.nx<>fi.nx) or (fw.ny<>fi.ny) or (fw.nc<>1) then cc.error('SImg_ApproxMosaic: Invalid dimensions of fw');
    if bg>1 then bg+=1; // degrees of freedom
    if mosaic<0 then mosaic:=0;

    { unmask extreme values }
    s:=SImg_GetQuantile(cc, fi, select(pc, q_low_pc, q_low));
    t:=SImg_GetQuantile(cc, fi, select(pc, q_high_pc, q_high));
    fw2.pt:=nil;
    SImg_New(cc, fw2, fw.nx,fw.ny,1, false);
    for x:=0 to fw.size-1 do
    begin
        r:=fi.pt[x];
        if (r<s) or (r>t) then fw2.pt[x]:=0
        else fw2.pt[x]:=fw.pt[x];
    end;

    { limit weight in order to reduce effects of incorrectly estimated noise }
    r:=0; s:=0;
    for i:=0 to fw2.size-1 do
    begin
        t:=fw2.pt[i];
        r+=sqr(t);
        s+=t;
    end;
    if s<simg_mina then cc.error('SImg_ApproxMosaic: Non-positive weight image');
    u:=r/s;
    r:=0; s:=0;
    for i:=0 to fw2.size-1 do
    begin
        t:=fw2.pt[i];
        if t>u then
        begin
            t:=u;
            fw2.pt[i]:=r;
        end;
        r+=sqr(t);
        s+=t;
    end;
    if s<simg_mina then cc.error('SImg_ApproxMosaic: Non-positive weight image');
    r/=s;
    for i:=0 to fw2.size-1 do
        if fw2.pt[i]>r then fw2.pt[i]:=r;


    { boundary overweight }
    if mosaic>1e-3 then
    begin
        mosaic/=1+mosaic;
        s:=2/(fw2.nx-1);
        t:=2/(fw2.ny-1);
        i:=0;
        for y:=0 to fw2.ny-1 do
        begin
            r:=y*t;
            r*=(2-r)*mosaic;
            for x:=0 to fw2.nx-1 do
            begin
                u:=x*s;
                fw2.pt[i]*=1-u*(2-u)*r;
                i+=1;
            end;
        end;
    end;

    { approximation }
//    SImg_WriteImageFile(cc,'approxMosaic-fw2.fits',fw2);
//    SImg_WriteImageFile(cc,'approxMosaic-fi.fits',fi);
    SImg_Approx2(cc, fi,fw2, bg, dm);
//    SImg_WriteImageFile(cc,'approxMosaic-fj.fits',fi);
//    cc.error('no error');

    SImg_Free(cc,fw2);
    SImg_DecDebugLevel;
end;


{*******************************************************************************}
{********* SImg_Imgapprox2 *****************************************************}
{*******************************************************************************}
{ Approximate fi by fa and a constant. Return the factor for fa.
    fi:  Input image and background corrected output (1 color)
    fa:  Approximation image (1 colors)
    fw:  Weight for approximation image
    result: factor for fa
}
function SImg_ImgApprox2(cc:CSImg_Config; fi,fa,fw:TSImg_FPImage):double;
var r,s,t,b0,b1 : double;
    a00,a01,a11 : double;
    i           : longint;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_Imgapprox2: Called with: fi='+SImg_InfoStr2(fi)+' fa='+SImg_InfoStr2(fa)+' fw='+SImg_InfoStr2(fw));

    if not SImg_Recover(cc,fi) then cc.error('SImg_Imgapprox: Invalid image: fi');
    if not SImg_Recover(cc,fa) then cc.error('SImg_Imgapprox: Invalid image: fa');
    if not SImg_Recover(cc,fw) then cc.error('SImg_Imgapprox: Invalid image: fw');
    if fa.size<>fi.size then cc.error('SImg_Imgapprox: Invalid dimensions of fa');
    if fw.size<>fi.size then cc.error('SImg_Imgapprox: Invalid dimensions of fw');

    a00:=0; a01:=0; a11:=0;
    b0:=0; b1:=0;
    for i:=0 to fi.size-1 do
    begin
        r:=fi.pt[i];
        s:=fw.pt[i];
        t:=fa.pt[i];
        b0+=t*r*s;
        b1+=r*s;
        a00+=sqr(t)*s;
        a01+=t*s;
        a11+=s;
    end;
    r:=a00*a11-sqr(a01);
    if abs(r)<simg_mina then result:=0
    else result:=(a11*b0-a01*b1)/r;

    SImg_DecDebugLevel;
end;


{*******************************************************************************}
{********* SImg_ImgFactor ******************************************************}
{*******************************************************************************}
//ifdef[approx_debug]
const SImg_ImgFactor_first : boolean = true;
//endif
{ Calculate image factor
    fi: input
    fr: reference image
    fw: pixel weight
    bg: Approximation method:
        1: constant
        2: gradient
        3: curved
        4+: 1+ radial polynomials
}
function SImg_ImgFactor(cc:CSImg_Config; fi,fr,fw:TSImg_FPImage; bg:longint):double;
const bgit    = 2;      // background filter iterations
      oit     = 1;
      penalty = 1e3;    // background approximation penalty for negative values
      mask_fr = 1.5;    // filter radius for background and object mask
var fi2,fw2,fr2,fw3,fj  : TSImg_FPImage;
    x,y,c,xo,yo,y2,i,j  : longint;
    i0,i1,i2,i3         : longint;
    d,e,f               : ^simgfp;
    a0,a1,a2,a3,r,s     : double;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_ImgFactor: Called with: fi='+SImg_InfoStr2(fi)+' fr='+SImg_InfoStr2(fr)+' fw='+SImg_InfoStr2(fw)+' bg='+int2str(bg)+' oit='+int2str(oit));
    if bg>1 then bg+=1; // degrees of freedom

    if not SImg_Recover(cc,fi) then cc.error('SImg_ImgFactor: Invalid image: fi');
    if not SImg_Recover(cc,fr) then cc.error('SImg_ImgFactor: Invalid image: fi');
    if not SImg_Recover(cc,fw) then cc.error('SImg_ImgFactor: Invalid image: fw');
    if (fw.nx<>fi.nx) or (fw.ny<>fi.ny) or (fw.nc<>1) then cc.error('SImg_ImgFactor: Incompatible dimensions of fw and fi');
    if (fr.nx<>fi.nx) or (fr.ny<>fi.ny) or (fr.nc<>1) then cc.error('SImg_ImgFactor: Incompatible dimensions of fr and fi');

    { weight limit }
    a0:=0; a1:=0;
    for i:=0 to fw.size-1 do
    begin
        r:=fw.pt[i];
        a0+=r;
        a1+=sqr(r);
    end;
    a2:=1e-6*a1/a0;     { weight limit }
    if a2<simg_mina then a2:=simg_mina;

    { down-scaling by factor 4 + median filter with diameter 5 }
    fi2.pt:=nil; fr2.pt:=nil; fw2.pt:=nil;
    SImg_New(cc, fi2, fi.nx shr 2, fi.ny shr 2, 1, false);
    SImg_New(cc, fr2, fi2.nx, fi2.ny, 1, false);
    SImg_New(cc, fw2, fi2.nx, fi2.ny, 1, false);
    xo:=(fi.nx - fi2.nx*4) div 2;
    yo:=(fi.ny - fi2.ny*4) div 2;
    j:=0;
    for y:=0 to fi2.ny-1 do
    begin
        y2:=(y*4+yo)*fi.sl;
        for x:=0 to fi2.nx-1 do
        begin
            i:=y2 + (x*4+xo)*fi.nc;
            d:=fi.pt + i;
            e:=fr.pt + i;
            f:=fw.pt + i;
            a0:=0;  a1:=0;  s:=1e-6;
            for i:=0 to 3 do
            begin
                for c:=0 to fi.nc-1 do
                begin
                    if f[0      ]>=a2 then i0:=1 else i0:=0;
                    if f[fi.nc  ]>=a2 then i1:=1 else i1:=0;
                    if f[2*fi.nc]>=a2 then i2:=1 else i2:=0;
                    if f[3*fi.nc]>=a2 then i3:=1 else i3:=0;
                    f+=1;
                    a0+=d[0]*i0 + d[fi.nc]*i1 + d[2*fi.nc]*i2 + d[3*fi.nc]*i3;
                    d+=1;
                    a1+=e[0]*i0 + e[fi.nc]*i1 + e[2*fi.nc]*i2 + e[3*fi.nc]*i3;
                    e+=1;
                    s+=i0+i1+i2+i3;
                end;
                c:=fi.sl-fi.nc;
                d+=c;  e+=c;  f+=c;
            end;
            r:=1/s;
            fi2.pt[j]:=a0*r;
            fr2.pt[j]:=a1*r;
            fw2.pt[j]:=s;
            j+=1;
        end;
    end;
//ifdef[approx_debug]
    if SImg_ImgFactor_first then
    begin
        SImg_WriteImageFile(cc,'imgFactor-fi2-0.fits',fi2);
        SImg_WriteImageFile(cc,'imgFactor-fr2-0.fits',fr2);
        SImg_WriteImageFile(cc,'imgFactor-fw2.fits',fw2);
    end;
//endif

    { background mask }
    fj.pt:=nil;  fw3.pt:=nil;
    SImg_Duplicate(cc, fr2, fj, true);
    SImg_Duplicate(cc, fw2, fw3, true);
    for i:=1 to bgit do
    begin
        SImg_Approx2(cc, fj,fw3, bg, true);
        for j:=0 to fj.size-1 do
            if fj.pt[j]<0 then fw3.pt[j]:=fw2.pt[j]*penalty
            else fw3.pt[j]:=fw2.pt[j];
    end;
    SImg_Approx2(cc, fj,fw3, bg, true);
    a0:=0;  a1:=simg_mina;
    for j:=0 to fj.size-1 do
    begin
        r:=fw2.pt[j];
        a0+=fj.pt[j]*r;
        a1+=r;
    end;
    a0/=a1;
    if a0<simg_mina then a0:=simg_mina;
    a0:=2/a0;
    for j:=0 to fj.size-1 do
    begin
        r:=fj.pt[j]*a0;
        if r<0 then r:=1
        else if r>1 then r:=0
        else r:=1-r;
        fw3.pt[j]:=fw2.pt[j]*r;         { background mask }
    end;
    SImg_Free(cc, fj);
    SImg_Filter(cc, fw3,fw3, mask_fr);

    { gradient removal }
    SImg_Approx2(cc, fi2, fw3, bg, true);
    SImg_Approx2(cc, fr2, fw3, bg, true);
//ifdef[approx_debug]
    if SImg_ImgFactor_first then
    begin
        SImg_WriteImageFile(cc,'imgFactor-fi2-1.fits',fi2);
        SImg_WriteImageFile(cc,'imgFactor-fr2-1.fits',fr2);
        SImg_WriteImageFile(cc,'imgFactor-fwb.fits',fw3);
    end;
{    a0:=0;  a1:=0;  a2:=0;
    for i:=0 to fw3.size-1 do
    begin
        r:=fw3.pt[i];
        a0+=fi2.pt[i]*r;
        a1+=fr2.pt[i]*r;
        a2+=r;
    end;
    cc.info('fi_bg='+real2str(a0/a2,5)+'  fr_bg='+real2str(a1/a2,5));}
//endif

    { object mask }
    a3:=0; a0:=simg_mina;
    for j:=0 to fr2.size-1 do
    begin
        r:=fw2.pt[j];
        a3+=fr2.pt[j]*r;
        a0+=r;
    end;
    a3/=a0;

    a1:=a3; a2:=0;
    for i:=0 to oit do
    begin
        a2:=a1;
        a0:=0;  a1:=simg_mina;
        for j:=0 to fr2.size-1 do
        begin
            r:=fr2.pt[j];
            if r>a2 then
            begin
                s:=fw2.pt[j];
                a0+=r*s;
                a1+=s;
            end;
        end;
        a1:=a0/a1;
//ifdef[approx_debug]
//      cc.info('a1_'+int2str(i)+'='+real2str(a1));
//endif
    end;

    { object brightness }
    s:=a1-a3;
    if s<simg_mina then s:=simg_mina;
    s:=1/s;
    for j:=0 to fr2.size-1 do
    begin
        r:=(fr2.pt[j]-a3)*s;
        if r<=0 then r:=0
        else if r>1 then r:=1;
        fw3.pt[j]:=r;
    end;
    SImg_Filter(cc, fw3,fw3, mask_fr);

    a0:=simg_mina; a1:=simg_mina;
    for j:=0 to fr2.size-1 do
    begin
        r:=fw3.pt[j]*(2*mask_fr);
        if r>1 then r:=1;
        r*=fw2.pt[j];
        a0+=fr2.pt[j]*r;
        a1+=fi2.pt[j]*r;
//ifdef[approx_debug]
        fw2.pt[j]:=r;
//endif
    end;
    result:=a1/a0;

//ifdef[approx_debug]
//  cc.info('fi/fr = '+real2str(result,5));
    if SImg_ImgFactor_first then SImg_WriteImageFile(cc,'imgFactor-fwo.fits',fw2);
    SImg_ImgFactor_first:=false;
//endif
    SImg_Free(cc, fw3);
    SImg_Free(cc, fi2);
    SImg_Free(cc, fr2);
    SImg_Free(cc, fw2);

    SImg_DecDebugLevel;
end;


{*************************************************************************}
{******* SImg_BgApprox5 **************************************************}
{*************************************************************************}
{ constant background approximation. Pixels with a weight less than thr*<weighted average weight> are set to weighted average.
    c      : color
    rad    : median filter radius
    result : result buffer for at least fi.nc values
}
function SImg_BgApprox5(cc:CSImg_Config; fi,wi:TSImg_FPImage; c:longint; rad:single):double;
const thr = 0.25;       { ignore threshold, see description above }
var  fj                 : TSImg_FPImage;
     d,e                : ^simgfp;
     i,wc               : longint;
     r,s,sw,sfw,sww     : double;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_BgApprox5; fi='+SImg_InfoStr2(fi)+' wi='+SImg_InfoStr2(fi)+' c='+int2str(c)+' rad='+real2str(rad,-1));
    if not SImg_Recover(cc,fi) then cc.error('SImg_BgApprox5: Invalid image: fi');
    if not SImg_Recover(cc,wi) then cc.error('SImg_BgApprox5: Invalid image: wi');
    if (wi.nx<>fi.nx) or (wi.ny<>fi.ny) or ((wi.nc<>1) and (wi.nc<>fi.nc)) then cc.error('SImg_BGApprox5: Incompatible image dimensions');
    if (c<0) or (c>=fi.nc) then cc.error('SImg_BGApprox5: Invalid color: '+int2str(c));

    if wi.nc=1 then wc:=0
    else wc:=c;
    if rad<1.5 then rad:=1.5;

    { averages, copy }
    fj.pt:=nil;
    SImg_New(cc, fj, fi.nx,fi.ny,1, false);
    sww:=0; sw:=0; sfw:=0;
    d:=fi.pt + c;
    e:=wi.pt + wc;
    for i:=0 to fi.size-1 do
    begin
        r:=e^;  e+=wi.nc;
        s:=d^;  d+=fi.nc;
        sw+=r;
        sww+=sqr(r);
        sfw+=s*r;
        fj.pt[i]:=s;
    end;
    if sw<simg_mina then cc.error('SImg_BGApprox5: Non-positive weight image');
    sw:=1/sw;
    sfw*=sw;
    sww*=sw*thr;

    { mask out pixels }
    e:=wi.pt + wc;
    for i:=0 to fi.size-1 do
    begin
        if e^<sww then fj.pt[i]:=sfw;
        e+=wi.nc;
    end;

    { median filter }
    i:=round(rad*2) div 3;
    SImg_Filter(cc, fj,fj, rad,-2,i, 0.5);

    r:=fj.pt[0];
    for i:=1 to fj.size-1 do
        if fj.pt[i]<r then r:=fj.pt[i];
    SImg_Free(cc, fj);

    result:=r;
    SImg_DecDebugLevel;
end;
