{  simg -- image processing library
   Version 20210727
   Copyright (C) 1999-2021 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

// define[debug_findStars]
// define[debug_starSubtract]

{*************************************************************************}
{******* SImg_Subtract_NoiseFloor ****************************************}
{*************************************************************************}
{ Calculate the noise floor from a background corrected image, also see SImg_NoiseFloor_int.
  Due to the larger background filter image contains more object information. Therefore quantiles are shifted left.
  fj : bg corrected: minimum filter with radius > 2*<superresolution factor> should be subtracted
  sig,offs : noise floor and offset
}
procedure SImg_StarSubtract_NoiseFloor(cc:CSImg_Config; fj: TSImg_FPImage; var sig,offs:double);
var qs          : array[0..1] of double;
    ps          : array[0..1] of double = (0.25, 0.57);         { assuming with stars }
begin
    if not SImg_Recover(cc,fj) then cc.error('SImg_StarSubtract_NoiseFloor: Invalid image: fj');
    SImg_GetQuantile(cc, fj, cc.threads,-1,ps,qs);
    sig:=qs[1]-qs[0];
    offs:=0.5*(qs[1]+qs[0]);
end;


{*************************************************************************}
{******* SImg_SubtractBackground *****************************************}
{*************************************************************************}
{ Subtract background from image, process saturated pixels and calculate noise floor and  offset
    fi       : input
    fi2      : background subtracted output, brightness for saturated pixels is result*(1+0.1*d) where d is the smallest distance to a non-saturated pixel
    brad     : Background interpolation radius
    bradn    : Background interpolation radius for noise estimation (use brad if bradn<0 of bradn>brad)
    nrad     : Noise filtering radius. 0:disable, <0 automatic quantile noise filter depending on brad
    sat      : brightest non-saturated pixel
    sig,offs : noise floor and offset
}
procedure SImg_SubtractBackground(cc:CSImg_Config; fi:TSImg_FPImage; var fi2:TSImg_FPImage; brad,bradn,nrad:double; var sat,sig,offs:double);
const sat_thr = 0.5;    { saturation threshold relative to brightest pixel }
      lq      = 0.001;  { lower quantile  }
var fj,fj2      : TSImg_FPImage;
    i,x,y       : longint;
    xa,xz,ya,yz : longint;
    r,s,t,u,v   : simgfp;
    st          : shortstring;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_SubtractBackground: fi='+SImg_InfoStr2(fi)+' fi2='+SImg_InfoStr2(fi2)+' brad='+real2str(brad)+' bradn='+real2str(bradn)+' nrad='+real2str(nrad));
    if not SImg_Recover(cc,fi) then cc.error('SImg_GetPSF: Invalid image: fi');
    if not Simg_OverWriteAlert(fi,fi2) then SImg_Free(cc,fi2);

    if (bradn<0.499) or (bradn>brad-1e-2) then bradn:=-1;

    st:='';
    if bradn>0 then st:=', noise estimation with interpolation radius '+real2str(bradn,-1);
    cc.info(cc.name+': Background subtraction with interpolation radius '+real2str(brad,-1)+st);

    { minimum / maximum, lower quantile }
    s:=SImg_GetQuantile(cc,fi,lq);
    fj.pt:=nil;
    SImg_New(cc, fj, fi.nx,fi.ny,fi.nc, false);
    t:=s;
    for i:=0 to fi.size-1 do
    begin
        r:=fi.pt[i];
        if r>t then t:=r;
        if r<s then r:=s;
        fj.pt[i]:=r;
    end;
    t:=s+sat_thr*(t-s);

    { bg filter }
    if nrad>0.2 then SImg_Filter(cc, fj,fj, nrad)
    else if nrad<=-simg_mina then
    begin
        if brad>=50 then SImg_Filter(cc, fj,fj, 2.5, -2,1,0.5)          { drop 12/25 }
        else if brad>=25 then SImg_Filter(cc, fj,fj, 1.5, -2,1,0.5)     { drop 4/9 }
        else if brad>=20 then SImg_Filter(cc, fj,fj, 1.5, -2,1,0.375)   { drop 3/9 }
        else if brad>=15 then SImg_Filter(cc, fj,fj, 1.5, -2,1,0.25)    { drop 2/9 }
        else if brad>=10 then SImg_Filter(cc, fj,fj, 1.5, 1,1);         { drop 1/9 }
    end;
//  SImg_WriteImageFile(cc,'subtractBackground-fj1.fits',fj);
    fj2.pt:=nil;
    if bradn>0 then
    begin
        SImg_FBGFilter(cc, fj,fj2, bradn, true, false);
        SImg_FBGFilter(cc, fj2,fj, sqrt(sqr(brad)-sqr(bradn)), true, true)
    end
//    SImg_WriteImageFile(cc,'subtractBackground-fj2.fits',fj);
    else SImg_FBGFilter(cc, fj,fj, brad, true, false);

    { subtract BG and handle saturated pixels }
    u:=1/t;
    s:=t-s;
    for i:=0 to fi.size-1 do
    begin
        r:=fi.pt[i];
        if r<t then
        begin
            r-=fj.pt[i];
            fj.pt[i]:=r;
        end
        else begin
            y:=i div fj.sl;
            x:=(i-y*fj.sl) div fj.nc;
            xa:=0;
            while (x+xa>0) and (fi.pt[i+xa*fi.nc]>t) do
                xa-=1;
            xz:=0;
            while (x+xz<fi.nx-1) and (fi.pt[i+xz*fi.nc]>t) do
                xz+=1;
            ya:=0;
            while (y+ya>0) and (fi.pt[i+ya*fi.sl]>t) do
                ya-=1;
            yz:=0;
            while (y+yz<fi.ny-1) and (fi.pt[i+yz*fi.sl]>t) do
                yz+=1;
            v:=fi.nx+fi.ny;
            for y:=ya to yz do
                for x:=xa to xz do
                begin
                    r:=fi.pt[i+y*fi.sl+x*fi.nc];
                    if r<t then
                    begin
                        r:=sqrt(sqr(y)+sqr(x))-1+r*u; { interpolated distance }
                        if r<v then v:=r;
                    end;
                end;
            fj.pt[i]:=s*(1 + v*0.1);
{           y:=i div fj.sl;
            x:=(i-y*fj.sl) div fj.nc;
            x-=2270;
            y-=1908;
            if sqr(x)+sqr(y)<10 then writeln(stderr,x,', ',y,'   ',v);}
        end;
    end;
    sat:=s;

    { subtract BG and handle saturated pixels }
    if fj2.pt=nil then SImg_StarSubtract_NoiseFloor(cc, fj, sig,offs)
    else begin
        for i:=0 to fi.size-1 do
            fj2.pt[i]:=fi.pt[i]-fj2.pt[i];
        SImg_StarSubtract_NoiseFloor(cc, fj2, sig,offs);
        SImg_Free(cc,fj2);
    end;

    SImg_Move(cc, fj,fi2);
    SImg_DecDebugLevel;
end;


{*************************************************************************}
{******* SImg_FindStars **************************************************}
{*************************************************************************}
type TSImg_FindStars_peak = record
            x,y : word;         { position }
            b   : simgfp;       { brightness }
            w,a : single;       { FWHM width and aspect ratio }
            s,v : boolean;      { saturated / valid }
        end;
     PSImg_FindStars_peak = ^TSImg_FindStars_peak;

{ find peaks }
threadprocedure SImg_FindStars_thread1(ti:longint; tn:longint; fi:TSImg_FPImage; gx:integer; gy:integer; minb:double; stars:PSImg_FindStars_peak);
const mbd = 20; { minimum distance to border }
var i,ia,iz,xi,xa,xz    : longint;
    yi,ya,yz,x,y        : longint;
    r,s                 : simgfp;
    d                   : ^simgfp;
begin
    ia:=(gx*gy*ti) div tn;
    iz:=(gx*gy*(ti+1)) div tn - 1;
    for i:=ia to iz do
    begin
        yi:=i div gx;
        xi:=i - yi*gx;
        xa:=(xi*fi.nx) div gx;
        xz:=((xi+1)*fi.nx) div gx - 1;
        ya:=(yi*fi.ny) div gy;
        yz:=((yi+1)*fi.ny) div gy - 1;
        x:=-1;  y:=-1;
        r:=0;
        for yi:=ya to yz do
        begin
            d:=fi.pt + yi*fi.sl;
            for xi:=xa to xz do
            begin
                s:=d[xi];
                if s>r then
                begin
                    r:=s;
                    x:=xi;
                    y:=yi;
                end;
            end;
        end;
        stars[i].x:=x;
        stars[i].y:=y;
        stars[i].b:=r;
        stars[i].v:=(x>=mbd) and (y>=mbd) and (x<fi.nx-mbd) and (y<fi.ny-mbd) and (r>minb);
    end;
end;

{ calculate FWHM's }
threadprocedure SImg_FindStars_thread2(ti:longint; tn:longint; fi:TSImg_FPImage; gx:integer; gy:integer; stars:PSImg_FindStars_peak; sat:double);
const k_fact = 2/pi;
      limit  = exp(ln(0.5)*4/(pi*k_fact));  { 0.25 @ k_fact=2/pi }
var i,x,y,ia,iz         : longint;
    j,k                 : int64;
    xi,xa,xz,yi,ya,yz   : longint;
    nx,ny               : longint;
    r                   : double;
    d                   : ^simgfp;
begin
    ia:=(gx*gy*ti) div tn;
    iz:=(gx*gy*(ti+1)) div tn - 1;
    nx:=fi.nx;
    ny:=fi.ny;
    d:=fi.pt;
    for i:=ia to iz do
        if stars[i].v  then
        begin
            x:=stars[i].x;
            y:=stars[i].y;
            d:=fi.pt + y*nx + x;
            { search boundaries }
            if stars[i].s then r:=sat*limit*0.5
            else r:=stars[i].b*limit*0.75;
            xa:=0;
            while ((x+xa)>0) and (d[xa]>r) do
                xa-=1;
            xz:=0;
            while ((x+xz)<nx-1) and (d[xz]>r) do
                xz+=1;
            ya:=0;
            while ((y+ya)>0) and (d[ya*nx]>r) do
                ya-=1;
            yz:=0;
            while ((y+yz)<ny-1) and (d[yz*nx]>r) do
                yz+=1;

            { search boundaries }
            d+=ya*nx;
            if stars[i].s then r:=sat*limit*0.8
            else r:=stars[i].b*limit;
            j:=0; k:=0;
            for yi:=ya to yz do
            begin
                for xi:=xa to xz do
                    if d[xi]>r then
                    begin
                        j+=sqr(yi)+sqr(xi);
                        k+=1;
                    end;
                d+=nx;
            end;
//          if (k<0) or (j<0) then simg_error('k='+int2str(k)+' j='+int2str(j)+' xa='+int2str(ya)+' xz='+int2str(yz)+' ya='+int2str(ya)+' yz='+int2str(yz));
            stars[i].w:=sqrt(k*k_fact);
            stars[i].a:=sqrt((6*j+0.5)/(sqr(k)+0.5));
//          if (abs(x-4502)<10) and (abs(y-8693)<10) then with stars[i] do writeln(stderr,'x=',x,' y=',y,' s=',s,' w=',w:0:2,' a=',a:0:2);
//          if stars[i].s then writeln(stderr,stars[i].x,'  ',stars[i].y,'  ',stars[i].a:0:3);
        end
        else begin
            stars[i].w:=1;
            stars[i].a:=1;
        end;
end;

{ Find save stars
    fi2    : input image (1 color image), should be BG correced
    stars  : output array, sorted by brightness
    sat    ; saturation level
    minb   : minimum brightness
    tol    : FWHM tolerance
    rad    : PSF radius
    result : FWHM
}
function SImg_FindStars(cc:CSImg_Config; fi2:TSImg_FPImage; var stars:array of TSImg_FindStars_peak; sat,minb, tol:double; rad:longint):double;
const fwhm_its = 3;     { FWHM iterations }
var i,xi,yi,gx,gy,tn,it : longint;
    n0,n1               : longint;
    a                   : simgfp;
    r,r1,t              : double;
    ss,s                : PSImg_FindStars_peak;
//ifdef[debug_findStars]
    fd                  : TSImg_FPImage;
//endif
threadvars;
begin
    if cc=nil then cc:=SImg_Config_Default;
    if not SImg_Recover(cc,fi2) then cc.error('SImg_GetPSF: Invalid image: fi2');
    if fi2.nc<>1 then cc.error('SImg_GetPSF: Invalid dimensions: fi2');
    if tol<1.1 then tol:=1.1;

    cc.info(cc.name+': Searching bright stars');

    { grid size }
    r:=sqrt(fi2.size/(high(stars)+1));
    gx:=round(3*fi2.nx/r)+10;
    gy:=round(3*fi2.ny/r)+10;
    getmem2(ss,gx*gy*sizeof(TSImg_FindStars_peak));

    { find local maxima }
    tn:=cc.threads;
    for i:=0 to tn-1 do
        dothread SImg_FindStars_thread1(i,tn, fi2, gx,gy, minb, ss);
    threadwait;
    for yi:=0 to gy-1 do
        for xi:=0 to gx-1 do
        begin
            s:=ss+yi*gx+xi;
            a:=s[0].b;
            if (yi>0   ) and (xi>0   ) and (a>s[-gx-1].b) then s[-gx-1].v:=false;
            if (yi>0   ) and               (a>s[-gx  ].b) then s[-gx  ].v:=false;
            if (yi>0   ) and (xi<gx-1) and (a>s[-gx+1].b) then s[-gx+1].v:=false;
            if               (xi>0   ) and (a>s[-1   ].b) then s[-1   ].v:=false;
            if               (xi<gx-1) and (a>s[ 1   ].b) then s[ 1   ].v:=false;
            if (yi<gy-1) and (xi>0   ) and (a>s[ gx-1].b) then s[ gx-1].v:=false;
            if (yi<gy-1) and               (a>s[ gx  ].b) then s[ gx  ].v:=false;
            if (yi<gy-1) and (xi<gx-1) and (a>s[ gx+1].b) then s[ gx+1].v:=false;
            s[0].s:=a>sat;
        end;

    { disable stars which are to close to borders }
    for i:=0 to gx*gy-1 do
        with ss[i] do
            if (x<rad) or (y<rad) or (x+rad>=fi2.nx) or (y+rad>=fi2.ny) then v:=false;

    { calculate fwhm's }
    for i:=0 to tn-1 do
        dothread SImg_FindStars_thread2(i,tn, fi2, gx,gy, ss, sat);
    threadwait;
//    SImg_FindStars_thread2(0,1, fi2, gx,gy, ss, sat);
    xi:=0; r:=0; n0:=0;
    for i:=0 to gx*gy-1 do
        with ss[i] do
        begin
            if not s then
            begin
//              if v then writeln(stderr,x,', ',y,':   w=',w:0:3,'  a=',a:0:3);
//              v:=v and (sqr(a*w)+1<sqr(tol)*(sqr(w)+1));      { FWHM regularization with 1 pixel in order to avoid problems due to discretization }
                if v then
                begin
                    r+=w;
                    xi+=1;
                end;
            end
//          else if sqr(a-0.1)>tol then v:=false;
            else if a>tol then v:=false;
            if v then n0+=1;
        end;
    for it:=0 to fwhm_its do
    begin
        if xi<1 then cc.error('No valid and non-saturated sources found within tolerance');
        r/=xi;
        t:=exp(ln(tol+0.7/r)*(2-it/fwhm_its));
//ifdef[debug_findStars]
        cc.info(cc.name+': tol='+real2str(t,-3)+'   fwhm='+real2str(r,-2)+'   n='+int2str(xi));
//endif
//      r0:=sqr(r/t);
//      if r0<0 then r0:=0 else r0:=sqrt(r0);
        r1:=r*t;
        r:=0; xi:=0;
        for i:=0 to gx*gy-1 do
            with ss[i] do
                if not s then
                begin
//                  if v and (w>=r0) and (w<=r1) then
                    if v and (w*a<=r1) then
                    begin
                        r+=w;
                        xi+=1;
                    end
                    else if it=fwhm_its then v:=false;
//                  if v and (abs(x-4502)<10) and (abs(y-8693)<10) then writeln(stderr,'x=',x,' y=',y,' s=',s,' w=',w:0:2,' a=',a:0:2,' r1=',r1:0:1,' b=',b);
                end;
    end;
    result:=r/xi;

    { sort output }
    n1:=n0;
    if high(stars)>=0 then
    begin
        xi:=-1; { idx last valid element }
        for i:=0 to gx*gy-1 do
            with ss[i] do
                if v then
                begin
                    n1-=1;
                    a:=b;
                    if xi<high(stars) then xi+=1
                    else if a<=stars[xi].b then continue;
                    yi:=xi;
                    while (yi>0) and (a>stars[yi-1].b) do
                    begin
                        stars[yi]:=stars[yi-1];
                        yi-=1;
                    end;
                    stars[yi]:=ss[i];
                end;
        for i:=xi+1 to high(stars) do
            stars[i].v:=false;
    end;
    cc.info(cc.name+': Filtered '+real2str((n1/n0)*100,-1)+'% ('+int2str(n1)+' / '+int2str(n0)+') sources.');

//ifdef[debug_findStars]
{    if fi2.size<50e6 then
    begin
        fd.pt:=nil;
        SImg_New(cc,fd, fi2.nx,fi2.ny,2);
        for i:=0 to gx*gy-1 do
            if ss[i].v then fd.pt[ ss[i].y*fd.sl + ss[i].x*2]:=ss[i].b;
        for i:=0 to high(stars)-1 do
            if stars[i].v then fd.pt[ stars[i].y*fd.sl + stars[i].x*2 + 1]:=stars[i].b;
        SImg_Filter(cc,fd,fd,result*FWHM_to_rad);
        SImg_WriteImageFile(cc,'findStars-1.fits',fd);

        SImg_Free(cc,fd);
    end;}
//endif

    freemem2(ss,gx*gy*sizeof(TSImg_FindStars_peak));
end;


{*************************************************************************}
{******* SImg_GetPSF *****************************************************}
{*************************************************************************}
{******* SImg_GetPSF_getdat **********************************************}
{ extract data from image and apply symmetry and monotony constraints
  ti    : thread index
  tn    : number of threads
  fi2   : input image
  stars : stars
  dat   : n x m x 2; data samples ( brightness + weight )
  n     : number of stars
  rad   : radius
  sat   : saturation / undefined limit
  sig   : noise floor as calculated by SImg_StarSubtract_NoiseFloor
}
threadprocedure SImg_GetPSF_getdat(ti:longint; tn:longint; fi2:TSImg_FPImage; stars:PSImg_FindStars_peak; dat:psimgfp; n:longint; rad:longint; sat:simgfp; sig:double);
const mon_tol   = 1;    { monotony tolerance at center }
var s                   : TSImg_FindStars_peak;
    q,t,u,v,dx,dy,t2,u2 : double;
    i,ia,iz,j,x,y,z     : longint;
    d,e                 : ^simgfp;
begin
    ia:=(ti*n) div tn;
    iz:=((ti+1)*n) div tn - 1;
    e:=dat + ((rad*2+1)*(rad+1))*ia*2;
    q:=1/sig;
    sig*=2;
    for i:=ia to iz do
    begin
        s:=stars[i];
        d:=fi2.pt + s.y*fi2.sl + s.x;
        for y:=-rad to rad do
        begin
            for x:=0 to rad do
            begin
                t:=d[y*fi2.sl+x];
                u:=d[-y*fi2.sl-x];
                if (x>1) or (abs(y)>1) then
                begin
                    if t>u then v:=u
                    else v:=t;
                    if v<0 then v:=sig
                    else v:=v*mon_tol*2/(1+sqrt(sqr(x)+sqr(y)))+sig;
                    t-=v;
                    u-=v;
                    z:=abs(y);
                    if x>z then z:=x;
                    z-=1;
                    dx:=x/z;
                    dy:=y/z;
                    while z>0 do
                    begin
                        j:=round(z*dy)*fi2.sl + round(z*dx);
                        t2:=d[j];
                        u2:=d[-j];
                        if u2<t2 then t2:=u2;
                        if t2<t then t:=t2;
                        if t2<u then u:=t2;
                        z-=1;
                    end;
                    t+=v;
                    u+=v;
                end;
                v:=0.5*(u+t);
                e[0]:=v;
                if (u>=sat) or (v>=sat) then e[0]:=0
                else begin
                    v:=v*q;
                    if v<1 then e[1]:=1
                    else e[1]:=1/(sqr(v-1)+1);
                end;
                e+=2;
            end;
        end;
    end;
end;

{******* SImg_GetPSF_getabc **********************************************}
{ Get factors, offsets and weight
  res   : m; reference PSF
  dat   : n x m x 2; data samples ( brightness + weight )
  abc   : n x 3; output buffer
  e     : exponent, weight is a^(-e)
}
procedure SImg_GetPSF_getabc(res,dat,abc:psimgfp; n,m:longint; e:double);
var s,f,w,w0,w1         : simgfp;
    r,s_f,s_s,s_1       : double;
    s_sf,s_sff,s_ssf    : double;
    i,j,i1              : longint;
begin
    w0:=0; w1:=0; i1:=0;
    for i:=0 to n-1 do
    begin
        s_1:=0; s_f:=0; s_s:=0;
        s_sf:=0; s_sff:=0; s_ssf:=0;
        for j:=0 to m-1 do
        begin
            s:=res[j];
            f:=dat[0];
            w:=dat[1];
            dat+=2;
            s_1+=w;
            s_f+=f*w;
            s_s+=s*w;
            s_sf+=s*f*w;
            s_sff+=s*sqr(f)*w;
            s_ssf+=sqr(s)*f*w;
        end;
        r:=s_1*s_sff - s_sf*s_f;
        if abs(r)<simg_mina then
        begin
            abc[i*3  ]:=0;
            abc[i*3+1]:=0;
            abc[i*3+2]:=0;
        end else
        begin
            r:=1/r;
            s:=r*(s_1*s_ssf-s_sf*s_s);
            f:=r*(s_sff*s_s-s_f*s_ssf);
            if s<0 then
            begin
                abc[i*3  ]:=0;
                abc[i*3+1]:=0;
                abc[i*3+2]:=0;
            end else
            begin
                abc[i*3  ]:=s;
                abc[i*3+1]:=f;
                if s>simg_mina then s:=exp(ln(s)*(-e))
                else s:=0;
                if s>w1 then
                begin
                    w0:=w1;     { 2nd largest weight }
                    w1:=s;
                    i1:=i;
                end;
                abc[i*3+2]:=s;
            end;
        end;
    end;
    { limit largest weight }
    if w1>2*w0 then abc[i1*3+2]:=2*w0
end;

{******* SImg_GetPSF_avg *************************************************}
{ calculate psf using average
  ti    : thread index
  tn    : number of threads
  res   : m; output PSF
  dat   : n x m; data samples
  abc   : n x 3; factor, offset, weight
}
threadprocedure SImg_GetPSF_avg(ti:longint; tn:longint; res:psimgfp; dat:psimgfp; abc:psimgfp; n:longint; m:longint);
var a,b         : simgfp;
    r,s,t,w     : double;
    i,j,ja,jz   : longint;
    d           : ^simgfp;
begin
    ja:=m*ti div tn;
    jz:=m*(ti+1) div tn-1;
    for j:=ja to jz do
    begin
        s:=0; t:=0;
        d:=dat+j*2;
        for i:=0 to n-1 do
        begin
            a:=abc[i*3];
            b:=abc[i*3+1];
            r:=d[0];
            w:=d[1]*abc[i*3+2];
            d+=2*m;
            s+=(a*r+b)*w;
            t+=w;
        end;
        if t<simg_mina then res[j]:=0
        else res[j]:=s/t;
    end;
end;

procedure SImg_GetPSF_avg(cc:CSImg_Config; res,dat,abc:psimgfp; n,m:longint);
var ti,tn       : longint;
threadvars;
begin
    tn:=cc.threads;
    for ti:=0 to tn-1 do
        dothread SImg_GetPSF_avg(ti,tn, res,dat,abc, n,m);
    threadwait;
end;

{******* SImg_GetPSF_quant ***********************************************}
{ calculate psf using averages between quantiles qmin and qmax
  ti    : thread index
  tn    : number of threads
  res   : m; output PSF
  dat   : n x m x 2; data samples ( brightness + weight )
  abc   : n x 3; factor, offset, weight
  qmin  : lower quantile
  qmax  : upper quantile
}
threadprocedure SImg_GetPSF_quant(ti:longint; tn:longint; res:psimgfp; dat:psimgfp; abc:psimgfp; n:longint; m:longint; qmin:double; qmax:double);
const hist_n = 512;
var q,r,s,t     : double;
    ra,rz,qa,qz : double;
    w           : double;
    i,j,ja,jz   : longint;
    hi          : longint;
    d           : ^simgfp;
    hist        : array[0..hist_n,0..1] of simgfp;  { [*,0]: weight; [*,1]:value*weight }
begin
    ja:=m*ti div tn;
    jz:=m*(ti+1) div tn-1;
    for j:=ja to jz do
    begin
        { get min/max }
        ra:=2; rz:=-2;
        d:=dat+2*j;
        for i:=0 to n-1 do
        begin
            if abc[i*3+2]*d[1]>simg_mina then
            begin
                r:=abc[i*3]*d[0]+abc[i*3+1];
                if r<ra then ra:=r;
                if r>rz then rz:=r;
            end;
            d+=2*m;
        end;

        rz-=ra;
        if rz<simg_mina then res[j]:=0
        else begin
            { get histogram }
            d:=dat+2*j;
            rz:=hist_n/rz;
            for i:=0 to hist_n do
            begin
                hist[i,0]:=0;
                hist[i,1]:=0;
            end;
            qz:=0;
            for i:=0 to n-1 do
            begin
                r:=abc[i*3]*d[0]+abc[i*3+1];
                w:=d[1]*abc[i*3+2];
                s:=(r-ra)*rz;
                if s<0 then s:=0;
                if s>hist_n then s:=hist_n;
                hi:=round(s);
                hist[hi,0]+=w;
                hist[hi,1]+=r*w;
                qz+=w;
                d+=2*m;
            end;

            { average between quantiles }
            qa:=qz*qmin;
            qz:=qz*qmax;
            i:=0; q:=0; r:=0; s:=0;
            while (i<=hist_n) and (q<qz) do
            begin
                t:=hist[i,0];
                q+=t;
                if q>qa then
                begin
                    r+=hist[i,1];
                    s+=t;
                end;
                i+=1;
            end;
            if s>simg_mina then res[j]:=r/s
            else res[j]:=0;
//          if s<=simg_mina then writeln(stderr,'qa=',qa,'   qz=',qz,'   q=',q,'   r=',r,'  s=',s);
        end;
    end;
end;

procedure SImg_GetPSF_quant(cc:CSImg_Config; res,dat,abc:psimgfp; n,m:longint; qmin,qmax:double);
var ti,tn       : longint;
threadvars;
begin
    tn:=cc.threads;
    for ti:=0 to tn-1 do
        dothread SImg_GetPSF_quant(ti,tn, res,dat,abc, n,m, qmin,qmax);
    threadwait;
end;


{******* SImg_GetPSF_normalize *******************************************}
{ normalize PSF such the sum is 1 and background is 0. Background is approximated by interpolation
  res   : PSF
  rad   : radius
}
procedure SImg_GetPSF_normalize(res:psimgfp; rad,rad2:longint);
const its = 100;        { negative crop compensation iterations }
var x,y,i,j,k,m,sr,sr2  : longint;
    x2,y2               : longint;
    r,s,a1,a2,b1,b2     : double;
    rr,rrd              : double;
    d                   : ^simgfp;
// define[SImg_GetPSF_normalize_debug]
//ifdef[SImg_GetPSF_normalize_debug]
    fi                  : TSImg_FPImage;
//endif
begin
    m:=(rad2*2+1)*(rad2+1);
    sr:=sqr(rad);
    sr2:=sqr(rad2);

    { background approximation by polynomials 1 and x*y which is used for the inner region }
    a1:=simg_mina;  a2:=simg_mina;
    b1:=0;  b2:=0;
    rrd:=1/(rad2+1-rad);
    for y:=-rad2 to rad2 do
    begin
        i:=1;
        d:=res + (y+rad2)*(rad2+1);
        for x:=0 to rad2 do
        begin
            j:=sqr(x)+sqr(y);
            if (j>sr) and (j<=sr2) then
            begin
                s:=i*(1-sqr((sqrt(j)-rad)*rrd));
                r:=d[x]*s;
                k:=x*y;
                a1+=s;
                a2+=sqr(int64(k))*s;
                b1+=r;
                b2+=r*k;
            end;
            i:=2;
        end;
    end;
    b1/=a1;
    b2/=a2;

//ifdef[SImg_GetPSF_normalize_debug]
    { debug image, approximation test }
    fi.pt:=nil;
    SImg_New(SImg_Config_Default, fi, 2*rad2+1,2*rad2+1,1, false);
    d:=fi.pt;
    a1:=0; a2:=0;
    for y:=-rad2 to rad2 do
    begin
        for x:=-rad2 to rad2 do
        begin
            j:=sqr(x)+sqr(y);
            if j<=sr then s:=1
            else if j>sr2 then s:=0
            else s:=1-sqr((sqrt(j)-rad)*rrd);
            if x<0 then r:=res[(rad2-y)*(rad2+1) - x]
            else r:=res[(rad2+y)*(rad2+1) + x];
            if j>sr then
            begin
                a1+=sqr(r)*s;
                a2+=sqr(r-b1-b2*x*y)*s;
            end;
            d^:=(r-b1-b2*x*y)*s;
            d+=1;
        end;
    end;
    writeln(stderr,'a1=',a1,'   a1-a2=',a1-a2);
    SImg_WriteImageFile(SImg_Config_Default,'psf_normalize-fa.fits',fi);
    d:=fi.pt;
    for y:=-rad2 to rad2 do
    begin
        for x:=-rad2 to rad2 do
        begin
            if x<0 then d^:=res[(rad2-y)*(rad2+1) - x]
            else d^:=res[(rad2+y)*(rad2+1) + x];
            d+=1;
        end;
    end;
    SImg_WriteImageFile(SImg_Config_Default,'psf_normalize-fi.fits',fi);
//endif

    { interpolation / approximation }
    rr:=1/rad;
    rrd:=1/(rad2-rad);
    for y:=-rad2 to rad2 do
    begin
        d:=res + (y+rad2)*(rad2+1);
        for x:=0 to rad2 do
        begin
            j:=sqr(x)+sqr(y);
            if j>sr then continue;  { outer values are set later because they are needed for interpolation }
            if j>0 then
            begin
                r:=sqrt(j);
                a1:=0; a2:=0;
                for k:=rad+1 to rad2 do
                begin
                    s:=k/r;
                    x2:=round(x*s);
                    y2:=round(y*s);
                    if sqr(x2)+sqr(y2)>sr then  { inner values may be overwritten }
                    begin
                        s:=1-sqr((k-rad-1)*rrd);
                        a1+=res[(y2+rad2)*(rad2+1) + x2]*s;
                        a2+=s;
                    end;
                end;
                r*=rr;
//              d[x]-=a1/a2;
                d[x]-=r * a1/a2                         { interpolation }
                      + (1-r) * (b1 + b2*x*y);          { inner approximation by polynomials }
            end
            else d[x]-=b1;
        end;
    end;
    for y:=-rad2 to rad2 do
    begin
        d:=res + (y+rad2)*(rad2+1);
        for x:=0 to rad2 do
            if sqr(x)+sqr(y)>sr then d[x]:=0;   { set outer values }
    end;

//ifdef[SImg_GetPSF_normalize_debug]
    { debug image }
    d:=fi.pt;
    for y:=-rad2 to rad2 do
    begin
        for x:=-rad2 to rad2 do
        begin
            if x<0 then d^-=res[(rad2-y)*(rad2+1) - x]
            else d^-=res[(rad2+y)*(rad2+1) + x];
            d+=1;
        end;
    end;
    SImg_WriteImageFile(SImg_Config_Default,'psf_normalize-bg.fits',fi);
    SImg_Free(SImg_Config_Default, fi);
//endif

    { negative crop compensation }
    for k:=1 to its do
    begin
        i:=0; r:=0;
        for y:=-rad to rad do
        begin
            d:=res + (y+rad2)*(rad2+1);
            j:=sr-sqr(y);
            for x:=0 to rad do
                if sqr(x)<=j then
                begin
                    s:=d[x];
                    if s<0 then
                    begin
                        d[x]:=0;
                        r+=s;
                    end;
                i+=1;
                end;
        end;
        r/=i;
//      writeln(stderr,'crop_comp=',r);

        for y:=-rad to rad do
        begin
            d:=res + (y+rad2)*(rad2+1);
            j:=sr-sqr(y);
            for x:=0 to rad do
                if sqr(x)<=j then d[x]+=r;
        end;
        if r>-simg_mina then break;
    end;

    { scale }
    i:=0; r:=0;
    for y:=-rad2 to rad2 do
    begin
        d:=res + (y+rad2)*(rad2+1);
        if y<0 then x:=0
        else x:=1;
        for x:=x to rad2 do
            r+=d[x];
    end;
    r:=1/(2*r+res[rad2*(rad2+1)]);
    for x:=0 to m-1 do
        res[x]*=r;
end;


{******* SImg_GetPSF_rad *************************************************}
{ return PSF width as radius
  res   : PSF
  rad   : radius
}
function SImg_GetPSF_rad(res:psimgfp; rad:longint):double;
const thr = 5e-3;       { ignore values less then this limit relative to value at center }
var x,y,i,j     : longint;
    r,s,t,u     : double;
    d           : ^simgfp;
begin
    { first sigma estimation with brightness filter }
    u:=res[rad*(rad+1)]*thr;
    s:=0; t:=0;
    for y:=-rad to rad do
    begin
        d:=res + (y+rad)*(rad+1);
        if y<=0 then x:=0
        else x:=1;
        for x:=x to rad do
        begin
            r:=d[x];
            if r>u then
            begin
                i:=sqr(x)+sqr(y);
                if i=0 then r*=2        { half-weight center pixel because we consider only every second pixel }
                else r/=0.25+sqrt(i);   { perimeter compensation because 2D-sum }
                s+=i*r;
                t+=r;
            end;
        end;
    end;

    { final sig estimation with brightness and radius filter }
    j:=2+round(s/t*sqr(ln(thr))); { assuming exponential distribution }
    s:=0; t:=0;
    for y:=-rad to rad do
    begin
        d:=res + (y+rad)*(rad+1);
        if y<=0 then x:=0
        else x:=1;
        for x:=x to rad do
        begin
            i:=sqr(x)+sqr(y);
            if i<=j then
            begin
                r:=d[x];
                if r>u then
                begin
                    if i=0 then r*=2            { half-weight center pixel because we consider only every second pixel }
                    else r/=0.25+sqrt(i);       { perimeter compensation because 2D-sum }
                    s+=i*r;
                    t+=r;
                end;
            end;
        end;
    end;

    result:=sqrt(2*s/t);        { sqr(rad) = 2*sqr(sig) = 2*s/t }
end;


{******* SImg_GetPSF *****************************************************}
{ Calculate PSF
    fi     : input image (1 color image)
    fpsf   : output PSF
    tol    : FWHM tolerance as factor
    rad    : PSF radius
    rm     : star mask. Or with mask at position found by SImg_FindStars (optional)
    mask   : see rm
}
procedure SImg_GetPSF(cc:CSImg_Config; fi:TSImg_FPImage; var fpsf:TSImg_FPImage; tol:double; rad:longint; rm:TSImg_RDImage; mask:byte);
const min_stars = 250;  { minimum desired number of stars for PSF calculation }
      minb      = 25;   { minimum brightness relative to noise floor }
var fi2         : TSImg_FPImage;
    stars       : array of TSImg_FindStars_peak;
    n,m,rad2    : longint;      { see below }
    abc         : ^simgfp;      { n x 3, factors offset weight}
    dat         : ^simgfp;      { n x m, data samples }
    res         : ^simgfp;      { 0..2*rad x 0..rad, result }
    sat,fwhm    : double;
    r,t,u,sig   : double;
    i,x,y       : longint;
    d,e         : ^simgfp;
threadvars;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_GetPSF: fi='+SImg_InfoStr2(fi)+' fpsf='+SImg_InfoStr2(fpsf)+' tol='+real2str(tol)+' rad='+int2str(rad)+' rm='+SImg_InfoStr2(rm)+' mask='+int2str(mask));
    if not SImg_Recover(cc,fi) then cc.error('SImg_GetPSF: Invalid image: fi');
    if fi.nc<>1 then cc.error('SImg_GetPSF: Invalid dimensions: fi');
    if not SImg_OverwriteAlert(fi,fpsf) then SImg_Free(cc,fpsf);
    if mask=0 then rm.pt:=nil;
    if rm.pt<>nil then
    begin
        if not SImg_Recover(cc,rm) and (rm.fmt<>SImg_RawByte) then cc.error('SImg_GetPSF: Invalid image: rm');
        if rm.size<>fi.size then cc.error('SImg_GetPSF: Invalid dimensions: rm');
    end;

    if tol<1.1 then tol:=1.1;
    if rad<3 then rad:=3;

    { background correction }
    fi2.pt:=nil;
    SImg_SubtractBackground(cc, fi,fi2, rad,0.25*rad,-1, sat,sig,r); { r=offs }
    for i:=0 to fi2.size-1 do
        fi2.pt[i]-=r;
    sat-=r;

    { find stars }
    n:=fi2.size div sqr(rad*2+1);
    if n<min_stars then n:=min_stars;
//    cc.info(cc.name+': n_0='+int2str(n));
    rad2:=round(rad*sqrt(pi/2)+0.5);    { with boundary for background approximation }
//    rad2:=round(rad*sqrt(2)+0.5);     { with boundary for background approximation }
    setlength(stars,n);
    fwhm:=SImg_FindStars(cc, fi2, stars, sat,minb*sig, tol, rad2);
    while (n>0) and not stars[n-1].v do
        n-=1;
    m:=(rad2*2+1)*(rad2+1);
    cc.info(cc.name+': Calculating PSF using '+int2str(n)+' sources with average FWHM '+real2str(fwhm,-2));

    { save stars }
    if (rm.pt<>nil) and (mask<>0) then
    begin
        for i:=0 to n-1 do
        begin
            x:=stars[i].y*rm.sl + stars[i].x;
            rm.ptb[x]:=rm.ptb[x] or mask;
        end;
    end;

    { initial psf }
    getmem2(res, m*sizeof(simgfp));
    t:=-1/sqr(fwhm*FWHM_TO_RAD);
    u:=0;
    d:=res;
    for y:=-rad2 to rad2 do
        for x:=0 to rad2 do
        begin
            r:=exp((sqr(x)+sqr(y))*t);
            d^:=r; d+=1;
            u+=r;
        end;
    u:=1/u;
    for i:=0 to m-1 do
        res[i]*=u;
//    writeln(stderr,'rad=',fwhm*FWHM_TO_RAD:0:3,'   rad0=',SImg_GetPSF_rad(res,rad):0:3);

    { extract data from image and apply symmetry and monotony constraints }
    getmem2(dat, m*n*2*sizeof(simgfp));
    y:=cc.threads;
    for x:=0 to y-1 do
        dothread SImg_GetPSF_getdat(x,y, fi2, @stars[0], dat, n, rad2, sat, sig);
    threadwait;
    SImg_Free(cc,fi2);

    { Iteration 1: only non-saturated stars, e=1 }
    getmem2(abc, n*3*sizeof(simgfp));
    SImg_GetPSF_getabc(res,dat,abc, n,m, 1);
    d:=dat + rad2*(rad2+1)*2;
    for i:=0 to n-1 do
        if d[i*m*2]>sat then abc[i*3+2]:=0;
//    SImg_GetPSF_avg(cc, res,dat,abc, n,m );
    SImg_GetPSF_quant(cc, res,dat,abc, n,m, 0.35,0.65);
    SImg_GetPSF_normalize(res, rad,rad2);
//    writeln(stderr,'rad1=',SImg_GetPSF_rad(res,rad):0:3);

    { Iteration 2: all stars, e=2 }
    SImg_GetPSF_getabc(res,dat,abc, n,m, 2);
//  SImg_GetPSF_avg(cc, res,dat,abc, n,m);
    SImg_GetPSF_quant(cc, res,dat,abc, n,m, 0.35,0.65);
    SImg_GetPSF_normalize(res, rad,rad2);

    { generate output }
    SImg_New(cc, fpsf, rad*2+1,rad*2+1,1, false);
    e:=fpsf.pt + rad*fpsf.sl + rad;
    for y:=-rad to rad do
    begin
        d:=res+(y+rad2)*(rad2+1);
        for x:=0 to rad do
        begin
            e[ y*fpsf.sl+x]:=d^;
            e[-y*fpsf.sl-x]:=d^;
            d+=1;
        end;
    end;
{    r:=0;
    for i:=0 to fpsf.size-1 do
        r+=fpsf.pt[i];
    writeln(stderr,'sum=',r:0:3);  }

    freemem2(res, m*sizeof(simgfp));
    freemem2(dat, m*n*2*sizeof(simgfp));
    freemem2(abc, n*3*sizeof(simgfp));
    setlength(stars,0);
    SImg_DecDebugLevel;
end;


{*************************************************************************}
{******* SImg_StarSubtract ***********************************************}
{*************************************************************************}
{******* SImg_StarSubtractGetMask **************************************}
{ pattern 2: line }
function SImg_StarSubtract_GetMask_pat2(cc:CSImg_Config; fs:TSImg_FPImage; xm,ym:longint):double;
var x,y : longint;
begin
    result:=0;
    for y:=0 to fs.ny-1 do
        for x:=0 to fs.nx-1 do
            result+=fs.pt[y*fs.sl+x]/sqrt((sqr(x-xm)+sqr(y-ym))*sqr(pi)+1);
end;

{ Find stars using deconvolution
    fi     : input from which maxima are calculated
    rm     : mask input and output
             bit 0: is set at center of new source
             bit 1: input star candidate
             bit 2: save input star
}
threadprocedure SImg_StarSubtract_GetMask_thread1(it:longint; tn:longint; fi:TSImg_FPImage; rm:TSImg_RDImage);
var r           : simgfp;
    x,y,ya,yz   : longint;
    a           : byte;
    d           : ^simgfp;
    b           : ^byte;
begin
    ya:=1 + ((rm.ny-2)*it) div tn;
    yz:=((rm.ny-2)*(it+1)) div tn;
    for y:=ya to yz do
    begin
        d:=fi.pt + y*fi.sl + 1;
        b:=rm.ptb + y*rm.sl + 1;
        for x:=1 to rm.nx-2 do
        begin
            a:=b^ and 254;
            r:=d^;      { wiener deconvolution }
            if ( a and 6=0 )            { no input star }
                and (r>0) and (r>d[-fi.sl-1]) and (r>d[-fi.sl]) and (r>d[-fi.sl+1]) and (r>d[-1]) and (r>d[1]) and (r>d[fi.sl-1]) and (r>d[fi.sl]) and (r>d[fi.sl+1])                                           { maximum with 3x3 area }
                and (b[-rm.sl-1] and 6=0) and (b[-rm.sl] and 6=0) and (b[-rm.sl+1] and 6=0) and (b[-1] and 6=0) and (b[1] and 6=0) and (b[rm.sl-1] and 6=0) and (b[rm.sl] and 6=0) and (b[rm.sl+1] and 6=0)     { no input star within 3x3 area }
                then a:=a or 1;
{           if ((abs(x-7658*2)<3) and (abs(y-5739*2)<3)) or
                ((abs(x-9870)<3) and (abs(y-14977)<3))
                then writeln(stderr,'x=',x,' y=',y,': r='+real2str(r),'  s='+real2str(s),'  b='+int2str(b^),'  a='+int2str(a),'  sat='+real2str(sat)); }
            b^:=a;
            d+=1;
            b+=1;
        end;
    end;
end;

{ Calculate i / (o*s)
    fi     : i
    ti     : o*s
}
threadprocedure SImg_StarSubtract_GetMask_thread2(it:longint; tn:longint; fi:TSImg_FPImage; ti:TSImg_FFTImage);
var x,y,ya,yz   : longint;
    r,s         : simgfp;
    d           : ^fftfp;
    e           : ^simgfp;
begin
    ya:=(ti.ny*it) div tn;
    yz:=(ti.ny*(it+1)) div tn - 1;
    for y:=ya to yz do
    begin
        d:=ti.pt + y*ti.nxc*2;
        x:=0;
        if y<fi.ny then
        begin
            e:=fi.pt + y*fi.sl;
            for x:=0 to fi.nx-1 do
            begin
                r:=e[x];
                if r<0 then r:=0;
                s:=d[x];
                if s<=1e-2*r then d[x]:=1e2
                else d[x]:=r/s;
            end;
            x:=fi.nx;
        end;
        for x:=x to ti.nxr-1 do
            d[x]:=0;
    end;
end;


{ Find initial source mask
    fi1    : Background subtracted input image
    fs     : PSF
    rm     : mask input and output.
             bit 1: star candidate input and output
             bi2 2: save star input and output
    fj     : initial solution for a (undefined if rm and 6=0, otherwise non-negative)
    getmask: bit 0: find new stars
             bit 1: filter input candidates, otherwise they will be set to save stars
    inf    : PSF info
    sat    : saturation threshold
    sig2   : brightness limit
    offset : fi2 offset
    tol    : tolerance
    ff     : filter strength
}
procedure SImg_StarSubtract_GetMask(cc:CSImg_Config; fi1,fs:TSImg_FPImage; rm:TSImg_RDImage; var fj:TSImg_FPImage; getmask:longint; inf:TSImg_PSFinfo; sat,sig2,offs,ff:double);
const brad_thr  = 0.05;  { threshold for determination of background filter size }
      rl_iter   = 3;     { Richardson-Lucy iteration for initial solution }
      min_sfit  = 4;     { minimum shape filter iterations }
      bgfact    = 0.25;  { background factor for fi3 at ff=1 }
      dc_snr    = 20;    { SNR parameter for wiener deconvolution }
      dc_q      = 0.999; { Use (convoluted) input for star center calculation above this quantile }
      r9        = 1/9;

var fi2,fs2             : TSImg_FPImage; { background subtracted input image and PSF, filter radius brad  }
    fi3                 : TSImg_FPImage; { denoised fi2 and PSFs (with and w/o tolerance), filter radius nrad }
    fi4,fs4             : TSImg_FPImage; { smoothed fi3 and PSF for shape filtering }
    ts,ti               : TSImg_FFTImage;
    m2,m4               : simgfp;        { maximum of fs2 and fs4 }
    brad,nrad,srad      : double;
    srad_prev           : double;
    q,r,s,t,u,offs2     : double;
    i,j,x,y,it          : longint;
    b,c                 : byte;
    bo                  : boolean;
    d,e                 : ^simgfp;
threadvars;
begin
    if cc=nil then cc:=SImg_Config_Default;
    if not SImg_Recover(cc,fi1) then cc.error('SImg_StarSubtract_GetMask: Invalid image: fi1');
    if fi1.nc<>1 then cc.error('SImg_StarSubtract_GetMask: Invalid dimensions: fi1');
    if not SImg_Recover(cc,fs) then cc.error('SImg_StarSubtract_GetMask: Invalid image: fs');
    if fs.nc<>1 then cc.error('SImg_StarSubtract_GetMask: Invalid dimensions: fs');
    if not SImg_Recover(cc,rm) and (rm.fmt<>SImg_RawByte) then cc.error('SImg_StarSubtract_GetMask: Invalid image: rm');
    if rm.size<>fi1.size then cc.error('SImg_StarSubtract_GetMask: Invalid dimensions: rm');

    { denoising filter }
    fs2.pt:=nil;
    nrad:=SImg_GetPSFWidth(cc,fs,0.25)/(4*sqrt(ln(4)));
    SImg_Filter(cc, fs,fs2, nrad);

    { background subtraction radius }
    brad:=SImg_GetPSFWidth(cc,fs2,brad_thr)*0.5;        { = FWHM at Gaussian PSF }
    cc.info(cc.name+': Background subtraction with interpolation radius '+real2str(brad,-1)+' and deconvolution');

    { subtract offset and crop negative values }
    fi3.pt:=nil;
    SImg_New(cc, fi3, fi1.nx,fi1.ny,1);
    for i:=0 to fi1.size-1 do
    begin
        r:=fi1.pt[i]-offs;
        if r<0 then r:=0;
        fi3.pt[i]:=r;
    end;
    SImg_Filter(cc, fi3,fi3, nrad);

    { subtract background }
    fi2.pt:=nil;
    SImg_FBGFilter(cc, fi3,fi2, brad, true);
    SImg_FBGFilter(cc, fs2,fs2, brad, true);
    t:=SImg_GetQuantile(cc, fi2, 0.75);
    u:=ff*bgfact;
    if u<0 then u:=0;
    for i:=0 to fi1.size-1 do
    begin
        s:=fi2.pt[i];
        r:=fi1.pt[i]-offs-s;
        if r<0 then r:=0;
        fi2.pt[i]:=r;
        if s<=t then fi3.pt[i]-=s
        else fi3.pt[i]-=s-(s-t)*u;
        rm.ptb[i]:=rm.ptb[i] and 254;
    end;
    for i:=0 to fs.size-1 do
    begin
        r:=fs.pt[i]-fs2.pt[i];
        if r<0 then r:=0;
        fs2.pt[i]:=r;
    end;
{    SImg_WriteImageFile(cc,'starSubtract-getmask-fi1.fits',fi1);
    SImg_WriteImageFile(cc,'starSubtract-getmask-fi2.fits',fi2);
    SImg_WriteImageFile(cc,'starSubtract-getmask-fs.fits',fs);
    SImg_WriteImageFile(cc,'starSubtract-getmask-fs2.fits',fs2); }

    { wiener deconvolution }
    offs2:=SImg_GetQuantile(cc, fi2, 0.5);
    for i:=0 to fi2.size-1 do
        fi2.pt[i]-=offs2;
    ts.pt:=nil; ti.pt:=nil;
    SImg_New_findbase(cc, ts, fi2.nx+fs2.nx-1, fi2.ny+fs2.ny-1);
    SImg_CopyR(cc, fs2,ts, 0, -inf.xm,-inf.ym);
    SImg_FFT_r2c(cc,ts);
    SImg_New(cc, ti, ts.nxr,ts.ny);
    SImg_Copy(cc, fi2,ti,0);
    SImg_FFT_r2c(cc,ti);
    s:=8*(2-sqrt(fs2.pt[inf.ym*fs2.sl+inf.xm]/ts.pt[0]));  { SNR for wiener regularization: 8..16 }
    SImg_FFTWiener(cc, dc_snr, ti,ti,ts);
    SImg_FFT_c2r(cc,ti);

    { Post-processing deconvolution result used to estimate star centers.
      <=0       : deconvolution result is used
      >=q(fc_q) : background subtracted input fi2 (=deconvolution input)
      >=0.5*sat : 3x3 pixel avg filter
    }
    fi4.pt:=nil;
    SImg_Duplicate(cc, fi2,fi4, true);
    q:=SImg_GetQuantile(cc,fi2,dc_q);
    if q<simg_mina then cc.error('SImg_StarSubtract_GetMask: Input image seems to contain no data');
    if q*8>sat then q:=sat/8;
    q:=1/q;
    r:=1/(0.5*sat*q-1);
    for y:=1 to fi4.ny-2 do
    begin
        d:=fi2.pt + y*fi2.sl;
        e:=fi4.pt + y*fi4.sl;
        j:=y*2*ti.nxc;
        for x:=1 to fi4.nx-2 do
        begin
            d+=1; e+=1;
            s:=d^;
            t:=s*q;
            if t<=0 then
            begin
                e^:=ti.pt[j+x];
                continue;
            end;
            if t<=1 then
            begin
                e^:=s*t + ti.pt[j+x]*(1-t);
                continue;
            end;
            t:=(t-1)*r;
            if t>1 then t:=1;
            e^:=(d[-fi2.sl-1] + d[-fi2.sl] + d[-fi2.sl+1] + d[-1] + s + d[+1] + d[fi2.sl-1] + d[fi2.sl] + d[fi2.sl+1])*r9*t + s*(1-t);
        end;
    end;
    SImg_Filter(cc, fi4,fi4, sqrt(0.5+sqr(SImg_GetPSFwidth(cc,fs,0.25)/8)) );   { denoising }
//    SImg_WriteImageFile(cc,'starSubtract-getmask-wiener.fits',fi4);

    { find new candidates center/maximum filter }
    if getmask and 1<>0 then
    begin
        for i:=0 to cc.threads-1 do
            dothread SImg_StarSubtract_GetMask_thread1(i,cc.threads, fi4,rm);
        threadwait;
    end;

    { set input star candidates either as save (getmask and 2=0) or as candidates }
    if getmask and 2=0 then c:=6        { input stars ara save }
    else c:=1;
    for i:=0 to rm.size-1 do
    begin
        b:=rm.ptb[i];
        if (b and 7<>0) and (fi1.pt[i]>0.5*sat) then rm.ptb[i]:=(b and 254) or 6        { saturated inputs (half limit due to smoothing) and candidates are marked as save }
        else if b and 6<>0 then rm.ptb[i]:=(b and 248) or c;                            { handle inputs }
    end;
//    SImg_WriteImageFile(cc,'starSubtract-getmask-mask0.pgm',rm);

    { initial solution using masked RL deconvolution. Must be ran before filter otherwise result may be to large. }
    cc.info(cc.name+': Calculating initial solution');
    for y:=0 to ti.ny-1 do
    begin
        j:=y*ti.nxc*2;
        x:=0;
        if y<rm.ny then
        begin
            i:=y*rm.sl;
            for x:=0 to rm.nx-1 do
                if rm.ptb[i+x] and 7=0 then
                begin
                    ti.pt[j+x]:=0;
                    fi4.pt[i+x]:=0;
                end
                else begin
                    r:=fi4.pt[i+x];
                    if r<0 then
                    begin
                        r:=0;
                        fi4.pt[i+x]:=r;
                    end;
                    ti.pt[j+x]:=0;
                end;
            x:=rm.nx;
        end;
        for x:=x to ts.nxr-1 do
            ti.pt[j+x]:=0;
    end;
    for it:=1 to rl_iter do
    begin
        SImg_FFT_r2c(cc, ti);
        SImg_Mul_C(cc, ti,ts, ti);
        SImg_FFT_c2r(cc, ti);
        for i:=0 to cc.threads-1 do
            dothread SImg_StarSubtract_GetMask_thread2(i,cc.threads, fi2,ti);
        threadwait;
        SImg_FFT_r2c(cc, ti);
        SImg_MulK_C(cc, ti,ts, ti);
        SImg_FFT_c2r(cc, ti);
        bo:=(getmask<>0) and (it=rl_iter);
        u:=sig2/inf.max;
        for y:=0 to ti.ny-1 do
        begin
            j:=y*ti.nxc*2;
            x:=0;
            if y<rm.ny then
            begin
                i:=y*rm.sl;
                for x:=0 to rm.nx-1 do
                begin
                    b:=rm.ptb[i+x] and 7;
                    if b=0 then ti.pt[j+x]:=0
                    else begin
                        r:=fi4.pt[i+x]*ti.pt[j+x];
                        if r<0 then r:=0;
                        ti.pt[j+x]:=r;
                        fi4.pt[i+x]:=r;
                        { brightness filter using initial solution from deconvolution }
{                       if ((abs(x-7658*2)<10) and (abs(y-5739*2)<10)) or
                           ((abs(x-4934*2)<10) and (abs(y-7488*2)<10))
                            then writeln('x=',x,' y=',y,': r='+real2str(r)); }
                        if bo and (b=1) and (r<u) then rm.ptb[i+x]:=rm.ptb[i+x] and 254;
                    end;
                end;
                x:=rm.nx;
            end;
            for x:=x to ts.nxr-1 do
                ti.pt[j+x]:=0;
        end;
//      SImg_WriteImageFile(cc,'starSubtract-getmask-fj-'+int2str(it)+'.fits',fi4);
    end;
    SImg_Free(cc,ti);
    SImg_Free(cc,ts);
    SImg_Move(cc, fi4,fj);

    { brightness filter using fi2 }
    if getmask<>0 then
    begin
        SImg_Filter(cc, fi2,fi4, nrad);
        for i:=0 to rm.size-1 do
            if (rm.ptb[i] and 1=1) and (fi2.pt[i]<sig2) and (fi4.pt[i]<sig2) then rm.ptb[i]:=rm.ptb[i] and 254;
    end;
    SImg_Free(cc, fi2);
    SImg_Free(cc, fi4);

    if getmask<>0 then
    begin
        { shape filter }
        fi4.pt:=nil; fs4.pt:=nil;
        srad:=2*nrad; srad_prev:=0;
        u:=SImg_StarSubtract_GetMask_pat2(cc, fs2 , inf.xm,inf.ym);
        m2:=fs2.pt[0];
        for i:=1 to fs2.size-1 do
            if fs2.pt[i]>m2 then m2:=fs2.pt[i];
        it:=0;
        repeat
            r:=sqrt(sqr(srad)-sqr(srad_prev));
            if fs4.pt=nil then SImg_Filter(cc, fs2,fs4, srad )
            else SImg_Filter(cc, fs4,fs4, r );
            if fi4.pt=nil then SImg_Filter(cc, fi3,fi4, srad)
            else SImg_Filter(cc, fi4,fi4, r);
            m4:=fs4.pt[0];
            for i:=1 to fs4.size-1 do
                if fs4.pt[i]>m4 then m4:=fs4.pt[i];
            q:=u/SImg_StarSubtract_GetMask_pat2(cc, fs4 , inf.xm,inf.ym);
            r:=1-m4/m2;
            cc.debug('SImg_StarSubtract_GetMask: srad='+real2str(srad,-2)+' fact='+real2str(q)+' sig2_fact='+real2str(r));
            { filter candidates  }
            r*=sig2;
            for i:=0 to rm.size-1 do
            begin
                b:=rm.ptb[i];
                if b and 3=1 then
                begin
                    s:=fi3.pt[i];
                    t:=fi4.pt[i];
                    if s<t+r then rm.ptb[i]:=b and 254
                    else begin
                        t-=offs2;
                        if t<0 then t:=0;
                        if s>t*q+offs2 then rm.ptb[i]:=(b and 254) or 2;
                    end;
                end;
            end;
            it+=1;
            { next step  }
            srad_prev:=srad;
            srad:=sqrt(sqr(nrad) + sqr(srad*1.2));
        until (srad*(0.2+ff)>brad*(1+ff)) and (it>min_sfit);
        SImg_Free(cc,fi4);
        SImg_Free(cc,fs4);
        for i:=0 to rm.size-1 do
        begin
            b:=rm.ptb[i];
            if b and 3=1 then rm.ptb[i]:=b and 254;
        end;
    end;
    SImg_Free(cc,fs2);
    SImg_Free(cc,fi3);
{    SImg_Copy(cc, ti,fi3, 0);
    for i:=0 to fi3.size-1 do
        if fi3.pt[i]<0 then fi3.pt[i]:=0;
    SImg_WriteImageFile(cc,'starSubtract-getmask-fj-0.fits',fi3);}
//    if getmask<>0 then SImg_WriteImageFile(cc,'starSubtract-getmask-mask1.pgm',rm);
end;

{******* SImg_StarSubtract_Getmask2 **************************************}
{ Find extended saturated areas
    fi2    : Background subtracted input image
    fs     : PSF
    rm     : Bit 5 (mask 32) is set at extended area and  Bits 1 and 2 are unset
    sat    : saturation level
}
procedure SImg_StarSubtract_GetMask2(cc:CSImg_Config; fi2,fs:TSImg_FPImage; rm:TSImg_RDImage; sat:simgfp);
const oe_ratio = 100;   { overexposure ratio }
var i,j,rad,x,y : longint;
    xa,ya,xz,yz : longint;
    s,t         : double;
begin
    if not SImg_Recover(cc,fi2) then cc.error('SImg_StarSubtract_GetMask2: Invalid image: fi2');
    if fi2.nc<>1 then cc.error('SImg_StarSubtract_GetMask2: Invalid dimensions: fi2');
    if not SImg_Recover(cc,fs) then cc.error('SImg_StarSubtract_GetMask2: Invalid image: fs');
    if fs.nc<>1 then cc.error('SImg_StarSubtract_GetMask: Invalid dimensions2: fs');
    if not SImg_Recover(cc,rm) and (rm.fmt<>SImg_RawByte) then cc.error('SImg_StarSubtract_GetMask2: Invalid image: rm');
    if rm.size<>fi2.size then cc.error('SImg_StarSubtract_GetMask2: Invalid dimensions: rm');
    s:=SImg_GetPSFWidth(cc,fs,1/oe_ratio)*0.5+0.7;
    rad:=round(2*s+3);
    s:=sat*(1+0.1*s); { see SImg_SubtractBackground }
    t:=sat*0.5;
    for i:=0 to fi2.size-1 do
        if fi2.pt[i]>s then
        begin
            y:=i div fi2.sl;
            x:=i - y*fi2.sl;
            xa:=x-rad; if xa<0 then xa:=0;
            ya:=y-rad; if ya<0 then ya:=0;
            xz:=x+rad; if xz>=fi2.nx then xz:=fi2.nx-1;
            yz:=y+rad; if yz>=fi2.ny then yz:=fi2.ny-1;
            for y:=ya to yz do
            begin
                j:=y*fi2.sl;
                for x:=xa to xz do
                    if fi2.pt[j+x]>t then rm.ptb[j+x]:=(rm.ptb[j+x] and 249) or 32;
            end;
        end;
end;

{******* SImg_StarSubtract_WidthFilter_approx ****************************}
{ Approximate fi2 at (x0,y0) by constant, normal PSF fpsf and a second PSF fr, a wider reference PSF.
    fi2    : background corrected input (1 color image)
    fpsf   : input PSF s
    fr0    : reference PSF 0
    fr1    : reference PSF 1
    tas    : a \ast s
    xc,yc  : PSF center position
    x0,y0  : position
    a      : PSF factor factor a (including optional regularization)
    sat    : saturation threshold. Return true if one of the pixels is equal or larger.
    rad    : test radius
    rr     : mixing factor for reference PSF: fr = fr0*(1-rr) + fr1*rr.
    ap,ar,b: result: factor for fpsf, factor for fr, constant b
    result : false if saturation or singularity occurred
}
function SImg_StarSubtract_WidthFilter_approx(const fi2,fpsf,fr0,fr1:TSImg_FPImage; const tas:TSImg_FFTImage; const xc,yc,x0,y0:longint; const a,sat,rad,rr:simgfp; out ap,ar,b:simgfp):boolean; inline;
const wreg  = 0.05;     { weight regularization }
      sqrt2 = sqrt(2);
var a11,a21,a31,a22,a32,a33     : double;
    b1,b2,b3,p,q,r,s,t,w,rsr    : double;
    x,y,xa,xz,ya,yz,xa2,xz2,sy  : longint;
    d,ep,e0,e1                  : ^simgfp;
    f                           : ^fftfp;
begin
    result:=false;

    rsr:=1/sqr(rad);
    x:=round(rad*sqrt2-0.501);
    if x<yc then ya:=y0-x else ya:=y0-yc;
    if ya<0 then ya:=0;
    if x<fpsf.ny-yc then yz:=y0+x else yz:=y0+fpsf.ny-1-yc;
    if yz>=fi2.ny then yz:=fi2.ny-1;
    if x<xc then xa:=x0-x else xa:=x0-xc;
    if xa<0 then xa:=0;
    if x<fpsf.nx-xc then xz:=x0+x else xz:=x0+fpsf.nx-1-xc;
    if xz>=fi2.nx then xz:=fi2.nx-1;

    { calculate matrix + vector }
    b1:=0;  b2:=0;  b3:=0;
    a11:=0; a21:=0; a31:=0;
    a22:=0; a32:=0; a33:=0;
    for y:=ya to yz do
    begin
        sy:=sqr(y-y0);
        d:=fi2.pt + y*fi2.sl;
        x:=(yc+y-y0)*fpsf.sl + xc-x0;
        ep:=fpsf.pt + x;
        e0:=fr0.pt + x;
        e1:=fr1.pt + x;
        f:=tas.pt + y*tas.nxc*2;
        x:=round(sqrt(2*sqr(rad)-sqr(y-y0))-0.5);
        xa2:=x0-x; if xa2<xa then xa2:=xa;
        xz2:=x0+x; if xz2>xz then xz2:=xz;
        for x:=xa2 to xz2 do
        begin
            q:=d[x];
            if q>sat then exit;
            w:=q*wreg;  if w<simg_mina then w:=simg_mina;
            p:=ep[x];                   { PSF }
            r:=e0[x]*(1-rr) + e1[x]*rr; { reference PSF }
            t:=a*p;     if t<0 then t:=0;
            s:=f[x];    if s<0 then s:=0;
            w:=(t+w)/(s+w);             { weight }
            q+=t-s;                     { corrected fi2 }
            t:=(sy+sqr(x-x0))*rsr;
            if t>1 then w*=2-t;
            a11+=w;
            a21+=p*w;
            a31+=r*w;
            a22+=sqr(p)*w;
            a32+=p*r*w;
            a33+=sqr(r)*w;
            q*=w;
            b1+=q;
            b2+=q*p;
            b3+=q*r;
        end;
    end;
//    printmatrix( solve( rmatrix(3,3, [a11,a21,a31, a21,a22,a32, a31,a32,a33]), rmatrix(1,3, [b1,b2,b3]) ) );
    if a33<simg_mina then exit;
    a33:=1/a33;
    s:=a32*a33; b2-=s*b3; a21-=s*a31;  a22-=s*a32;
    s:=a31*a33; b1-=s*b3; a11-=s*a31;
    r:=a22*a11 - sqr(a21);
    if r<simg_mina then exit;
    r:=1/r;
    b :=(a22*b1 - a21*b2)*r;   { constant = background }
    ap:=(a11*b2 - a21*b1)*r;   { factor for PSF fpsf }
    ar:=(b3-b*a31-ap*a32)*a33; { factor for reference PSF fr }
    result:=true;
end;


{******* SImg_AspectRatio ************************************************}
{ reg in pixels^2 }
function SImg_GetAspectRatio(const m,n,reg:simgfp):double; inline;
var r,s : double;
begin
    r:=2*pi*m;
    s:=sqr(n);
    if s>=r then
    begin
        result:=1;
        exit;
    end;
    r:=(r+reg)/(s+reg);
    result:=r+sqrt(sqr(r)-1);
end;


{******* SImg_StarSubtract_WidthFilter_width ****************************}
{ Width and shape test at (x0,y0)
    fi2    : background corrected input (1 color image)
    fpsf   : input PSF s
    tas    : a \ast s
    x0,y0  : position minimum distance to border must be 1
    a,b    : PSF factor factor a (including optional regularization), background b
    rad    : test radius
    thr    : brightness threshold
    m,n    : output
    result : false if if result is undefined
}
function SImg_StarSubtract_WidthFilter_width(const fi2,fpsf:TSImg_FPImage; const tas:TSImg_FFTImage; const xc,yc,x0,y0:longint; const a,b,rad,thr:simgfp; out m,n:double):boolean; inline;
const sqrt2 = sqrt(2);
var r,s,t,xr,yr,rsr     : double;
    x,y,xa,xz,ya,yz     : longint;
    xa2,xz2,x1,y1,sr    : longint;
    d,e                 : ^simgfp;
    f                   : ^fftfp;
begin
    result:=false;

    rsr:=1/sqr(rad);
    x:=round(rad*sqrt2-0.501);
    if x<yc then ya:=y0-x else ya:=y0-yc;
    if ya<0 then ya:=0;
    if x<fpsf.ny-yc then yz:=y0+x else yz:=y0+fpsf.ny-1-yc;
    if yz>=fi2.ny then yz:=fi2.ny-1;
    if x<xc then xa:=x0-x else xa:=x0-xc;
    if xa<0 then xa:=0;
    if x<fpsf.nx-xc then xz:=x0+x else xz:=x0+fpsf.nx-1-xc;
    if xz>=fi2.nx then xz:=fi2.nx-1;

    { search maximum within 3x3 region }
    s:=0;
    for y:=-1 to 1 do
    begin
        d:=fi2.pt  + (y+y0)*fi2.sl + x0;
        e:=fpsf.pt + (y+yc)*fpsf.sl + xc;
        f:=tas.pt  + (y+y0)*tas.nxc*2 + x0;
        for x:=-1 to 1 do
        begin
            r:=d[x]+a*e[x]-f[x];
            if r>s then s:=r;
        end;
    end;
{    if ((abs(x0-7658*2)<10) and (abs(y0-5739*2)<10)) or
        ((abs(x0-4934*2)<10) and (abs(y0-7488*2)<10))
        then writeln(stderr,'x=',x0,' y=',y0,' a=',real2str(a),' s=',real2str(s),' b=',real2str(b),' thr=',real2str(thr));}
    s:=(s-b)*thr;
    if s<simg_mina then exit;
    s:=2/s;

    { calculate result }
    n:=0; m:=0;
    xr:=0; yr:=0;
    for y:=ya to yz do
    begin
        y1:=y-y0;
        d:=fi2.pt  + y*fi2.sl;
        e:=fpsf.pt + (yc+y-y0)*fpsf.sl + xc-x0;
        f:=tas.pt  + y*tas.nxc*2;
        x:=round(sqrt(2*sqr(rad)-sqr(y-y0))-0.5);
        xa2:=x0-x; if xa2<xa then xa2:=xa;
        xz2:=x0+x; if xz2>xz then xz2:=xz;
        for x:=xa2 to xz2 do
        begin
            r:=(d[x]+a*e[x]-f[x]-b)*s-1;
            if r>0 then
            begin
                if r>1 then r:=1;
                x1:=x-x0;
                sr:=sqr(x1)+sqr(y1);
                t:=sr*rsr;
                if t>1 then r*=2-t;
                n+=r;
                m+=sr*r;
                xr+=x1*r;
                yr+=y1*r;
            end;
        end;
    end;
    if n>simg_mina then
    begin
        r:=1/n;
        xr*=r; yr*=r;
//      write(stderr,'  xc=',xr:0:1,' yc=',yr:0:1,'  ');
        if xr<-0.5 then xr:=-0.5
        else if xr>0.5 then xr:=0.5;
        if yr<-0.5 then yr:=-0.5
        else if yr>0.5 then yr:=0.5;
        m-=(sqr(xr)+sqr(yr))*n;
    end;
    result:=true;
end;


{******* SImg_StarSubtract_WidthFilter_tests *****************************}
procedure SImg_StarSubtract_WidthFilter_tests(cc:CSImg_Config);
const rad = 20;
      rot = -30;        { in degree }
      w1  = 12;
      w2  = 5;
      wr  = 10;
var fpsf,f1,fr  : TSImg_FPImage;
    tas         : TSImg_FFTImage;
    i,x,y       : longint;
    s,c         : double;
    a,b,r       : simgfp;
begin
    if cc=nil then cc:=SImg_Config_Default;

    { f1: constant image }
    f1.pt:=nil;
    SImg_New(cc, f1, 5*rad+1, 7*rad+1, 1);
    for i:=0 to f1.size-1 do
        f1.pt[i]:=1;

    { fpsf: PSF }
    fpsf.pt:=nil;
    SImg_New(cc, fpsf, 6*rad+1, 4*rad+1, 1);
    c:=cos(rot*pi/180);
    s:=sin(rot*pi/180);
    i:=0;
    for y:=-2*rad to 2*rad do
        for x:=-3*rad to 3*rad do
        begin
            fpsf.pt[i]:=exp(-sqr((c*x-s*y)/w1)-sqr((s*x+c*y)/w2));
            i+=1;
        end;
    SImg_WriteImageFile(cc,'corrTest-psf.fits',fpsf);

    { fr: reference PSF }
    fr.pt:=nil;
    SImg_Filter(cc, fpsf,fr, wr);
    SImg_WriteImageFile(cc,'corrTest-r.fits',fr);

    { tas: a*s }
    tas.pt:=nil;
    SImg_New_findbase(cc, tas, fpsf.nx, fpsf.ny);
    SImg_CopyC(cc, fpsf,tas, 0);

    { approx tests  }
    SImg_StarSubtract_WidthFilter_approx(f1  ,fpsf,fr,fr, tas, 3*rad,2*rad, 3*rad,2*rad, 1,10, 2*rad,0.5, a,r,b);
    writeln(stderr,'constant: '+real2str(a)+'  '+real2str(r)+'  '+real2str(b));

    SImg_StarSubtract_WidthFilter_approx(fpsf,fpsf,fr,fr, tas, 3*rad,2*rad, 3*rad,2*rad, 1,10, 2*rad,0.5, a,r,b);
    writeln(stderr,'PSF     : '+real2str(a)+'  '+real2str(r)+'  '+real2str(b));

    SImg_StarSubtract_WidthFilter_approx(fr  ,fpsf,fr,fr, tas, 3*rad,2*rad, 3*rad,2*rad, 1,10, 2*rad,0.5, a,r,b);
    writeln(stderr,'ref. PSF: '+real2str(a)+'  '+real2str(r)+'  '+real2str(b));

    SImg_StarSubtract_WidthFilter_approx(fr  ,fpsf,fr,fr, tas, 3*rad,2*rad, 3*rad,2*rad, 1.5,10, 2*rad,0.5, a,r,b);
    writeln(stderr,'normal. : '+real2str(a)+'  '+real2str(r)+'  '+real2str(b));

    SImg_Duplicate(cc, fpsf,f1, true);
    for i:=0 to f1.size-1 do
        f1.pt[i]+=2+3*fr.pt[i];
    SImg_StarSubtract_WidthFilter_approx(f1  ,fpsf,fr,fr, tas, 3*rad,2*rad, 3*rad,2*rad, 1,10, 2*rad,0.5, a,r,b);
    writeln(stderr,'mix     : '+real2str(a)+'  '+real2str(r)+'  '+real2str(b));

    SImg_Filter(cc, fr,f1, 0.2*wr);
    SImg_StarSubtract_WidthFilter_approx(f1  ,fpsf,fr,fr, tas, 3*rad,2*rad, 3*rad,2*rad, 1,10, 2*rad,0.5, a,r,b);
    writeln(stderr,'wider   : '+real2str(a)+'  '+real2str(r)+'  '+real2str(b));

    SImg_Filter(cc, fpsf,f1, 0.8*wr);
    SImg_StarSubtract_WidthFilter_approx(f1  ,fpsf,fr,fr, tas, 3*rad,2*rad, 3*rad,2*rad, 1,10, 2*rad,0.5, a,r,b);
    writeln(stderr,'narrower: '+real2str(a)+'  '+real2str(r)+'  '+real2str(b));

    { width tests  }
    SImg_StarSubtract_WidthFilter_width(fpsf,fpsf, tas, 3*rad,2*rad, 3*rad,2*rad, 1,0,rad,0.5, s,c);
    writeln(stderr,'width test fpsf: n='+real2str(c,-1)+' m='+real2str(s,-1)+' ar='+real2str(SImg_GetAspectRatio(s,c,1e-3),-4));
    SImg_StarSubtract_WidthFilter_width(fr  ,fpsf, tas, 3*rad,2*rad, 3*rad,2*rad, 1,0,rad,0.5, s,c);
    writeln(stderr,'width test fr  : n='+real2str(c,-1)+' m='+real2str(s,-1)+' ar='+real2str(SImg_GetAspectRatio(s,c,1e-3),-4));

    SImg_Free(cc, fpsf);
    SImg_Free(cc, f1);
    SImg_Free(cc, fr);
    SImg_Free(cc, tas);
end;


{******* SImg_StarSubtract_WidthFilter **********************************}
{   ti,tn  : thread index, thread number
    fi2    : background corrected input (1 color image)
    fpsf   : input PSF s
    r0,r1  : reference PSF in center, at corner
    fa     : approximation factor a
    rm     : mask image
    tas    : a \ast s
    inf    : psf info
    sat    : saturation limit
    nor    : normalization relative to a
    sr0,sr1: search radius in center corners (half the search radius for approximation test
    wthr   : threshold for width/shape test
    rm0,rm1: square root of reference m in center, corner
}
threadprocedure SImg_StarSubtract_WidthFilter_thread(ti:longint; tn:longint; fi2:TSImg_FPImage; fpsf:TSImg_FPImage; fr0:TSImg_FPImage; fr1:TSImg_FPImage; fa:TSImg_FPImage; rm:TSImg_RDImage; tas:TSImg_FFTImage; inf:TSImg_PSFInfo; sat:simgfp; nor:simgfp; sr0:simgfp; sr1:simgfp; wthr:simgfp; rm0:simgfp; rm1:simgfp);
const breg   = 0.4;             { Brightness normalization at saturation relative to a }
var a,b,an,a1,a2        : simgfp;
    rr,sr,rsr,ran       : simgfp;
    m,n                 : double;
    i,ia,iz,x0,y0,x,y   : longint;
    xa,xz,ya,yz         : longint;
    e                   : ^simgfp;
    f                   : ^fftfp;

begin
    ia:=(rm.size*ti) div tn;
    iz:=(rm.size*(ti+1)) div tn - 1;
    rsr:=1/(sqr(fi2.nx div 2)+sqr(fi2.ny div 2));
    ran:=inf.max/sat;
    for i:=ia to iz do
    begin
        if (rm.ptb[i] and 6)=2 then
        begin
            y0:=i div fi2.sl;
            x0:=i-y0*fi2.sl;
            if (x0<=0) or (y0<=0) or (x0>=fi2.nx-1) or (y0>=fi2.ny-1) then continue;
            rr:=(sqr(x0-fi2.nx div 2)+sqr(y0-fi2.nx div 2))*rsr;  { 0 in center, 1 in corner }
            sr:=sr0*(1-rr) + sr1*rr;
            a:=fa.pt[i];
            { normalization for bright pixels }
            if a<=0 then an:=0
            else an:=a*( breg*sqrt(a*ran) + nor);

            { tests }
            if not SImg_StarSubtract_WidthFilter_approx(fi2,fpsf,fr0,fr1, tas, inf.xm,inf.ym, x0,y0, a,sat,sr*2,rr, a1,a2,b) then continue;
{           if ((abs(x0-7658*2)<10) and (abs(y0-5739*2)<10)) or
                ((abs(x0-4934*2)<10) and (abs(y0-7488*2)<10))
                then writeln(stderr,'x=',x0,' y=',y0,' b=',real2str(b),' a1=',real2str(a1),' a2=',real2str(a2),' a=',real2str(a),' an=',real2str(an)); }
            if (a1+an>0) and SImg_StarSubtract_WidthFilter_width(fi2,fpsf, tas, inf.xm,inf.ym, x0,y0, a+an,b, sr,wthr, m,n) then
            begin
{               if ((abs(x0-7658*2)<10) and (abs(y0-5739*2)<10)) or
                    ((abs(x0-4934*2)<10) and (abs(y0-7488*2)<10))
                    then writeln(stderr,'x=',x0,' y=',y0,' b=',real2str(b),' a1=',real2str(a1),' a2=',real2str(a2),' a=',real2str(a),' an=',real2str(an),' m='+real2str(m),' limit='+real2str(sqr(rm0*(1-rr)+rm1*rr)));}
                if m<=sqr(rm0*(1-rr)+rm1*rr) then continue;
            end;

{           if ((abs(x0-7658*2)<10) and (abs(y0-5739*2)<10)) or
               ((abs(x0-4934*2)<10) and (abs(y0-7488*2)<10))
               then writeln('x=',x0,' y=',y0,': rm');}
            { remove source, not totally thread save }
            rm.ptb[i]:=rm.ptb[i] and 249;
            ya:=y0-inf.ym;
            if ya<0 then ya:=0;
            yz:=y0+fpsf.ny-1-inf.ym;
            if yz>=fi2.ny then yz:=fi2.ny-1;
            xa:=x0-inf.xm;
            if xa<0 then xa:=0;
            xz:=x0+fpsf.nx-1-inf.xm;
            if xz>=fi2.nx then xz:=fi2.nx-1;
            for y:=ya to yz do
            begin
                e:=fpsf.pt + (inf.ym+y-y0)*fpsf.sl + inf.xm-x0;
                f:=tas.pt + y*tas.nxc*2;
                for x:=xa to xz do
                    f[x]-=a*e[x];
            end;
            fa.pt[i]:=0;
        end;
    end;
end;

{   fi2    : background corrected input (1 color image)
    fpsf   : input PSF s
    fa     : approximation factor a
    rm     : mask image
    tas    : a \ast s
    inf    : PSF info
    sat    : saturation limit
    tol    : PSF tolerance in center
    tolc   : PSF tolerance in corners
    ff     : filter strength (clipped to valid range of 0..2)
    nor    : normalization relative to a
}
procedure SImg_StarSubtract_WidthFilter(cc:CSImg_Config; fi2,fpsf,fa:TSImg_FPImage; rm:TSImg_RDImage; tas:TSImg_FFTImage; const inf:TSImg_PSFInfo; sat,tol0,tolc,ff,nor:simgfp);
const nmin = 15;               { minimum number of pixels brighter than threshold for width/shape test }
var fr0,fr1     : TSImg_FPImage;
    sr0,sr1     : simgfp;
    i,x,y,j0,j1 : longint;
    r,s,t,xc,yc : double;
    n0,n1,m0,m1 : double;
    wthr        : double;
threadvars;
begin
    if cc=nil then cc:=SImg_Config_Default;
    if not SImg_Recover(cc,fi2) then cc.error('SImg_StarSubtract_WidthFilter: Invalid image: fi2');
    if fi2.nc<>1 then cc.error('SImg_StarSubtract_WidthFilter: Invalid dimensions: fi2');
    if not SImg_Recover(cc,fpsf) then cc.error('SImg_StarSubtract_WidthFilter: Invalid image: fpsf');
    if (fpsf.nc<>1) or (fpsf.nx<9) or (fpsf.ny<9) then cc.error('SImg_StarSubtract_WidthFilter: Invalid dimensions: fpsf');
    if not SImg_Recover(cc,rm) and (rm.fmt<>SImg_RawByte) then cc.error('SImg_StarSubtract_WidthFilter: Invalid image: rm');
    if rm.size<>fi2.size then cc.error('SImg_StarSubtract_WidthFilter: Invalid dimensions: rm');
    if fa.nc<>1 then cc.error('SImg_StarSubtract_WidthFilter: Invalid dimensions: fa');
    if fa.size<>fi2.size then cc.error('SImg_StarSubtract_WidthFilter: Invalid dimensions: fa');
    if not SImg_Integrity(tas) then cc.error('SImg_StarSubtract_WidthFilter: Invalid image: tas');
    if (tas.nxr<fi2.nx) or (tas.ny<fi2.ny) then cc.error('SImg_StarSubtract_WidthFilter: Invalid dimensions: tas');

    if tol0<1 then tol0:=1;
    if tolc<tol0 then tolc:=tol0;
    if ff<0 then ff:=0;
    if ff>2 then ff:=2;

    { reference PSF's }
    fr0.pt:=nil;  fr1.pt:=nil;
    SImg_Filter(cc, fpsf,fr0, SIG_TO_RAD*sqrt(0.25 + sqr(inf.sig)*(sqr(tol0)-1)) );  { 0.25 = 4/12-1/12 }
    SImg_Filter(cc, fpsf,fr1, SIG_TO_RAD*sqrt(0.25 + sqr(inf.sig)*(sqr(tolc)-1)) );  { 0.25 = 4/12-1/12 }

    { width/shape test: wthr, m0, r0 }
    t:=0;
    for i:=0 to fr0.size-1 do
        if fr0.pt[i]>t then t:=fr0.pt[i];
    wthr:=(0.5-ff*0.2)/0.9; { desired brightness threshold for width/shape test }
    repeat
        wthr*=0.9;
        r:=2/(wthr*t);
        n0:=0; m0:=0;
        xc:=0; yc:=0;
        j0:=0;
        i:=0;
        for y:=0 to fr0.ny-1 do
            for x:=0 to fr0.nx-1 do
            begin
                s:=fr0.pt[i]*r-1;
                if s>0 then
                begin
                    if s>1 then s:=1;
                    n0+=s;
                    xc+=x*s;
                    yc+=y*s;
                    m0+=(sqr(x)+sqr(y))*s;
                    j0+=1;
                end;
                i+=1;
            end;
    until (n0>=nmin) or (wthr<1e-6);
    m0-=(sqr(xc)+sqr(yc))/n0;
    sr0:=sqrt(j0/pi);

    { m1, sr1 }
    t:=0;
    for i:=0 to fr1.size-1 do
        if fr1.pt[i]>t then t:=fr1.pt[i];
    r:=2/(wthr*t);
    n1:=0; m1:=0;
    xc:=0; yc:=0;
    j1:=1;
    i:=0;
    for y:=0 to fr1.ny-1 do
        for x:=0 to fr1.nx-1 do
        begin
            s:=fr1.pt[i]*r-1;
            if s>0 then
            begin
                if s>1 then s:=1;
                n1+=s;
                xc+=x*s;
                yc+=y*s;
                m1+=(sqr(x)+sqr(y))*s;
                j1+=1;
            end;
        i+=1;
        end;
    m1-=(sqr(xc)+sqr(yc))/n1;
    sr1:=sqrt(j1/pi);
    cc.debug('SImg_StarSubtract_WidthFilter: sr0='+real2str(sr0,-1)+' sr1='+real2str(sr1,-2)+' wthr='+real2str(wthr*100,-1)+'%  n0='+real2str(n0,-1)+' m0='+real2str(m0,-1)+'  n1='+real2str(n1,-1)+' m1='+real2str(m1,-1));

    m0:=sqrt(m0);
    m1:=sqrt(m1);
    for i:=0 to cc.threads-1 do
        dothread SImg_StarSubtract_WidthFilter_thread(i,cc.threads, fi2,fpsf,fr0,fr1,fa, rm, tas, inf, sat,nor, sr0,sr1 ,wthr,m0,m1);
    threadwait;

    SImg_Free(cc, fr0);
    SImg_Free(cc, fr1);
end;


{******* SImg_SubtractStars_PSFErr ***************************************}
{ Calculate PSF error which is the weighted error of shifted PSF
    fpsf   : input PSF
    fe     : error function
    sig    : RMS width of PSF
    Gaussian curve used as weight function
}
procedure SImg_StarSubtract_PSFErr(cc:CSImg_Config; fpsf:TSImg_FPImage; var fe:TSImg_FPImage; sig,tol:double);
const fn  = 3; // filter steps
var fj,fk               : TSImg_FPImage;
    r                   : double;
    i,j,x,y,xs,ys       : longint;
begin
    if cc=nil then cc:=SImg_Config_Default;
    if not SImg_Recover(cc,fpsf) then cc.error('SImg_StarSubtract_PSFErr: Invalid image: fpsf');
    if fpsf.nc<>1 then cc.error('SImg_StarSubtract_PSFErr: Invalid dimensions: fpsf');
    if tol<1.1 then tol:=1.1;
    fj.pt:=nil;  fk.pt:=nil;
    SImg_Duplicate(cc, fpsf,fj, true);
    { filter in fn steps and use maximum }
    for j:=1 to fn do
    begin
        SImg_Filter(cc, fpsf,fk, SIG_TO_RAD*sig*sqrt(sqr(tol)-1)*j/fn);
        for i:=0 to fj.size-1 do
            if fk.pt[i]>fj.pt[i] then fj.pt[i]:=fk.pt[i];
    end;
    SImg_Free(cc,fk);
    { shift 1 pixel in each direction  }
    for ys:=-1 to 1 do
        for xs:=-1 to 1 do
        begin
            if (ys=0) and (xs=0) then continue;
            j:=fpsf.sl*ys+xs;
            for y:=1 to fpsf.ny-2 do
            begin
                i:=y*fpsf.sl+1;
                for x:=1 to fpsf.nx-2 do
                begin
                    r:=abs(fpsf.pt[i]-fpsf.pt[i+j]);
                    if r>fj.pt[i] then fj.pt[i]:=r;
                    i+=1;
                end;
            end;
        end;
    SImg_Move(cc, fj,fe);
end;


{******* SImg_StarSubtract_PSFCorrection ********************************}
{ Calculate PSF correction
    fr  : residual (input and output)
    fw  : weight, must be between 0 and 1
    fs  : deconvoluted stars
    rad : correction radius
    Residual is approximated by:
        1. W_0.5[ (r*w \ast s^0.5^T) / ( ( w \ast s^0.5^T ) * ( \sum s^0.5 ) ) ] \ast s^0.5
        2. W_1  [ (r*w \ast s^T    ) / ( ( w \ast s^T     ) * ( \sum s     ) ) ] \ast s
        3. W_2  [ (r*w \ast s^2^T  ) / ( ( w \ast s^2^T   ) * ( \sum s^2   ) ) ] \ast s^2
    where W_* are the window functions which subtract the estimated background and sets the function to zero if r>2*rad where r is the radius.
}
procedure SImg_StarSubtract_PSFCorrection(cc:CSImg_Config; fr,fw,fs:TSImg_FPImage; rad:longint);
//define[starSubtract_correct_wiener]
//ifdef[starSubtract_correct_wiener]
const itn = 4;          { correction iterations }
//else
const itn = 5;          { correction iterations }
//endif
      swn = 4;          { underweight and only partially correct brightest swn stars }
      neg_weight = 2;   { weight factor at 0. weight factor at >=sig is 1 }
type star = record
        b   : simgfp;
        i   : longint;
     end;
var ts,ti               : array[0..2] of TSImg_FFTImage;
    fw1,fi,fj,fb        : TSImg_FPImage;
    fbw         : TSImg_FPImage;
//ifdef[debug_starSubtract]
    fc                  : TSImg_FPImage;
//endif
    smax                : array[0..swn] of star;
    x,y,i,j,k,it        : longint;
    x0,y0,xa,ya,xz,yz   : longint;
    r,s,t,u,v           : double;
    a00,a01,a02,a11,a12 : double;
    a22,b0,b1,b2        : double;
    mm,mv               : matrix;
begin
    if cc=nil then cc:=SImg_Config_Default;
    if not SImg_Recover(cc,fr) then cc.error('SImg_StarSubtract_PSFCorrection: Invalid image: fr');
    if (fr.nc<>1) or (fr.ny<12*rad) or (fr.nx<12*rad) then cc.error('SImg_StarSubtract_PSFCorrection: Invalid dimensions: fr');
    if not SImg_Recover(cc,fw) then cc.error('SImg_StarSubtract_PSFCorrection: Invalid image: fw');
    if (fw.nc<>1) or (fw.ny<>fr.ny) or (fw.nx<>fr.nx) then cc.error('SImg_StarSubtract_PSFCorrection: Invalid dimensions: fw');
    if not SImg_Recover(cc,fs) then cc.error('SImg_StarSubtract_PSFCorrection: Invalid image: fs');
    if (fs.nc<>1) or (fs.ny<>fr.ny) or (fs.nx<>fr.nx) then cc.error('SImg_StarSubtract_PSFCorrection: Invalid dimensions: fs');
    if rad<5 then rad:=5;

    { find the swn+1 brightest stars which are not fully corrected }
    for i:=0 to swn do
        smax[i].b:=0;
    k:=0; it:=0;
    for i:=0 to fs.size-1 do
    begin
        r:=fs.pt[i];
        if r>=simg_mina then it+=1;
        if r>smax[k].b then
        begin
            j:=k;
            while (j>0) and (r>smax[j-1].b) do
                j-=1;
            for x:=k downto j+1 do
                smax[x]:=smax[x-1];
            if k<swn then k+=1;
            smax[j].b:=r;
            smax[j].i:=i;
        end;
    end;
    if 20*k>it then k:=round(it*0.05);

    { underweight boundaries }
    fw1.pt:=nil;
    SImg_Duplicate(cc, fw,fw1, true);
    v:=1/sqr(rad);
    for y:=0 to fw1.ny-1 do
    begin
        i:=y*fw1.sl;
        if y+1<rad then y0:=y+1
        else if fw1.ny-y<rad then y0:=fw1.ny-y
        else y0:=rad;
        r:=y0*v;
        for x:=0 to rad-2 do
            fw1.pt[i+x]*=(x+1)*r;
        if y0<>rad then
        begin
            s:=r*rad;
            for x:=rad-1 to fw1.nx-rad do
                fw1.pt[i+x]*=s;
        end;
        for x:=fw1.nx-rad+1 to fw1.nx-1 do
            fw1.pt[i+x]*=(fw1.nx-x)*r;
    end;

    { underweight bright stars }
    v:=smax[k].b;
    u:=1/sqr(rad);
    for i:=0 to k-1 do
    begin
        s:=1-sqr(v/smax[i].b);
        x0:=smax[i].i mod fw1.nx;
        y0:=smax[i].i div fw1.nx;
        xa:=x0-2*rad; if xa<0 then xa:=0;
        xz:=x0+2*rad; if xz>=fw1.nx then xz:=fw1.nx-1;
        ya:=y0-2*rad; if ya<0 then ya:=0;
        yz:=y0+2*rad; if yz>=fw1.ny then yz:=fw1.ny-1;
        for y:=ya to yz do
        begin
            j:=y*fw1.sl;
            for x:=xa to xz do
            begin
                r:=(sqr(x-x0)+sqr(y-y0))*u;
                if r<1 then r:=1
                else if r<4 then r:=sqr(2-sqrt(r))
                else r:=0;
                fw1.pt[x+j]*=1-r*s;
            end;
        end;
    end;
//    SImg_WriteImageFile(cc,'starSubtract-fw1.fits',fw1);

    { FFT of:
      s^0.5 = s_0  -> ts[0]
      s     = s_1  -> ts[1]
      s^2   = s_2  -> ts[2] }
    ts[0].pt:=nil; ts[1].pt:=nil; ts[2].pt:=nil;
    SImg_New_findbase(cc, ts[0], fr.nx+4*rad-2, fr.ny+4*rad-2);
    SImg_New(cc, ts[1], ts[0].nxr,ts[0].ny);
    SImg_New(cc, ts[2], ts[0].nxr,ts[0].ny);
    SImg_CopyR(cc, fs,ts[1], 0, 1-2*rad,1-2*rad);
    t:=smax[k].b;
    for i:=0 to ts[0].ny*ts[0].nxc*2-1 do
    begin
        r:=ts[1].pt[i];
        if r>t then
        begin
            r:=t;
            ts[1].pt[i]:=t;
        end;
        if r>0 then
        begin
            s:=sqrt(r); ts[0].pt[i]:=s;
            r:=sqr(r);  ts[2].pt[i]:=r;
        end;
    end;
    SImg_FFT_r2c(cc,ts[0]);
    SImg_FFT_r2c(cc,ts[1]);
    SImg_FFT_r2c(cc,ts[2]);
    for i:=0 to 2 do
    begin
        ti[i].pt:=nil;
        SImg_New(cc, ti[i], ts[0].nxr,ts[0].ny);
    end;

    { fb:  window function
      fbw: weight function for background approximation
    }
    fb.pt:=nil; fbw.pt:=nil;
    SImg_New(cc, fb, 4*rad-1,4*rad-1, 1 );
    SImg_New(cc, fbw, fb.nx,fb.nx,3);
    u:=1/sqr(rad);
    i:=0;
    for y:=1-2*rad to 2*rad-1 do
    begin
        j:=(y+2*rad-1)*ts[0].nxc*2;
        for x:=1-2*rad to 2*rad-1 do
        begin
            r:=(sqr(x)+sqr(y))*u;
            if r<=1 then s:=1
            else if r>=4 then s:=0
            else s:=2-sqrt(r);
            r*=s;
            fb.pt[i]:=s;
            ti[0].pt[j]:=s;
            fbw.pt[i*3  ]:=r;
            fbw.pt[i*3+1]:=r;
            fbw.pt[i*3+2]:=r;
            i+=1; j+=1;
        end;
    end;
//    SImg_WriteImageFile(cc,'starSubtract-fb.fits',fb);
//    SImg_WriteImageFile(cc,'starSubtract-fbw.fits',fbw);

    { fj[c=0] = w \ast s^0.5^T
      fj[c=1] = w \ast s^T
      fj[c=2] = w \ast s^2^T   }
    fj.pt:=nil;
    SImg_New(cc, fj, fb.nx,fb.ny, 3);
    SImg_CopyC(cc, fw1,ti[0], 0);
    SImg_FFT_r2c(cc, ti[0]);
    for j:=0 to 2 do
    begin
        SImg_MulK_C(cc, ti[0],ts[j], ti[1]);
        SImg_FFT_c2r(cc, ti[1]);
        SImg_Copy(cc, ti[1],fj, j);
    end;
    for j:=0 to 2 do { regularization }
    begin
        r:=0;
        for i:=0 to fj.nx*fj.ny-1 do
        begin
            s:=fj.pt[i*3+j];
            if s>=0 then r+=s
            else fj.pt[i*3+j]:=0;       { may happen due to rounding errors }
        end;
        r/=fj.nx*fj.ny;
        s:=ts[j].pt[0];
        for i:=0 to fj.nx*fj.ny-1 do
        begin
            k:=i*3+j;
//ifdef[starSubtract_correct_wiener]
            fj.pt[k]:=s/(fj.pt[k]+r*1e-2);
//else
            fj.pt[k]:=1/((fj.pt[k]+r*1e-3)*s);
//endif
        end;
    end;

    { iterations }
    am2(mm, zmatrix(3,3));
    am2(mv, zmatrix(1,3));
    fi.pt:=nil;
    SImg_New(cc, fi, fb.nx,fb.ny, 3);
//ifdef[debug_starSubtract]
    fc.pt:=nil;
    SImg_New(cc, fc, fb.nx,fb.ny, 3);
//endif
    for it:=1 to itn do
    begin
        { ti[0] = r*w }
        t:=0; v:=0;
        for y:=0 to ts[0].ny-1 do
        begin
            j:=y*ts[0].nxc*2;
            x:=0;
            if y<fr.ny then
            begin
                i:=y*fr.sl;
                for x:=0 to fr.nx-1 do
                begin
                    r:=fr.pt[i];
                    s:=fw1.pt[i];
                    if r<0 then s*=neg_weight;
                    ti[0].pt[x+j]:=r*s;
//ifdef[debug_starSubtract]
                    t+=sqr(r)*s;
                    v+=s;
//endif
                    i+=1;
                end;
                x:=fr.nx;
            end;
            for x:=x to ts[0].nxr-1 do
                ti[0].pt[x+j]:=0;
        end;

        { convergence check + backup }
//ifdef[debug_starSubtract]
//      cc.info(cc.name+': RMS residual after iteration '+int2str(it-1)+': '+real2str(sqrt(t/v),6)+'   approximation factors: '+real2str(mv.pt[0])+', '+real2str(mv.pt[1])+', '+real2str(mv.pt[2]));
//endif

        { fi[c=0] = r*w \ast s^0.5^T
          fi[c=1] = r*w \ast s^T
          fi[c=2] = r*w \ast s^2 }
        SImg_FFT_r2c(cc, ti[0]);
        for j:=0 to 2 do
        begin
//ifdef[starSubtract_correct_wiener]
            SImg_FFTWiener(cc, 2, ti[0],ti[1], ts[j]);
//else
            SImg_MulK_C(cc, ti[0],ts[j], ti[1]);
//endif
            SImg_FFT_c2r(cc, ti[1]);
            SImg_Copy(cc, ti[1],fi, j);
        end;
//      SImg_WriteImageFile(cc,'starSubtract-fi-'+int2str(it)+'-0.fits',fi);

        {  fi[c=0] = (r*w \ast s^0.5^T) / ( ( w \ast s^0.5^T ) * ( \sum s^0.5 ) )
           fi[c=1] = (r*w \ast s^T    ) / ( ( w \ast s^T     ) * ( \sum s     ) )
           fi[c=2] = (r*w \ast s^2^T  ) / ( ( w \ast s^2^T   ) * ( \sum s^2   ) ) }
         for i:=0 to fi.size-1 do
            fi.pt[i]*=fj.pt[i];
//      SImg_WriteImageFile(cc,'starSubtract-fi-'+int2str(it)+'-1.fits',fi);

        { background subtraction }
        SImg_Approx2(cc, fi,fbw, 4, true);
//      SImg_WriteImageFile(cc,'starSubtract-fi-'+int2str(it)+'-2.fits',fi);

        { window function W_c: background estimation and applying carrier }
        for j:=0 to 2 do
        begin
            { window function }
            for i:=0 to fb.size-1 do
                fi.pt[i*3+j]*=fb.pt[i];

            { ti[j] = fi[c=j] \ast s_j }
            SImg_CopyC(cc, fi,ti[j], j);
            SImg_FFT_r2c(cc, ti[j]);
            SImg_Mul_C(cc, ti[j],ts[j], ti[j]);
            SImg_FFT_c2r(cc,ti[j]);
        end;
//      SImg_WriteImageFile(cc,'starSubtract-fi-'+int2str(it)+'-2.fits',fi);

        { approximation }
        a00:=0; a01:=0; a02:=0;
        a11:=0; a12:=0; a22:=0;
        b0:=0; b1:=0; b2:=0;
        for y:=0 to fr.ny-1 do { calculate vector and matrix on x<=y}
        begin
            j:=y*ts[0].nxc*2;
            i:=y*fr.sl;
            for x:=0 to fr.nx-1 do
            begin
                r:=fr.pt[i];
                s:=fw1.pt[i];
                if r<0 then s*=neg_weight;
                t:=ti[0].pt[j];
                u:=ti[1].pt[j];
                v:=ti[2].pt[j];
                a00+=sqr(t)*s;
                a01+=t*u*s;
                a02+=t*v*s;
                a11+=sqr(u)*s;
                a12+=u*v*s;
                a22+=sqr(v)*s;
                b0+=r*t*s;
                b1+=r*u*s;
                b2+=r*v*s;
                i+=1; j+=1;
            end;
        end;
        mm.pt[0]:=a00;  mm.pt[1]:=a01;  mm.pt[2]:=a02;
        mm.pt[3]:=a01;  mm.pt[4]:=a11;  mm.pt[5]:=a12;
        mm.pt[6]:=a02;  mm.pt[7]:=a12;  mm.pt[8]:=a22;
        mv.pt[0]:=b0;   mv.pt[1]:=b1;   mv.pt[2]:=b2;

//      printmatrix(mv);
//      printmatrix(mm);
        am(mv, solve(mm,mv));
//      printmatrix(mv);
        r:=mv.pt[0];
        s:=mv.pt[1];
        t:=mv.pt[2];
        for y:=0 to fr.ny-1 do
        begin
            j:=y*ts[0].nxc*2;
            i:=y*fr.sl;
            for x:=0 to fr.nx-1 do
            begin
                fr.pt[i]-=r*ti[0].pt[j] + s*ti[1].pt[j] + t*ti[2].pt[j];
                j+=1; i+=1;
            end;
        end;
//ifdef[debug_starSubtract]
        for i:=0 to fc.nx*fc.ny-1 do
        begin
            fc.pt[i*3  ]+=r*fi.pt[i*3  ];
            fc.pt[i*3+1]+=s*fi.pt[i*3+1];
            fc.pt[i*3+2]+=t*fi.pt[i*3+2];
        end;
//endif
    end;
    SImg_Free(cc,fi);
    SImg_Free(cc,fj);
    SImg_Free(cc,fbw);
    SImg_Free(cc,fw1);
    for i:=0 to 2 do
    begin
        SImg_Free(cc,ts[i]);
        SImg_Free(cc,ti[i]);
    end;
    free(mm);
    free(mv);

//ifdef[debug_starSubtract]
    for j:=0 to 2 do
    begin
        for i:=0 to fb.size-1 do
            fb.pt[i]:=fc.pt[i*3+j];
        SImg_WriteImageFile(cc,'starSubtract-psfcorr-'+int2str(j)+'.fits',fb);
    end;
    SImg_Free(cc,fc);
//endif
    SImg_Free(cc,fb);
end;


{******* SImg_StarSubtract **********************************************}
{ Subtract stars
    fi     : input (1 color image)
    fpsf   : input PSF
    rms    : mask input and output. Possible inputs:
        1. undefined
        2. mask
            bit 0: interpolated pixels (output only)
            bit 1: star (input and output)
            bit 2: save star (input and output )
        3. floating point (SImg_RawFP) containing deconvoluted stars that will be brightness filtered
    fob    : output image: background
    fos    : output image: deconvoluted stars
    fod    : output image: difference between input and fob where saturated regions are extrapolated
    ofob   : true: output fob; false: do not touch fob
    ofos   : true: output fos; false: do not touch fos
    ofod   : true: output fod; false: do not touch fod
    getmask: bit 0: find new stars
             bit 1: filter input candidates, otherwise they will be set to save stars.
    tol    : psf width tolerance as factor in center
    tolc   : psf width tolerance as factor in corners
    ff     : Source filter strength >=0, high values ==> less objects are considered as stars
    nf     : Noise correction factor. Usually 1.
    ithr   : Interpolation threshold relative to noise
    iwdt   : Interpolation width
    cwdt   : PSF correction width relative to PSF. Valid 0..1
    tn     : PSF correction tiles into direction of larger dimension; <0.5 to disable PSF correction
}
function SImg_StarSubtract(cc:CSImg_Config; fi,fpsf:TSImg_FPImage; var rms:TSImg_RDImage; var fob,fos,fod:TSImg_FPImage; ofob,ofos,ofod:boolean; getmask:longint; tol,tolc: double; ff,nf,ithr,iwdt,cwdt:double; tn:double):TSImg_StarSubtract_Info;
const itn0      = 3;    { getmask<>0 : 3*itn approximation iterations, first filtering after itn0 iterations
                          getmask=0  : 2*itn approximation iterations }
      il_ext    = 0.25; { Extended interpolation threshold for detection of unlikely results }
      ipol_fd   = 3;    { interpolation filter radius divider }
      sig_thr   = 0.25; { threshold for calculation of background interpolation radius used for estimation of noise floor }
      lc_fact   = 2;    { lower constraint offset relative to sig2 at saturation }
var inf,inf1            : TSImg_PSFInfo;        { info of fpsf and fpsf2 }
    rm                  : TSImg_RDImage;
    fi2                 : TSImg_FPImage;        { background subtracted input }
    fj,fk,fl,fm,fn,fo   : TSImg_FPImage;
    fpsf1,fpsfw,fw      : TSImg_FPImage;
    fk2,fj2,fw2,fk3     : TSImg_FPImage;        { tiles (fk2,fj2,fw2) and mosaic output (fk3) }
    fc11,fc12,fc22      : TSImg_FPImage;        { matrix components }
    ts,tss2,ts2         : TSImg_FFTImage;       { Fourier transforms of s, s*s2 and s2 }
    ti,tj               : TSImg_FFTImage;
    sat,rsat            : double;
    i,j,x,y,y2,it,itn   : longint;
    q,r,s,a,b,c         : double;
    sig,sig2,fr,offs    : double;               { noise estimation, detection threshold for a, filter factor, offset for brightness limit, constraint for b }
    irad,crad           : double;               { max. interpolation radius and PSF correction radius }
    w1,w2,w3,w4,w5      : double;
    w6,w7,w8,w9         : double;
    mb                  : byte;
    tinf                : TSImg_TileInfo;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_StarSubtract: fi='+SImg_InfoStr2(fi)+' fpsf='+SImg_InfoStr2(fpsf)+' rms='+SImg_InfoStr2(rms)+select(ofob,' fob='+SImg_InfoStr2(fob),'')+select(ofos,' fos='+SImg_InfoStr2(fos),'')+select(ofod,' fod='+SImg_InfoStr2(fod),'')+
        ' getmask='+int2str(getmask)+' tolc='+real2str(tolc)+' tol='+real2str(tol)+' ff='+real2str(ff)+' nf='+real2str(nf)+' ithr='+real2str(ithr)+' iwdt='+real2str(iwdt,-2)+' cwdt='+real2str(cwdt,-2)+' tn='+realstr(tn,-1));
    if not SImg_Recover(cc,fi) then cc.error('SImg_StarSubtract: Invalid image: fi');
    if fi.nc<>1 then cc.error('SImg_StarSubtract: Invalid dimensions: fi');
    if not SImg_Recover(cc,fpsf) then cc.error('SImg_StarSubtract: Invalid image: fpsf');
    if (fpsf.nc<>1) or (fpsf.nx<9) or (fpsf.ny<9) then cc.error('SImg_StarSubtract: Invalid dimensions: fpsf');
    if rms.pt<>nil then
    begin
        if (not SImg_Recover(cc,rms)) or ((rms.fmt<>SImg_RawByte) and (rms.fmt<>SImg_RawFP)) then cc.error('SImg_StarSubtract: Invalid image: rm');
        if rms.size<>fi.size then cc.error('SImg_StarSubtract: Invalid dimensions: rm');
    end;
    if ofob and (not SImg_OverwriteAlert(fpsf,fob) and not SImg_OverwriteAlert(fi,fob)) then SImg_Free(cc,fob);
    if ofos and (not SImg_OverwriteAlert(fpsf,fos) and not SImg_OverwriteAlert(fi,fos)) then SImg_Free(cc,fos);
    if ofod and (not SImg_OverwriteAlert(fpsf,fod) and not SImg_OverwriteAlert(fi,fod)) then SImg_Free(cc,fod);

    tol:=abs(tol);
    if tol<1.1 then tol:=1.1;
    if tolc<tol then tolc:=tol;
    if iwdt<-1 then iwdt:=-1;
    if cwdt>1 then cwdt:=1;
    if ithr<1 then ithr:=1;
    if ff<0 then ff:=0;
    if nf<0.1 then nf:=0.1;

    { normalize PSF + PSF info }
    r:=0;
    for i:=0 to fpsf.size-1 do
        r+=fpsf.pt[i];
    if r<simg_mina then cc.error('SImg_StarSubtract: Invalid PSF, sum not positive');
    r:=1/r;
    for i:=0 to fpsf.size-1 do
        fpsf.pt[i]*=r;
    SImg_GetPSFInfo(cc,fpsf,inf);
//ifdef[debug_starSubtract]
    cc.info('PSF : '+SImg_PSFInfoString(inf));
//else
    cc.info(cc.name+': PSF info: '+SImg_PSFInfoString(inf));
//endif

    { fpsfw: PSF weight / window function }
    fpsfw.pt:=nil;
    SImg_New(cc, fpsfw, fpsf.nx,fpsf.ny,1, false);
    for i:=0 to fpsf.size-1 do
    begin
        r:=fpsf.pt[i];
        if r>0 then r:=sqrt(r)
        else r:=0;
        fpsfw.pt[i]:=r;
    end;
    SImg_GetPSFInfo(cc,fpsfw,inf1);
//ifdef[debug_starSubtract]
    SImg_WriteImageFile(cc,'starSubtract-fpsfw.fits',fpsfw);
    cc.info
//else
    cc.debug
//endif
        ('psfw: '+SImg_PSFInfoString(inf1));

    { background subtraction (minimum filter) }
    fpsf1.pt:=nil; fi2.pt:=nil;
    r:=inf.sig;                         { noise filter radius }
    SImg_Filter(cc, fpsf,fpsf1, r);
    SImg_GetPSFInfo(cc,fpsf1,inf1);
    inf.xm:=inf1.xm;    { will be used as center }
    inf.ym:=inf1.ym;
//ifdef[debug_starSubtract]
    cc.info
//else
    cc.debug
//endif
        ('psf1: '+SImg_PSFInfoString(inf1));
    SImg_SubtractBackground(cc,  fi,  fi2, inf1.zrad, sqrt(1+sqr(0.5*SImg_GetPSFWidth(cc,fpsf1,sig_thr))),  r,  sat, sig, offs);
    SImg_Free(cc,fpsf1);
    sig*=nf;
    sig2:=sig*sqrt(2)/sqrt(0.5+sqr(inf.sig));
    fr:=sqrt(2*ff);
//ifdef[debug_starSubtract]
    cc.info
//else
    cc.debug
//endif
        (cc.name+': sig='+real2str(sig,5)+'  sig2='+real2str(sig2,5)+'  offs='+real2str(offs,5)+'  fr='+real2str(fr,-2)+'  sat/sig='+real2str(sat/sig,5));
    result.sig:=sig;
    result.sig2:=sig2;
    result.sat:=sat;

    { input and output star mask, get initial mask }
    fj.pt:=nil; rm.pt:=nil;
    if (rms.pt<>nil) and (rms.fmt=SImg_RawFP) then
    begin
        SImg_New(cc, rm, SImg_RawByte, fi.nx,fi.ny,1, false);
        SImg_New(cc, fj, fi.nx,fi.ny,1, false); { create new instance in order to avoid trouble in the calling code }
        if getmask and 2=0 then mb:=6
        else mb:=2;
        r:=sig2/inf.max;
        for i:=0 to fj.size-1 do
        begin
            s:=rms.ptp[i];
            if s<r then
            begin
                fj.pt[i]:=0;
                rm.ptb[i]:=0;
            end
            else begin
                fj.pt[i]:=s;
                if fi2.pt[i]<sat then rm.ptb[i]:=mb
                else rm.ptb[i]:=6;
            end;
        end;
        SImg_Free(cc,rms);
        rms:=rm;
        if getmask and 1<>0 then SImg_StarSubtract_GetMask(cc, fi2,fpsf, rm, fj, getmask, inf, sat,sig2,offs,ff)
    end else
    begin
        if rms.pt=nil then SImg_New(cc, rms, SImg_RawByte, fi.nx,fi.ny,1, true);
        rm:=rms;
        SImg_StarSubtract_GetMask(cc, fi2,fpsf, rm, fj, getmask, inf, sat,sig2,offs,ff);
    end;
    if getmask<>0 then itn:=3*itn0
    else itn:=2*itn0;
    SImg_StarSubtract_GetMask2(cc, fi2,fpsf,rm, sat);

    { weight image. Same weight as in SImg_GetPSF }
    fw.pt:=nil;
    SImg_New(cc, fw, fi2.nx,fi2.ny,1, false);
    q:=1/(sig*ithr);
    a:=offs+sig;
    for i:=0 to fi2.size-1 do
    begin
        r:=fi2.pt[i];
        if r>sat then fw.pt[i]:=0
        else begin
            r:=(r-a)*q-1;
            if r<0 then fw.pt[i]:=1
            else fw.pt[i]:=1/(1+sqr(r));
        end;
    end;

    { ts, tss2, ts2 }
    cc.info(cc.name+': Calculate Fourier transforms of filters');
    ts2.pt:=nil; tss2.pt:=nil;
    ti.pt:=nil; { used for Fourier transform of sqr(s)*s2 }
    ts.pt:=nil;
    SImg_New_findbase(cc, ts, fi2.nx+fpsf.nx-1, fi2.ny+fpsf.ny-1);
    SImg_New(cc, ti,  ts.nxr,ts.ny);
    SImg_New(cc, ts2, ts.nxr,ts.ny);
    SImg_New(cc, tss2,ts.nxr,ts.ny);
    i:=0;
    for y:=0 to fpsf.ny-1 do
    begin
        y2:=y-inf.ym;
        if y2<0 then y2+=ts.ny;
        for x:=0 to fpsf.nx-1 do
        begin
            j:=x-inf.xm;
            if j<0 then j+=ts.nxr;
            j+=y2*ti.nxc*2;
            r:=fpsf.pt[i];
            s:=fpsfw.pt[i];
            ts.pt[j]:=r;
            ts2.pt[j]:=s;
            tss2.pt[j]:=r*s;
            ti.pt[j]:=sqr(r)*s;
            i+=1;
        end;
    end;
//    SImg_FFTWrite(cc,'starSubtract-ts.fits',ts);
    SImg_FFT_r2c(cc,ts);
    SImg_FFT_r2c(cc,ts2);
    SImg_FFT_r2c(cc,tss2);
    SImg_FFT_r2c(cc,ti);

    { matrix components }
    tj.pt:=nil;
    SImg_New(cc, tj, ts.nxr,ts.ny);
    SImg_Copy(cc, fw,tj, 0);
    SImg_FFT_r2c(cc, tj);

    fc11.pt:=nil; fc12.pt:=nil; fc22.pt:=nil;
    SImg_MulK_C(cc, tj,ti, ti);
    SImg_FFT_c2r(cc,ti);
    SImg_New(cc, fc11, fi.nx,fi.ny,1);
    SImg_Copy(cc, ti,fc11, 0);

    SImg_MulK_C(cc, tj, tss2, ti);
    SImg_FFT_c2r(cc,ti);
    SImg_New(cc, fc12, fi.nx,fi.ny,1);
    SImg_Copy(cc, ti,fc12, 0);

    SImg_MulK_C(cc, tj, ts2, tj);
    SImg_FFT_c2r(cc,tj);
    SImg_New(cc, fc22, fi.nx,fi.ny,1);
    SImg_Copy(cc, tj,fc22, 0);
    SImg_Free(cc, tj);

    { a_0*s }
    SImg_CopyC(cc, fj,ti, 0);
    SImg_FFT_r2c(cc,ti);
    SImg_Mul_C(cc, ti, ts, ti);
    SImg_FFT_c2r(cc,ti);

    { iterations }
    rsat:=1/sat;
    tj.pt:=nil;
    for it:=1 to itn-1 do
    begin
        cc.info(cc.name+': Iteration '+int2str(it)+'/'+int2str(itn-1));
        { calculate a and b at star centers and filter stars }
        for y:=0 to ti.ny-1 do
        begin
            j:=y*ti.nxc*2;
            x:=0;
            if y<fi.ny then
            begin
                i:=y*fi2.sl;
                for x:=0 to fi2.nx-1 do
                    ti.pt[j+x]:=(fi2.pt[i+x]-ti.pt[j+x])*fw.pt[i+x];
                x:=fi2.nx;
            end;
            for x:=x to ts.nxr-1 do
                ti.pt[j+x]:=0;  { weight on borders is 0 }
        end;
        SImg_FFT_r2c(cc,ti);
        SImg_New(cc, tj, ts.nxr,ts.ny);
        SImg_MulK_C(cc, ti, ts2, tj);   { tj=d_2 }
        SImg_FFT_c2r(cc,tj);
        SImg_MulK_C(cc, ti, tss2, ti);  { tj=d_1 }
        SImg_FFT_c2r(cc,ti);
{       for y:=0 to ti.ny-1 do
        begin
            j:=y*ts.nxc*2;
            for x:=0 to ti.nxr-1 do
            begin
                r:=tc11.pt[j];
                s:=tc12.pt[j];
                b:=r*tc22.pt[j]-sqr(s);
                if (abs(b)<1e-100) or (r<1e-100) or (y>=fi2.ny) or (x>=fi2.nx) then
                begin
                    ti.pt[j]:=0;
                    tj.pt[j]:=0;
                end else
                begin
//                  a:=(tc22.pt[j]*ti.pt[j]-tc12.pt[j]*tj.pt[j])/b;
//                  a:=ti.pt[j]/tc11.pt[j];
//                  b:=(tc11.pt[j]*tj.pt[j]-tc12.pt[j]*ti.pt[j])/b;
                    a:=ti.pt[j];
                    b:=(r*tj.pt[j]-s*a)/b;
                    if b<0 then b:=0;
                    a:=(a-b*s)/r;
                    ti.pt[j]:=a;
                    tj.pt[j]:=b;
                end;
                j+=1;
            end;
        end;
        SImg_FFTWrite(cc,'starSubtract-a.fits',ti);
        SImg_FFTWrite(cc,'starSubtract-b.fits',tj);}
        q:=2*it/itn;
        if q>1 then q:=1;
        for y:=0 to ts.ny-1 do
        begin
            j:=y*ts.nxc*2;
            x:=0;
            if (y>0) and (y<fi.ny-1) then { ignore borders }
            begin
                ti.pt[j]:=0;
                tj.pt[j]:=0;
                j+=1;
                i:=y*fj.sl+1;
                for x:=1 to fi.nx-2 do
                begin
                    a:=0;
                    mb:=rm.ptb[i];
                    if mb and 6<>0 then
                    begin
                        r:=fc11.pt[i];
                        s:=fc12.pt[i];
                        b:=r*fc22.pt[i]-sqr(s);
                        if (abs(b)<1e-50) or (abs(r)<1e-50) then rm.ptb[i]:=mb and 249  { disable any star if a and b are undefined }
                        else begin
                            { lower constraint for b }
                            c:=fi2.pt[i]*rsat;
                            if c>1 then c:=1;
                            if c<0 then c:=0;
                            c:=offs + sig2*lc_fact*sqrt(c);
                            { approximate star residual at y by a(y) \ast s(x-y) + b(y) }
                            a:=ti.pt[j];
                            b:=(r*tj.pt[j]-s*a)/b;
                            if b<c then b:=c;
                            a:=(a-b*s)/r;
{                           if ((abs(x-7321)<5) and (abs(y-11436)<5))
                                then writeln(stderr,int2str(x)+','+int2str(y)+': a='+real2str(a+fj.pt[i])+' b='+real2str(b));}
                            if (getmask<>0) and (mb and 4=0) then
                            begin
                                { filter stars }
                                if (a>0) or (it=itn-1) then r:=(a+fj.pt[i])*inf.max
                                else r:=fj.pt[i]*inf.max;
                                w1:=fw.pt[i-fi2.sl-1];  w2:=fw.pt[i-fi2.sl];  w3:=fw.pt[i-fi2.sl+1];
                                w4:=fw.pt[i       -1];  w5:=fw.pt[i       ];  w6:=fw.pt[i       +1];
                                w7:=fw.pt[i+fi2.sl-1];  w8:=fw.pt[i+fi2.sl];  w9:=fw.pt[i+fi2.sl+1];
                                c:=w1+w2+w3+w4+w5+w6+w7+w8+w9;
                                if c>simg_mina then
                                begin
                                    s:=(  fi2.pt[i-fi2.sl-1]*w1 + fi2.pt[i-fi2.sl]*w2 + fi2.pt[i-fi2.sl+1]*w3
                                        + fi2.pt[i       -1]*w4 + fi2.pt[i       ]*w5 + fi2.pt[i       +1]*w6
                                        + fi2.pt[i+fi2.sl-1]*w7 + fi2.pt[i+fi2.sl]*w8 + fi2.pt[i+fi2.sl+1]*w9 )/c;
                                    s-=r;
                                    if b>s then s:=b;  { b>offs ==> s>offs }
                                end else s:=b;
                                s-=offs;
//                              if ((abs(x-7321)<5) and (abs(y-11436)<5))
{                               if ((abs(x-1401)<5) and (abs(y-491)<5))
                                    then writeln(stderr,int2str(x)+','+int2str(y)+': r='+real2str(r)+'   s='+real2str(s)+' q='+real2str(q)+'   fr='+real2str(fr)+' sig2='+real2str(sig2));}
                                if r<q*(sig2+q*fr*s) then
                                begin
                                    rm.ptb[i]:=mb and 249;
                                    a:=0;
                                end;
                            end
                            else begin
                                { positive constraint }
                                c:=fj.pt[i];
                                if a<-c then a:=-c;
                            end;
//                          fj.pt[i+1]:=b;  { address can always be used because adjacent star centers and stars on border are not allowed }
                        end;
                    end;
                    ti.pt[j]:=a;
                    tj.pt[j]:=abs(a);
                    i+=1; j+=1;
                end;
                x:=fi.nx-1;
            end;
            for x:=x to ts.nxr-1 do
            begin
                ti.pt[j]:=0;
                tj.pt[j]:=0;
                j+=1;
            end;
        end;

        { a mixing and iteration: a_k = a_(k-1) + a * |a|*s(0)/(|a| \ast s) where s(0) should be th maximum of s }
        SImg_FFT_r2c(cc,tj);
        SImg_Mul_C(cc, tj, ts, tj);
        SImg_FFT_c2r(cc,tj);
//      SImg_FFTWrite(cc,'starSubtract-as.fits',tj);
        i:=0;
        for y:=0 to fj.ny-1 do
        begin
            j:=y*ts.nxc*2;
            for x:=0 to fj.nx-1 do
            begin
                if rm.ptb[i] and 6<>0 then
                begin
                    r:=ti.pt[j];
                    r:=fj.pt[i]+r*abs(r)*inf.max/tj.pt[j];
                    fj.pt[i]:=r;
                    ti.pt[j]:=r;
                end;
                i+=1; j+=1;
            end;
        end;
        SImg_Free(cc,tj);

        { a_k*s }
        SImg_FFT_r2c(cc,ti);
        SImg_Mul_C(cc, ti, ts, ti);
        SImg_FFT_c2r(cc,ti);

        { width filter }
        if (getmask<>0) and (it>=itn0) then
        begin
            r:=2-it/itn0;
            if r<0 then r:=0;
            SImg_StarSubtract_WidthFilter(cc, fi2,fpsf,fj, rm, ti, inf, sat, tol,tolc,ff, r);
        end;
    end;
    SImg_Free(cc,tss2);
    SImg_Free(cc,fc11);
    SImg_Free(cc,fc12);
    SImg_Free(cc,fc22);
    SImg_Free(cc,fpsfw);
    SImg_Free(cc,fw);

    if (not ofob) and (not ofod) then
    begin
        SImg_Free(cc,ts);
        SImg_Free(cc,ts2);
        SImg_Free(cc,ti);
        SImg_Free(cc,fi2);
        { skip interpolation }
        if not ofos then SImg_Free(cc,fj)
        else begin
            for i:=0 to fj.size-1 do
                if rm.ptb[i] and 6=0 then fj.pt[i]:=0; { makes fj valid }
            SImg_Move(cc,fj,fos);
        end;
        SImg_DecDebugLevel;
        exit;
    end;

    { PSF error, fos }
    SImg_StarSubtract_PSFErr(cc,fpsf,fpsf1, inf.sig,tol);
    irad:=SImg_GetPSFWidth(cc,fpsf ,exp(ln(0.1)*(1.5+iwdt)))*0.5;
    crad:=inf1.zrad*cwdt;
    if (crad<2) and (tn>0.4999) then
    begin
        cc.warning(cc.name+': PSF correction radius to small, disabling PSF correction');
        tn:=0;
    end;
    cc.info(cc.name+': Calculating interpolation and PSF correction masks. Maximum interpolation radius: '+real2str(irad,-1)+'. Correction radius: '+real2str(crad,-1)+'.');
//ifdef[debug_starSubtract]
    SImg_WriteImageFile(cc,'starSubtract-psferr.fits',fpsf1);
//endif
    SImg_Clear(cc,ts);
    SImg_CopyR(cc, fpsf1,ts, 0, -inf.xm,-inf.ym);
    i:=0;
    SImg_Clear(cc,ts2);
    for y:=0 to fj.ny-1 do
    begin
        j:=y*ts.nxc*2;
        for x:=0 to fj.nx-1 do
        begin
            if rm.ptb[i] and 6<>0 then ts2.pt[j]:=fj.pt[i]
            else fj.pt[i]:=0; { makes fj valid }
            i+=1; j+=1;
        end;
    end;
    SImg_FFT_r2c(cc,ts);
    SImg_FFT_r2c(cc,ts2);
    SImg_Mul_C(cc, ts, ts2, ts2);
    SImg_FFT_c2r(cc,ts2);
    SImg_Free(cc,ts);
    SImg_Free(cc,fpsf1);

    { limiting interpolation size }
    fk.pt:=nil;  fl.pt:=nil;
    SImg_New(cc, fk, fj.nx,fj.ny,1);
    i:=0;
    for y:=0 to fj.ny-1 do
    begin
        j:=y*ts.nxc*2;
        for x:=0 to fj.nx-1 do
        begin
            r:=ts2.pt[j];
            if r<0 then r:=0; { may happen due to FFT rounding errors }
            fk.pt[i]:=r;
            i+=1; j+=1;
        end;
    end;
    SImg_Free(cc,ts2);
    SImg_FBGFilter(cc, fk,fw, irad, true, false);
    SImg_FBGFilter(cc, fw,fl, crad, true, true);  { fl<=fw }

    { interpolation mask and weight for PSF correction }
    a:=sig*ithr;
    b:=0.25*sig;
    fm.pt:=nil;  SImg_New(cc, fm, fi.nx,fi.ny,1);       { lower constraint }
    fn.pt:=nil;  SImg_New(cc, fn, fi.nx,fi.ny,1);       { weight component }
    for y:=0 to fj.ny-1 do
    begin
        i:=y*fj.sl;
        j:=y*ts.nxc*2;
        for x:=0 to fj.nx-1 do
        begin
            mb:=rm.ptb[i] and 230;
            q:=fk.pt[i];
            r:=q-fw.pt[i];
            c:=q-fl.pt[i];      { q >= c >= r }
            if c>b then mb:=mb or 8;                    { interpolation pass 1: PSF correction }
            if (r>a*il_ext) or (c>a) then mb:=mb or 16; { interpolation pass 2: detection of unlikely values: negative or differs to much form 1st pass interpolation }
            if c>2*b then c:=2*b;
            if r<=a then c*=a
            else begin
                mb:=mb or 1;                            { interpolation pass 3: final }
                c*=2*a-r;
            end;
            r:=fi2.pt[i];
            if r>sat then
            begin
                c:=0;
                if mb and 32=0 then mb:=mb or 1;
            end;
            if c<simg_mina then fw.pt[i]:=0
            else fw.pt[i]:=c;
            s:=r-ti.pt[j];      { residual }
// define[starSubtract_ipolTest]
//ifdef[starSubtract_ipolTest]
//          rm.ptb[i]:=mb or ((mb and 8) shr 3);
            rm.ptb[i]:=mb or ((mb and 16) shr 4);
            fm.pt[i]:=-simg_large;
//else
            rm.ptb[i]:=mb;
            fm.pt[i]:=s;
//endif
            fk.pt[i]:=s;
            fl.pt[i]:=s;
//          s-=offs;
            s:=(s-offs)*2;
//          q:=q*2+sig;
            q+=sig;
            if s<=q then s:=q;
            fn.pt[i]:=sqr(q)/(sqr(q)+sqr(s-q));
            i+=1; j+=1;
        end;
    end;
    if not ofod then SImg_Free(cc,ti);

//ifndef[starSubtract_ipolTest]
    cc.info(cc.name+': Calculating lower constraint');
    SImg_InterpolateMasked(cc, fl, rm, 1+8, ipol_fd);
    SImg_InterpolateMasked(cc, fm, rm, 1+16, ipol_fd);
    for i:=0 to fk.size-1 do
    begin
        r:=fl.pt[i];
        if r<fm.pt[i] then fm.pt[i]:=r;
// define[starSubtract_corrTest]
//ifdef[starSubtract_corrTest]
        if fw.pt[i]<simg_mina then
        begin
            fl.pt[i]:=fk.pt[i];
            fk.pt[i]:=0
        end else
//endif
        fk.pt[i]-=r;
    end;

    { psf correction }
    if tn>0.4999 then
    begin
        { weight }
        fo.pt:=nil; SImg_Filter(cc, fn,fo, sqrt(2+sqr(SImg_GetPSFWidth(cc, fpsf, 0.2)*0.5)) );
        for i:=0 to fw.size-1 do
        begin
            r:=fn.pt[i];
            s:=fo.pt[i];
            if r<s then fw.pt[i]*=r
            else fw.pt[i]*=s;
        end;
        SImg_Free(cc,fn);
        SImg_Free(cc,fo);
//ifdef[starSubtract_corrTest]
        SImg_WriteImageFile(cc,'starSubtract-corr.fits',fk);
        SImg_WriteImageFile(cc,'starSubtract-corr_w.fits',fw);
//endif

        { tiles }
        if tn<1 then tn:=1;
        if fi.nx>fi.ny then
        begin
            tinf.nx:=round(tn);
            tinf.ny:=round(fi.ny/fi.nx*tn);
        end
        else begin
            tinf.ny:=round(tn);
            tinf.nx:=round(fi.nx/fi.ny*tn);
        end;
        i:=round(inf1.zrad*2);
        j:=(fpsf.nx+fpsf.ny) div 4;
        if i>j then i:=j;
        tinf.bb:=fpsf.nx+fpsf.ny;
        i:=fk.nx div (tinf.nx*2)-1;
        if tinf.bb>i then tinf.bb:=i;
        i:=fk.ny div (tinf.ny*2)-1;
        if tinf.bb>i then tinf.bb:=i;
        tinf.bu:=0;
        fk3.pt:=nil;
        if (tinf.nx>1) or (tinf.ny>1) then SImg_New(cc, fk3, fk.nx,fk.ny,1)
        else fk3:=fk;
        fk2.pt:=nil; fw2.pt:=nil; fj2.pt:=nil;
        for y:=0 to tinf.ny-1 do
            for x:=0 to tinf.nx-1 do
            begin
                cc.info(cc.name+': PSF correction for tile '+int2str(y*tinf.nx+x+1)+' / '+int2str(tinf.nx*tinf.ny));
                SImg_GetTile(cc, tinf, fk,fk2, x,y);
                SImg_GetTile(cc, tinf, fw,fw2, x,y);
                SImg_GetTile(cc, tinf, fj,fj2, x,y);
//              if (x=0) and (y=1) then
                SImg_StarSubtract_PSFCorrection(cc, fk2,fw2,fj2, round(crad+1));
                if fj.pt<>fj2.pt then SImg_Free(cc,fj2);
                if fw.pt<>fw2.pt then SImg_Free(cc,fw2);
                SImg_BlendTile(cc, tinf, fk2,fk3, x,y);
            end;
        if fk2.pt<>fk.pt then SImg_Free(cc, fk2);
        SImg_Move(cc, fk3,fk);

//ifdef[starSubtract_corrTest]
        for i:=0 to fk.size-1 do
            if fw.pt[i]>simg_mina then fw.pt[i]:=fk.pt[i];
        SImg_WriteImageFile(cc,'starSubtract-corr_o.fits',fw);
//endif
    end;
    SImg_Free(cc,fw);
    SImg_Free(cc,fn);
    SImg_Free(cc,fo);
    if not ofos then SImg_Free(cc,fj);

    for i:=0 to fk.size-1 do
    begin
        r:=fk.pt[i]+fl.pt[i];
        fk.pt[i]:=r;
        fl.pt[i]:=r;
    end;

    { 2nd pass interpolation }
    SImg_InterpolateMasked(cc, fl, rm, 1+16, ipol_fd);
//endif

    { 3rd pass interpolation }
    it:=0;
    for i:=0 to fk.size-1 do
    begin
        r:=fm.pt[i];
        if r<offs then r:=offs;
        q:=fl.pt[i];
        if q<offs then q:=offs;
        s:=fk.pt[i];
        mb:=rm.ptb[i];
        if (mb and 17=16) and (abs(s-q)>sig) then
        begin
            mb:=mb or 1;
            rm.ptb[i]:=mb;
        end
        else if s<r then s:=r;
        q:=fi.pt[i]-fi2.pt[i];
        fk.pt[i]:=q+s;
        fm.pt[i]:=q;
        if mb and 1 = 1 then it+=1;
    end;
    SImg_Free(cc,fl);
    cc.info(cc.name+': Interpolation of '+real2str(it/fk.size*100,-1)+'% of all pixels');
    SImg_InterpolateMasked(cc, fk, rm,1, ipol_fd);
    SImg_Free(cc,fi2);

    j:=0;
    s:=random;
    a:=sqrt(3)*sig;
    for i:=0 to fk.size-1 do
    begin
        mb:=rm.ptb[i];
//ifdef[debug_starSubtract]
        rm.ptb[i]:=(mb and 63) or (mb shl 6);
//else
        rm.ptb[i]:=mb and 7;
//endif
        if mb and 1=0 then continue;  { neither interpolated nor lower constraint was violated }
        q:=fk.pt[i];
        r:=q-fm.pt[i];
        if r<0 then continue;
        if r>a then r:=a;
        fk.pt[i]:=q+(2*s-1)*r;
        j+=1;
        if j and 1023<>0 then s:=frac(pi*s+0.876)       { cheap random }
        else s:=random;
    end;

    { output difference and extrapolate saturated areas }
    if ofod then
    begin
        a:=1.2/sat;
        for y:=0 to fj.ny-1 do
        begin
            i:=y*fj.sl;
            j:=y*ts.nxc*2;
            for x:=0 to fj.nx-1 do
            begin
                r:=fi.pt[i]-fk.pt[i];
                if r<0 then r:=0;
                s:=r*a-0.2;
                if s>0 then
                begin
                    q:=ti.pt[j];
                    if s<1 then s:=sqr(s)
                    else s:=1;
                    r:=r*(1-s)+q*s;
                end;
                fm.pt[i]:=r;
                i+=1; j+=1;
            end;
        end;
        SImg_Move(cc,fm,fod);
        SImg_Free(cc,ti);
    end
    else SImg_Free(cc,fm);

    if ofob then SImg_Move(cc,fk,fob)
    else SImg_Free(cc,fk);

    if ofos then SImg_Move(cc,fj,fos)
    else SImg_Free(cc,fj);

    SImg_DecDebugLevel;
end;

