{  umatrix2 -- pascal units for matrix operations (real and complex)
   Version 20210727
   Copyright (C) 2000-2021 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{$mode objfpc}
//ifdef[complex]
//define[@roc][complex]
//define[@troc][complex]
//else
//define[@roc][real]
//define[@troc][matreal]
//endif

//ifdef[complex]
unit ucmatrix2;
//else
unit urmatrix2;
//endif

interface
//ifdef[complex]
const bemath : boolean = true;  { i or j for imaginary unit }
//endif

type matreal = extended;
//ifdef[complex]
    complex = record
        r,i  : matreal;  { real, imaginary }
    end;
//endif
    matrix = record
        { reference counter
          Methods that accept this type should dispose the variable if reference count is 0.
          If methods passes the variable to other methods the reference counter should be increased
          an decreased before exit }
        ref  : longint;
        x,y  : longint;   { columns, rows }
        pt   : ^@troc;
    end;
    errorproc = procedure(l:longint; const s:string);  { l:level (0=error, 1=warning); s:message }

    string31 = string[31];
    string15 = string[15];
    string63 = string[63];

{********* string conversion ***************************************************}
function realstr(d:double;i:longint):string31;       { real to string; i: # of digits }
function realstr(d:double):string31;                 { real to string }
function intstr(i:longint):string15;                 { integer to string }
//ifdef[complex]
function complexstr(c:complex;i:longint):string63;   { complex to string; i: # of digits }
function complexstr(c:complex):string63;             { complex to string }
//endif

{********* error handling ******************************************************}
var lasterror : string;                 { last error message }
procedure error(const s:string);        { s:error message }
procedure warning(const s:string);      { s:error message }
procedure stderror(l:longint; const s:string); { s:error message }
procedure InitErrorProc(ep:errorproc);  { inits new error handler }
procedure RestoreErrorProc;             { restores last error handler }

{********* assignment and genaration *******************************************}
procedure free(var a:matrix);                      { deallocate matrix }

{ standart assignment:
  If referene count of a is 0, the pointer is copied and reference count is set to 1.
  Otherwise new space is allocated, data is copied and reference count is set to 1
  if b.pt<>nil it is disposed;
}
procedure am(var b:matrix;const a:matrix);

{ Assignment for uninitialized variables (b is not disposed):
  If reference count of a is 0, the pointer is copied and reference count is set to 1.
  Otherwise new space is allocated, data is copied and reference count is set to 1
}
procedure am2(var b:matrix;const a:matrix);

function zmatrix(xn,yn:longint):matrix;            { generates matrix with all values zero and reference counter of 0 }
function clone(const a:matrix):matrix;             { generates a clone with reference counter of 0 }
function rmatrix(x,y:longint;dat: array of const):matrix;  {  generates real matrix with with values dat and reference counter of 0 }
function ident(i:longint):matrix;                  { identity matrix of dimension i with reference counter of 0  }
//ifdef[complex]
function cmatrix(x,y:longint;dat: array of const):matrix;  { generates complex matrix with with values dat and reference counter of 0 }
//endif

{********* general *************************************************************}
procedure printmatrix(var out:text; a:matrix; j:longint);   { prints matrix a; j # of digits }
procedure printmatrix(var out:text; a:matrix);              { prints matrix a }
procedure printmatrix(a:matrix; j:longint);   { prints matrix a; j # of digits }
procedure printmatrix(a:matrix);              { prints matrix a }

{********* operations **********************************************************}
function trans(a:matrix):matrix;           { transpose a }
operator * (a,b:matrix) c:matrix;          { matrix multiplication }
operator * (a:matrix;b:matreal) c:matrix;  { multiplies matrix with real }
operator * (b:matreal;a:matrix) c:matrix;  { multiplies matrix with real }
operator + (a,b:matrix) c:matrix;          { matrix addition }
operator - (a,b:matrix) c:matrix;          { matrix subtraction }
operator - (a:matrix) c:matrix;            { negative matrix }
operator / (a:matrix;b:matreal) c:matrix;  { divide matrix by real }
operator = (a,b:matrix) c:boolean;
//ifdef[complex]
operator * (a:matrix;b:complex) c:matrix;  { multiplies matrix with complex }
operator * (b:complex;a:matrix) c:matrix;  { multiplies matrix with complex }
operator / (a:matrix;b:complex) c:matrix;  { divide matrix by complex }
//endif

{********* norms ***************************************************************}
function norm1(a:matrix):matreal;          { Manhattan norm }
function norm2(a:matrix):matreal;          { Euclidic norm }
function norminf(a:matrix):matreal;        { maximum norm }

{********* other operations ****************************************************}
function solve(a,b:matrix):matrix;         { solves ax=b }
function inverse(a:matrix):matrix;         { calculates the inverse matrix of a }
function cholesky(a:matrix):matrix;        { Cholesky decomposition (R such that A=R^T * R, A must be symmetric and positive definite) }
procedure qrhh(a:matrix;var q1,r1:matrix); { QR decomposition (Householder): A = QR }
procedure qr2(a:matrix;var q1,r1:matrix);  { QR decomposition such that Q = AR }
function inverse_l(l:matrix):matrix;       { inverts a left/lower triangle matrix (no check whether triangle matrix) }
function lcdelete(a:matrix;i,j:longint):matrix;  { removes column i and line j }
function det(a:matrix):@troc;              { determinant of a }
function xpoly(a:matrix):matrix;           { characteristic polynomial }

{*******************************************************************************}
{*******************************************************************************}
{********* implementation ******************************************************}
{*******************************************************************************}
{*******************************************************************************}
implementation
const MaxErrorProc=1023;

var errorstack   : array[0..maxerrorproc] of errorproc;
    errorstackpt : longint;


{*******************************************************************************}
{********* macros, inlines *****************************************************}
{*******************************************************************************}
//define[@comp(][,#1,#2)][#0.pt[(#2)*#0.x+#1]]
//ifdef[complex]
//define[@rcomp(][,#1,#2)][#0.pt[(#2)*#0.x+#1].r]
//define[@icomp(][,#1,#2)][#0.pt[(#2)*#0.x+#1].i]
//else
//define[@rcomp(][,#1,#2)][#0.pt[(#2)*#0.x+#1]]
//endif

function min(a,b:longint):longint;{inline;}
begin
    if a<b then min:=a
    else min:=b;
end;

function max(a,b:longint):longint;{inline;}
begin
    if a>b then max:=a
    else max:=b;
end;


{*******************************************************************************}
{********* string conversion ***************************************************}
{*******************************************************************************}
function realstr(d:double;i:longint):string31;
var r  : matreal;
    j,k  : longint;
    s  : string;
begin
    realstr:='';
    r:=1;
    for j:=1 to i do
        r*=10;
    if (abs(d)>=r) or ((abs(d)<0.001) and (d<>0)) then
    begin
        i-=1;
        str(d,s);
        j:=1;
        while (j<=length(s)) and (s[j]<>'.') do
            inc(j);
        k:=1;
        while (k<=length(s)) and (upcase(s[k])<>'E') do
            inc(k);
        if s[k+2]='0' then delete(s,k+2,1);
        if (i<1) and (k<>j) then delete(s,j,k-j)
        else if (i>0) and (k-j-1-i>0) then delete(s,j+1+i,k-j-1-i);
        realstr:=s;
    end
    else begin
        r:=1;
        for j:=0 to i do
        begin
            if abs(d)<r then
            begin
                str(d:0:i-j,realstr);
                break;
            end;
            r*=10;
        end;
    end;
end;

function realstr(d:double):string31;
begin
    realstr:=realstr(d,5);
end;

function intstr(i:longint):string15;
begin
    str(i,intstr);
end;

//ifdef[complex]
function complexstr(c:complex;i:longint):string63;
begin
    complexstr:=realstr(c.r,i);
    if c.i<>0 then
    begin
        if c.i>=0 then complexstr+='+'+realstr(c.i,i)
        else complexstr+=realstr(c.i,i);
        if bemath then complexstr+='i'
        else complexstr+='j';
    end;
end;

function complexstr(c:complex):string63;
begin
    complexstr:=complexstr(c,4);
end;
//endif


{*******************************************************************************}
{********* error handling ******************************************************}
{*******************************************************************************}
procedure error(const s:string);
begin
    lasterror:=s;
    if (errorstackpt>=0) and (errorstack[errorstackpt]<>nil)
    then errorstack[errorstackpt](0,s);
end;

procedure warning(const s:string);
begin
    lasterror:=s;
    if (errorstackpt>=0) and (errorstack[errorstackpt]<>nil)
    then errorstack[errorstackpt](1,s);
end;

procedure stderror(l:longint;const s:string);
begin
    writeln(stderr,s);
    if l<=0 then halt(1);
end;

procedure faterror(const s:string);
begin
    writeln(stderr,'Fatal: ',s);
    halt(2);
end;

procedure InitErrorProc(ep:errorproc);
begin
    if (errorstackpt<0) then faterror('InitErrorProc: error stack pointer underflow');
    if (errorstackpt>=maxerrorproc) then faterror('InitErrorProc: error stack pointer overflow');
    inc(errorstackpt);
    errorstack[errorstackpt]:=ep;
    lasterror:='';
end;

procedure RestoreErrorProc;
begin
    if (errorstackpt<=0) then faterror('RestoreErrorProc: error stack pointer underflow');
    if (errorstackpt>maxerrorproc) then faterror('RestoreErrorProc: error stack pointer overflow');
    errorstack[errorstackpt]:=nil;
    dec(errorstackpt);
    lasterror:='';
end;


{*******************************************************************************}
{********* assignment and generation *******************************************}
{*******************************************************************************}
procedure free(var a:matrix);
begin
    if a.pt<>nil then
    begin
        freemem(a.pt);
        a.pt:=nil;
    end;
    a.ref:=-1;
end;

procedure am(var b:matrix; const a:matrix);
begin
    if b.pt<>nil then freemem(b.pt);
    b.x:=a.x;
    b.y:=a.y;
    b.ref:=1;
    if a.ref<=0 then b.pt:=a.pt
    else begin
        getmem(b.pt, b.x*b.y*sizeof(@troc));
        move(a.pt^, b.pt^, b.x*b.y*sizeof(@troc));
    end;
end;

procedure am2(var b:matrix; const a:matrix);
begin
    b.x:=a.x;
    b.y:=a.y;
    b.ref:=1;
    if a.ref<=0 then b.pt:=a.pt
    else begin
        getmem(b.pt, b.x*b.y*sizeof(@troc));
        move(a.pt^, b.pt^, b.x*b.y*sizeof(@troc));
    end;
end;

function zmatrix(xn,yn:longint):matrix;
var j : longint;
begin
     getmem(zmatrix.pt,xn*yn*sizeof(@troc));
     for j:=0 to xn*yn-1 do
     begin
//ifdef[complex]
        zmatrix.pt[j].r:=0;
        zmatrix.pt[j].i:=0;
//else
        zmatrix.pt[j]:=0;
//endif
    end;
    zmatrix.x:=xn;
    zmatrix.y:=yn;
    zmatrix.ref:=0;
end;

function clone(const a:matrix):matrix;
begin
    getmem(clone.pt, a.x*a.y*sizeof(@troc));
    clone.x:=a.x;
    clone.y:=a.y;
    clone.ref:=0;
    move(a.pt^, clone.pt^, a.x*a.y*sizeof(@troc));
end;

function rmatrix(x,y:longint; dat:array of const):matrix;
var j : longint;
begin
    if high(dat)<>x*y-1 then warning('RMatrix: illegal number of arguments');
    rmatrix:=zmatrix(x,y);
    for  j:=0 to x*y-1 do
    begin
        case dat[j].vtype of
//ifdef[complex]
            vtinteger  : rmatrix.pt[j].r:=dat[j].vinteger;
            vtextended : rmatrix.pt[j].r:=dat[j].vextended^;
//else
            vtinteger  : rmatrix.pt[j]:=dat[j].vinteger;
            vtextended : rmatrix.pt[j]:=dat[j].vextended^;
//endif
            else warning('RMatrix: illegal parameter no. '+intstr(j+1));
        end;
    end;
end;

function ident(i:longint):matrix;
var j : longint;
begin
    if i<1 then error('ident: dimesion must be at least 1');
    ident:=zmatrix(i,i);
    for j:=0 to i-1 do
        @rcomp(ident,j,j):=1;
end;

//ifdef[complex]
function cmatrix(x,y:longint;dat: array of const):matrix;
var j : longint;
    r : matreal;
begin
    if high(dat)<>x*y*2-1 then warning('CMatrix: illegal number of parameters');
    cmatrix:=zmatrix(x,y);
    for  j:=0 to x*y*2-1 do
    begin
        r:=0;
        case dat[j].vtype of
            vtinteger  : r:=dat[j].vinteger;
            vtextended : r:=dat[j].vextended^;
            else warning('CMatrix: illegal parameter no. '+intstr(j+1));
        end;
        if (j and 1)=0 then cmatrix.pt[j div 2].r:=r
        else cmatrix.pt[j div 2].i:=r;
    end;
end;
//endif


{*******************************************************************************}
{********* general *************************************************************}
{*******************************************************************************}
procedure printmatrix(var out:text; a:matrix; j:longint);
var i,x,y  : longint;
    s,t    : string;
    b      : byte absolute s;
begin
    i:=0;
    for x:=0 to a.x*a.y-1 do
    begin
        y:=length(@rocstr(a.pt[x],j));
        if y>i then i:=y;
    end;
    fillchar(s,sizeof(s),32);
    for y:=0 to a.y-1 do
    begin
        for x:=0 to a.x-1 do
        begin
            t:=@rocstr(@comp(a,x,y),j);
            b:=i-length(t);
            write(out,s,t,'  ');
        end;
        writeln(out);
    end;
    if a.ref=0 then free(a);
end;

procedure printmatrix(var out:text; a:matrix);
begin
    printmatrix(out,a,4);
end;

procedure printmatrix(a:matrix; j:longint);
begin
    printmatrix(output,a,j);
end;

procedure printmatrix(a:matrix);
begin
    printmatrix(output,a);
end;

{*******************************************************************************}
{********* operations **********************************************************}
{*******************************************************************************}
function trans(a:matrix):matrix;
var x,y : longint;
begin
    trans:=zmatrix(a.y,a.x);
    for y:=0 to a.y-1 do
        for x:=0 to a.x-1 do
            @comp(trans,y,x):=@comp(a,x,y);
    if a.ref=0 then free(a);
end;

operator * (a,b:matrix) c:matrix;
var x,y,i  : longint;
//ifdef[complex]
    r,s,t  : complex;
//else
    r      : matreal;
//endif
begin
    c:=zmatrix(b.x,a.y);
    if b.y<>a.x then error('*: incompatible matrices')
    else begin
        for y:=0 to c.y-1 do
            for x:=0 to c.x-1 do
            begin
//ifdef[complex]
                r.i:=0; r.r:=0;
                for i:=0 to b.y-1 do
                begin
                    s:=@comp(a,i,y);
                    t:=@comp(b,x,i);
                    r.r+=s.r*t.r-s.i*t.i;
                    r.i+=s.r*t.i+s.i*t.r;
                end;
//else
                r:=0;
                for i:=0 to b.y-1 do
                    r+=@comp(a,i,y)*@comp(b,x,i);
//endif
                @comp(c,x,y):=r;
            end;
    end;
    if a.ref=0 then free(a);
    if b.ref=0 then free(b);
end;

operator * (a:matrix;b:matreal) c:matrix;
var i : longint;
begin
    am2(c,a);
    for i:=0 to a.x*a.y-1 do
    begin
//ifdef[complex]
        c.pt[i].r*=b;
        c.pt[i].i*=b;
//else
        c.pt[i]*=b;
//endif
    end;
    c.ref:=0;
end;

operator * (b:matreal;a:matrix) c:matrix;
var i : longint;
begin
    am2(c,a);
    for i:=0 to a.x*a.y-1 do
    begin
//ifdef[complex]
        c.pt[i].r*=b;
        c.pt[i].i*=b;
//else
        c.pt[i]*=b;
//endif
    end;
    c.ref:=0;
end;

operator + (a,b:matrix) c:matrix;
var i  : longint;
begin
    am2(c,a);
    if (b.x<>a.x) or (a.y<>b.y) then error('+: incompatible matrices')
    else begin
        for i:=0 to c.x*c.y-1 do
        begin
//ifdef[complex]
            c.pt[i].r+=b.pt[i].r;
            c.pt[i].i+=b.pt[i].i;
//else
            c.pt[i]+=b.pt[i];
//endif
         end;
        c.ref:=0;
    end;
    if b.ref=0 then free(b);
end;

operator - (a,b:matrix) c:matrix;
var i  : longint;
begin
    am2(c,a);
    if (b.x<>a.x) or (a.y<>b.y) then error('+: incompatible matrices')
    else begin
        for i:=0 to c.x*c.y-1 do
        begin
//ifdef[complex]
            c.pt[i].r-=b.pt[i].r;
            c.pt[i].i-=b.pt[i].i;
//else
            c.pt[i]-=b.pt[i];
//endif
         end;
        c.ref:=0;
    end;
    if b.ref=0 then free(b);
end;

operator - (a:matrix) c:matrix;
var i : longint;
begin
    am2(c,a);
    for i:=0 to a.x*a.y-1 do
    begin
//ifdef[complex]
        c.pt[i].r:=-c.pt[i].r;
        c.pt[i].i:=-c.pt[i].i;
//else
        c.pt[i]:=-c.pt[i];
//endif
    end;
    c.ref:=0;
end;

operator / (a:matrix;b:matreal) c:matrix;
var i : longint;
begin
    b:=1/b;
    am2(c,a);
    for i:=0 to a.x*a.y-1 do
    begin
//ifdef[complex]
        c.pt[i].r*=b;
        c.pt[i].i*=b;
//else
        c.pt[i]*=b;
//endif
    end;
    c.ref:=0;
end;

operator = (a,b:matrix) c:boolean;
var i : longint;
begin
    c:=(a.x=b.x) and (a.y=b.y);
    i:=0;
    while c and (i<a.x*a.y) do
    begin
//ifdef[complex]
        c:=(a.pt[i].r=b.pt[i].r) and (a.pt[i].i=b.pt[i].i);
//else
        c:=a.pt[i]=b.pt[i];
//endif
        i+=1;
    end;
    if a.ref=0 then free(a);
    if b.ref=0 then free(b);
end;

//ifdef[complex]
operator * (a:matrix;b:complex) c:matrix;
var i : longint;
    d : complex;
begin
    am2(c,a);
    for i:=0 to a.x*a.y-1 do
    begin
        d:=a.pt[i];
        c.pt[i].r:=d.r*b.r-d.i*b.i;
        c.pt[i].i:=d.r*b.i+d.i*b.r;
    end;
    c.ref:=0;
end;

operator * (b:complex;a:matrix) c:matrix;
var i : longint;
    d : complex;
begin
    am2(c,a);
    for i:=0 to a.x*a.y-1 do
    begin
        d:=a.pt[i];
        c.pt[i].r:=d.r*b.r-d.i*b.i;
        c.pt[i].i:=d.r*b.i+d.i*b.r;
    end;
    c.ref:=0;
end;

operator / (a:matrix;b:complex) c:matrix;
var i : longint;
    d : complex;
    r : matreal;
begin
    r:=1/(sqr(b.r)+sqr(b.i));
    b.r*=r; b.i*=-r;
    am2(c,a);
    for i:=0 to a.x*a.y-1 do
    begin
        d:=a.pt[i];
        c.pt[i].r:=d.r*b.r-d.i*b.i;
        c.pt[i].i:=d.r*b.i+d.i*b.r;
    end;
    c.ref:=0;
end;
//endif

{*******************************************************************************}
{********* norms ***************************************************************}
{*******************************************************************************}
//ifdef[complex]
function norminf(a:matrix):matreal;
var i   : longint;
    r   : matreal;
begin
    norminf:=0;
    for i:=0 to a.x*a.y-1 do
    begin
        r:=sqr(a.pt[i].i)+sqr(a.pt[i].r);
        if r>norminf then norminf:=r;
    end;
    norminf:=sqrt(norminf);
    if a.ref=0 then free(a);
end;

function norm1(a:matrix):matreal;
var i : longint;
begin
    norm1:=0;
    for i:=0 to a.x*a.y-1 do
        norm1+=sqrt(sqr(a.pt[i].i)+sqr(a.pt[i].r));
    if a.ref=0 then free(a);
end;

function norm2(a:matrix):matreal;
var i : longint;
begin
    norm2:=0;
    for i:=0 to a.x*a.y-1 do
        norm2+=sqr(a.pt[i].i)+sqr(a.pt[i].r);
    norm2:=sqrt(norm2);
    if a.ref=0 then free(a);
end;

//else

function norminf(a:matrix):matreal;
var i   : longint;
begin
    norminf:=0;
    for i:=0 to a.x*a.y-1 do
        if abs(a.pt[i])>norminf then norminf:=abs(a.pt[i]);
    if a.ref=0 then free(a);
end;

function norm1(a:matrix):matreal;
var i : longint;
begin
    norm1:=0;
    for i:=0 to a.x*a.y-1 do
        norm1+=abs(a.pt[i]);
    if a.ref=0 then free(a);
end;

function norm2(a:matrix):matreal;
var i : longint;
begin
    norm2:=0;
    for i:=0 to a.x*a.y-1 do
        norm2+=sqr(a.pt[i]);
    norm2:=sqrt(norm2);
    if a.ref=0 then free(a);
end;
//endif

{*******************************************************************************}
{********* other operations ****************************************************}
{*******************************************************************************}
{********* solve ***************************************************************}
function solve(a,b:matrix):matrix;
var px,py       : ^longint;
    c           : matrix;
    x,y,i,j,k,l : longint;
    r,s         : matreal;
//ifdef[complex]
    d,e,f       : complex;
//endif
begin
    result.pt:=nil;    // makes the compiler happy
    if a.y<>b.y then
    begin
        error('Solve: incompatible matrices');
        exit;
    end;
    solve:=zmatrix(b.x,a.x);
    c:=zmatrix(a.x+b.x,a.y);
    i:=0;
    for y:=0 to c.y-1 do
    begin
        for x:=0 to a.x-1 do
        begin
            c.pt[i]:=@comp(a,x,y);
            i+=1;
        end;
        for x:=0 to b.x-1 do
        begin
            c.pt[i]:=@comp(b,x,y);
            i+=1;
        end;
    end;
    getmem(px,c.x*sizeof(longint));
    getmem(py,a.y*sizeof(longint));
    for x:=0 to c.x-1 do
        px[x]:=x;
    for y:=0 to a.y-1 do
        py[y]:=y;

//ifdef[complex]
    for i:=0 to min(a.x,a.y)-1 do
    begin
        r:=-1; j:=i; k:=i;
        for y:=i to a.y-1 do
            for x:=i to a.x-1 do
            begin
                d:=@comp(c,px[x],py[y]);
                s:=sqr(d.r)+sqr(d.i);
                if s>r then
                begin
                    j:=x; k:=y;
                    r:=s;
                end;
            end;
        l:=px[j]; px[j]:=px[i]; px[i]:=l;
        l:=py[k]; py[k]:=py[i]; py[i]:=l;
        d:=@comp(c,px[i],py[i]);
        if r>0 then r:=1/r;
        e.r:=d.r*r; e.i:=-d.i*r;
        d.r:=1; d.i:=0;
        if r>0 then
        begin
            @comp(c,px[i],py[i]):=d;
            for x:=i+1 to c.x-1 do
            begin
                d:=@comp(c,px[x],py[i]);
                f.r:=d.r*e.r-d.i*e.i;
                f.i:=d.r*e.i+d.i*e.r;
                @comp(c,px[x],py[i]):=f;
            end;
            for y:=i+1 to a.y-1 do
            begin
                e:=@comp(c,px[i],py[y]);
                d.r:=0; d.i:=0; @comp(c,px[i],py[y]):=d;
                for x:=i+1 to c.x-1 do
                begin
                    d:=@comp(c,px[x],py[i]);
                    f:=@comp(c,px[x],py[y]);
                    f.r-=d.r*e.r-d.i*e.i;
                    f.i-=d.r*e.i+d.i*e.r;
                    @comp(c,px[x],py[y]):=f;
                end;
            end;
        end;
    end;

    for x:=0 to solve.x-1 do
    begin
        for y:=a.x to a.y-1 do
        begin
            d:=@comp(c,a.x+x,py[y]);
            if sqr(d.r)+sqr(d.i)>0 then warning('Solve: singularity detected');
        end;
        for y:=min(a.x,a.y)-1 downto 0 do
        begin
            d:=@comp(c,a.x+x,py[y]);
            for i:=y+1 to a.x-1 do
            begin
                e:=@comp(c,px[i],py[y]);
                f:=@comp(solve,x,px[i]);
                d.r-=e.r*f.r-e.i*f.i;
                d.i-=e.r*f.i+e.i*f.r;
            end;
            e:=@comp(c,px[y],py[y]);
            if sqr(e.r)+sqr(e.i)=0 then
            begin
                if (sqr(d.r)+sqr(d.i)>0) then warning('Solve: singularity detected');
                d.r:=1;
                e.i:=0;
            end;
            @comp(solve,x,px[y]):=d;
        end;
    end;

//else

    for i:=0 to min(a.x,a.y)-1 do
    begin
        r:=-1; j:=i; k:=i;
        for y:=i to a.y-1 do
            for x:=i to a.x-1 do
            begin
                s:=@comp(c,px[x],py[y]);
                if abs(s)>r then
                begin
                    j:=x; k:=y;
                    r:=s;
                end;
            end;
            l:=px[j]; px[j]:=px[i]; px[i]:=l;
            l:=py[k]; py[k]:=py[i]; py[i]:=l;
            if abs(r)>0 then
            begin
                r:=1/r;
                @comp(c,px[i],py[i]):=1;
                for x:=i+1 to c.x-1 do
                    @comp(c,px[x],py[i])*=r;
                for y:=i+1 to a.y-1 do
                begin
                    s:=@comp(c,px[i],py[y]);
                    @comp(c,px[i],py[y]):=0;
                    for x:=i+1 to c.x-1 do
                        @comp(c,px[x],py[y])-=@comp(c,px[x],py[i])*s;
                end;
            end;
    end;

    for x:=0 to solve.x-1 do
    begin
        for y:=a.x to a.y-1 do
            if abs(@comp(c,a.x+x,py[y]))>0 then warning('Solve: singularity detected');
        for y:=min(a.x,a.y)-1 downto 0 do
        begin
            r:=@comp(c,a.x+x,py[y]);
            for i:=y+1 to a.x-1 do
                r-=@comp(c,px[i],py[y])*@comp(solve,x,px[i]);
            if @comp(c,px[y],py[y])=0 then
            begin
                if (r<>0) then warning('Solve: singularity detected');
                r:=1;
            end;
            @comp(solve,x,px[y]):=r;
        end;
    end;

//endif

    freemem(px,c.x*sizeof(longint));
    freemem(py,a.y*sizeof(longint));
    free(c);
    if a.ref=0 then free(a);
    if b.ref=0 then free(b);
end;


{********* inverse *************************************************************}
function inverse(a:matrix):matrix;
begin
    if a.y<>a.x then error('Inverse: no square matrix');
    inverse:=solve(a,ident(a.x));
end;


{********* Cholesky ************************************************************}
function cholesky(a:matrix):matrix;       { Cholesky decomposition: A = R^T R }
var l     : matrix;
    c,e   : @troc;
    b     : boolean;
    i,j,k : longint;
//ifdef[complex]
    d,f   : complex;
//endif
begin
    if a<>trans(a) then error('Cholesky: matrix must be symmetrical');
    l:=zmatrix(a.x,a.x);

    b:=true;
    for i:=0 to a.x-1 do
    begin
        c:=@comp(a,i,i);
//ifdef[complex]

        for j:=0 to i-1 do
        begin
            d:=@comp(l,j,i);
            c.r-=sqr(d.r)-sqr(d.i);
            c.i-=2*d.r*d.i;
        end;
        d.r:=sqrt(sqr(c.r)+sqr(c.i));
        c.i:=d.r-c.r; if c.i<0 then c.i:=0; { rounding errors } c.i:=sqrt(0.5*c.i);
        c.r:=d.r+c.r; if c.r<0 then c.r:=0; { rounding errors } c.r:=sqrt(0.5*c.r);
        @comp(l,i,i):=c;
        d.r:=sqr(c.r)+sqr(c.i);
        if d.r<1e-200 then
        begin
            d.r:=0;
            if b then warning('Cholesky: singularity detected');
            b:=false;
        end
        else d.r:=1/d.r;
        c.r*=d.r; c.i*=-d.r;

        for j:=i+1 to a.x-1 do
        begin
            e:=@comp(a,i,j);
            for k:=0 to i-1 do
            begin
                d:=@comp(l,k,i);
                f:=@comp(l,k,j);
                e.r-=d.r*f.r-d.i*f.i;
                e.i-=d.r*f.i+d.i*f.r;
            end;
            d.r:=e.r*c.r-e.i*c.i;
            d.i:=e.r*c.i+e.i*c.r;
            @comp(l,i,j):=d;
        end;

//else

        for j:=0 to i-1 do
            c-=sqr(@comp(l,j,i));
        if c<0 then
        begin
            if b then warning('Cholesky: A is not positive definite or singular');
            c:=0;
        end;
        c:=sqrt(c);
        @comp(l,i,i):=c;
        if c<1e-100 then
        begin
            c:=0;
            if b then warning('Cholesky: singularity detected');
            b:=false;
        end
        else c:=1/c;

        for j:=i+1 to a.x-1 do
        begin
            e:=@comp(a,i,j);
            for k:=0 to i-1 do
                e-=@comp(l,k,i)*@comp(l,k,j);
            @comp(l,i,j):=e*c;
        end;

//endif
    end;

    result:=l;
    if a.ref=0 then free(a);
end;

{********* qrhh ****************************************************************}
procedure qrhh(a:matrix;var q1,r1:matrix);       { QR decomposition (Householder): A = QR }
var q2,r2  : matrix;
    i,j,k  : longint;
    r      : @troc;
//ifdef[complex]
    c,d    : complex;
//endif
begin
    q2:=clone(a);
    r2:=zmatrix(a.x,a.x);
    for k:=0 to a.x-1 do
    begin
//ifdef[complex]
        for j:=0 to k-1 do
        begin
            r.r:=0; r.i:=0;
            for i:=0 to a.y-1 do
            begin
                c:=@comp(q2,k,i);
                d:=@comp(q2,j,i);
                r.r+=c.r*d.r - c.i*d.i;
                r.i+=c.i*d.r + c.r*d.i;
            end;
            for i:=0 to a.y-1 do
            begin
                c:=@comp(q2,j,i);
                @comp(q2,k,i).r-=c.r*r.r - c.i*r.i;
                @comp(q2,k,i).i-=c.i*r.r + c.r*r.i;
            end;
            @comp(r2,k,j):=r
        end;
        r.r:=0; r.i:=0;
        for i:=0 to a.y-1 do
        begin
            c:=@comp(q2,k,i);
            r.r+=sqr(c.r)-sqr(c.i);
            r.i+=2*c.r*c.i;
        end;
{ r:=sqrt(r) }
        d.r:=sqrt(sqr(r.r)+sqr(r.i));
        r.i:=d.r-r.r; if r.i<0 then r.i:=0; { rounding errors } r.i:=sqrt(0.5*r.i);
        r.r:=d.r+r.r; if r.r<0 then r.r:=0; { rounding errors } r.r:=sqrt(0.5*r.r);
        @comp(r2,k,k):=r;
{ r:=1/r }
        d.r:=sqr(r.r)+sqr(r.i);
        if d.r<1e-100 then d.r:=0
        else d.r:=1/d.r;
        r.r*=d.r; r.i*=-d.r;
        for i:=0 to a.y-1 do
        begin
            c:=@comp(q2,k,i);
            @comp(q2,k,i).r:=c.r*r.r - c.i*r.i;
            @comp(q2,k,i).i:=c.i*r.r + c.r*r.i;
        end;

//else

        for j:=0 to k-1 do
        begin
            r:=0;
            for i:=0 to a.y-1 do
                r+=@comp(q2,k,i)*@comp(q2,j,i);
            for i:=0 to a.y-1 do
                @comp(q2,k,i)-=@comp(q2,j,i)*r;
            @comp(r2,k,j):=r;
        end;
        r:=0;
        for i:=0 to a.y-1 do
            r+=sqr(@comp(q2,k,i));
        r:=sqrt(r);
        @comp(r2,k,k):=r;
        if r<1e-100 then r:=0
        else r:=1/r;
        for i:=0 to a.y-1 do
            @comp(q2,k,i)*=r;

//endif
    end;
    if a.ref=0 then free(a);
    am2(q1,q2);
    am2(r1,r2);
end;

{********* qr2 *****************************************************************}
procedure qr2(a:matrix;var q1,r1:matrix);       { QR decomposition such that Q = AR }
var q2,r2  : matrix;
    i,j,k  : longint;
    r,s    : @troc;
//ifdef[complex]
    c,d    : complex;
//endif
begin
    q2:=clone(a);
    r2:=zmatrix(a.x,a.x);
    for k:=0 to a.x-1 do
    begin
//ifdef[complex]

        s.r:=1e-40; s.i:=0;
        for i:=0 to a.y-1 do
        begin
            c:=@comp(q2,k,i);
            s.r+=sqr(c.r)-sqr(c.i);
            s.i+=2*c.r*c.i;
        end;
        s.r:=(sqr(s.r)+sqr(s.i))*1e-12;
        for j:=0 to k-1 do
        begin
            r.r:=0; r.i:=0;
            for i:=0 to a.y-1 do
            begin
                c:=@comp(q2,k,i);
                d:=@comp(q2,j,i);
                r.r+=c.r*d.r - c.i*d.i;
                r.i+=c.i*d.r + c.r*d.i;
            end;
            for i:=0 to a.y-1 do
            begin
                c:=@comp(q2,j,i);
                @comp(q2,k,i).r-=c.r*r.r - c.i*r.i;
                @comp(q2,k,i).i-=c.i*r.r + c.r*r.i;
            end;
            for i:=0 to j do
            begin
                c:=@comp(r2,j,i);
                @comp(r2,k,i).r-=c.r*r.r - c.i*r.i;
                @comp(r2,k,i).i-=c.i*r.r + c.r*r.i;
            end;
        end;
        r.r:=0; r.i:=0;
        for i:=0 to a.y-1 do
        begin
            c:=@comp(q2,k,i);
            r.r+=sqr(c.r)-sqr(c.i);
            r.i+=2*c.r*c.i;
         end;
        { r:=sqrt(r) }
        d.r:=sqrt(sqr(r.r)+sqr(r.i));
        r.i:=d.r-r.r; if r.i<0 then r.i:=0; { rounding errors } r.i:=sqrt(0.5*r.i);
        r.r:=d.r+r.r; if r.r<0 then r.r:=0; { rounding errors } r.r:=sqrt(0.5*r.r);
        { r:=1/r }
        d.r:=sqr(r.r)+sqr(r.i);
        if d.r<s.r then d.r:=0
        else d.r:=1/d.r;
        r.r*=d.r; r.i*=-d.r;
        @comp(r2,k,k):=r;
        for i:=0 to a.y-1 do
        begin
            c:=@comp(q2,k,i);
            @comp(q2,k,i).r:=c.r*r.r - c.i*r.i;
            @comp(q2,k,i).i:=c.i*r.r + c.r*r.i;
        end;
        for i:=0 to k-1 do
        begin
            c:=@comp(r2,k,i);
            @comp(r2,k,i).r:=c.r*r.r - c.i*r.i;
            @comp(r2,k,i).i:=c.i*r.r + c.r*r.i;
        end;

//else

        s:=1e-40;
        for i:=0 to a.y-1 do
            s+=sqr(@comp(q2,k,i));
        s*=1e-12;
        for j:=0 to k-1 do
        begin
            r:=0;
            for i:=0 to a.y-1 do
                r+=@comp(q2,k,i)*@comp(q2,j,i);
            for i:=0 to a.y-1 do
                @comp(q2,k,i)-=@comp(q2,j,i)*r;
            for i:=0 to j do
                @comp(r2,k,i)-=@comp(r2,j,i)*r
        end;
        r:=0;
        for i:=0 to a.y-1 do
            r+=sqr(@comp(q2,k,i));
        if r<s then r:=0
        else r:=1/sqrt(r);
        @comp(r2,k,k):=r;
        for i:=0 to a.y-1 do
            @comp(q2,k,i)*=r;
        for i:=0 to k-1 do
            @comp(r2,k,i)*=r;

//endif
    end;

    if a.ref=0 then free(a);
    am2(q1,q2);
    am2(r1,r2);
end;

{********* inverse_l ***********************************************************}
function inverse_l(l:matrix):matrix;
var i,j,k  : longint;
    b      : matrix;
    c      : @troc;
    ne     : boolean;
//ifdef[complex]
    d      : complex;
//endif
begin
    b:=clone(l);

    ne:=true;
//ifdef[complex]
    for i:=0 to b.x-1 do
    begin
        c:=@comp(b,i,i);
        d.r:=sqr(c.r)+sqr(c.i);
        if abs(d.r)<1e-200 then
        begin
            d.r:=0;
            if ne then warning('inverse_l: singularity detected');
            ne:=false;
        end
        else d.r:=1/d.r;
        c.r*=d.r; c.i*=-d.r;
        @comp(b,i,i):=c;
        for j:=0 to i-1 do
        begin
            d:=@comp(b,j,i);
            @comp(b,j,i).r:=d.r*c.r-d.i*c.i;
            @comp(b,j,i).i:=d.r*c.i+d.i*c.r;
        end;
    end;

    for i:=0 to b.x-1 do
    begin
        for j:=i+1 to b.x-1 do
        begin
            c:=@comp(b,i,j);
            for k:=0 to i-1 do
            begin
                d:=@comp(b,k,i);
                @comp(b,k,j).r-=d.r*c.r-d.i*c.i;
                @comp(b,k,j).i-=d.r*c.i+d.i*c.r;
            end;
            d:=@comp(b,i,i);
            @comp(b,i,j).r:=-d.r*c.r+d.i*c.i;
            @comp(b,i,j).i:=-d.r*c.i-d.i*c.r;
        end;
    end;

//else
    for i:=0 to b.x-1 do
    begin
        c:=@comp(b,i,i);
        if abs(c)<1e-100 then
        begin
            c:=0;
            if ne then warning('inverse_l: singularity detected');
            ne:=false;
        end
        else c:=1/c;
        @comp(b,i,i):=c;
        for j:=0 to i-1 do
            @comp(b,j,i)*=c;
    end;

    for i:=0 to b.x-1 do
    begin
        for j:=i+1 to b.x-1 do
        begin
            c:=@comp(b,i,j);
            for k:=0 to i-1 do
                @comp(b,k,j)-=@comp(b,k,i)*c;
            @comp(b,i,j):=-c*@comp(b,i,i);
        end;
    end;
//endif

    result:=b;
    if l.ref=0 then freemem(l.pt);
end;

{********* lcdelete ************************************************************}
function lcdelete(a:matrix;i,j:longint):matrix;
var x,y,k : longint;
begin
    if (i<0) or (i>=a.x) or (j<0) or (j>=a.y) then
    begin
        warning('lcdelete: arguments out of range');
        lcdelete:=a;
        exit;
    end;

    lcdelete:=zmatrix(a.x-1,a.y-1);
    k:=0;
    for y:=0 to lcdelete.y-1 do
        for x:=0 to lcdelete.x-1 do
        begin
            if y<j then
            begin
                if x<i then lcdelete.pt[k]:=@comp(a,x,y)
                else lcdelete.pt[k]:=@comp(a,x+1,y);
            end
            else begin
                if x<i then lcdelete.pt[k]:=@comp(a,x,y+1)
                else lcdelete.pt[k]:=@comp(a,x+1,y+1);
            end;
            k+=1;
        end;
    if a.ref=0 then freemem(a.pt);
end;


{********* det *****************************************************************}
function det(a:matrix):@troc;
var i   : longint;
//ifdef[complex]
    r,s : complex;
//endif
begin
    inc(a.ref);
//ifdef[complex]
    if a.x<>a.y then
    begin
        det.r:=0;
        det.i:=0;
    end
    else if a.x=1 then det:=a.pt[0]
    else if a.x=2 then
    begin
        det.r:=a.pt[0].r*a.pt[3].r-a.pt[0].i*a.pt[3].i - (a.pt[1].r*a.pt[2].r-a.pt[1].i*a.pt[2].i);
        det.i:=a.pt[0].r*a.pt[3].i+a.pt[0].i*a.pt[3].r - (a.pt[1].r*a.pt[2].i+a.pt[1].i*a.pt[2].r);
    end
    else begin
        det.r:=0;
        det.i:=0;
        for i:=0 to a.x-1 do
        begin
            r:=det(lcdelete(a,i,0));
            s:=@comp(a,i,0);
            if i mod 2=0 then
            begin
                det.r+=r.r*s.r-r.i*s.i;
                det.i+=r.r*s.i+r.i*s.r;
            end
            else begin
                det.r-=r.r*s.r-r.i*s.i;
                det.i-=r.r*s.i+r.i*s.r;
            end;
        end;
    end;
//else
    if a.x<>a.y then det:=0
    else if a.x=1 then det:=a.pt[0]
    else if a.x=2 then det:=a.pt[0]*a.pt[3] - a.pt[1]*a.pt[2]
    else begin
        det:=0;
        for i:=0 to a.x-1 do
            if i mod 2=0 then det+=@comp(a,i,0)*det(lcdelete(a,i,0))
            else det-=@comp(a,i,0)*det(lcdelete(a,i,0));
    end;
//endif
    if a.ref=1 then freemem(a.pt);
end;

{********* xpoly ***************************************************************}
function xpoly_int(a,b:matrix;l:longint):matrix;
var i,j : longint;
    c   : matrix;
    r   : matreal;
begin
    inc(a.ref);
    inc(b.ref);
    c.pt:=nil;
    am2(xpoly_int,zmatrix(l,1));
    if a.x<>a.y then exit
    else if a.x=1 then
    begin
        xpoly_int.pt[1]:=b.pt[0];
        xpoly_int.pt[0]:=a.pt[0];
    end else
    begin
        for i:=0 to a.x-1 do
        begin
            if i mod 2=0 then r:=1
            else r:=-1;
            am(c,xpoly_int(lcdelete(a,i,0),lcdelete(b,i,0),l));
            am(xpoly_int,xpoly_int+(r*c*@comp(a,i,0))); // disposes xpoly_int
            am(c,r*c*@comp(b,i,0));
            for j:=1 to xpoly_int.x-1 do
            begin
//ifdef[complex]
                xpoly_int.pt[j].r+=c.pt[j-1].r;
                xpoly_int.pt[j].i+=c.pt[j-1].i;
//else
                xpoly_int.pt[j]+=c.pt[j-1];
//endif
            end;
        end;
    end;
    free(c);
    if a.ref=1 then freemem(a.pt);
    if b.ref=1 then freemem(b.pt);
    xpoly_int.ref:=0;
end;

function xpoly(a:matrix):matrix;
begin
    xpoly:=xpoly_int(a, -ident(a.x), a.x+1);
end;


{*******************************************************************************}
{********* init / final ********************************************************}
{*******************************************************************************}
initialization
    errorstackpt:=0;
    errorstack[0]:=@stderror;
    lasterror:='';
end.
